package fiber;



import java.io.PrintWriter;
import san.application.IApplication;
import san.core.IApplicationAPI;
import san.core.fiber.api.IInfoAPI;
import san.ident.Guid;



/**
 *
 * @author Jiri Nohavec
 */
public class Fiber implements IApplication {



    private PrintWriter out;



    private void printTitle() {
        out.println("+-------------------------------+");
        out.println("| Fiber scheduler demonstration |");
        out.println("+-------------------------------+");
        out.println();
    }



    private void printHelp() {
        out.println("Usage:");
        out.println("    > run fiber <feature> [thread_count]");
        out.println();
        out.println("Features:");
        out.println("    (S) Sleep ----------- Suspending threads");
        out.println("    (C) Condition ------- Threads waiting on condition");
        out.println("    (M) Monitor --------- Accessing shared context with monitor");
        out.println("    (N) MonitorNo ------- Same as previous, but without monitor");
        out.println("    (D) DeadlockDetect -- Finding monitor deadlocks");
        out.println("    (P) DeadlockPrevent - Preventing monitor deadlocks");
        out.println("    (T) DeadlockTest ---- Acquire monitor with TSL method");
        out.println("    (I) Info ------------ Scheduler information and statistic");
        out.println();
        out.println("Feature real usages:");
        out.println("    (X) PiSingle -------- Separate PI calculating");
        out.println("    (Y) PiMulti --------- Shared PI calculating");
        out.println();
    }



    private void printInfo(IApplicationAPI applicationAPI) {
        IInfoAPI infoAPI = applicationAPI.getInfo();
        out.println("Scheduler info:");
        out.println("    * Applications started count ------- " + infoAPI.getApplicationsStartedCount());
        out.println("    * Applications finished count ------ " + infoAPI.getApplicationsFinishedCount());
        out.println("    * Applications failed count -------- " + infoAPI.getApplicationsFailedCount());
        out.println();
        out.println("    * Capsules started count ----------- " + infoAPI.getCapsulesStartedCount());
        out.println("    * Capsules finished count ---------- " + infoAPI.getCapsulesFinishedCount());
        out.println("    * Capsules failed count ------------ " + infoAPI.getCapsulesFailedCount());
        out.println();
        out.println("    * Special call started count ------- " + infoAPI.getSpecialCallStartedCount());
        out.println("    * Special call finished count ------ " + infoAPI.getSpecialCallFinishedCount());
        out.println("    * Special call failed count -------- " + infoAPI.getSpecialCallFailedCount());
        out.println();
        out.println("    * Startup time --------------------- " + infoAPI.getStartupTime());
        out.println("    * Run time ------------------------- " + infoAPI.getRunTime());
        out.println();
        out.println("    * Physical CPU count --------------- " + infoAPI.getPhysicalCPUCount());
        out.println("    * Virtual CPU count ---------------- " + infoAPI.getVirtualCPUCount());
        out.println("    * Max capsules running ------------- " + infoAPI.getMaxCapsulesRunning());
        out.println();
        out.println("    * Scheduler sleep interval --------- " + infoAPI.getSchedulerSleepInterval() + " [ms]");
        out.println("    * Scheduler strategy --------------- " + infoAPI.getSchedulerStrategy());
        out.println();
        out.println("    * Deadlock detector enabled -------- " + infoAPI.isDeadlockDetectorEnabled());
        out.println("    * Deadlock detector sleep interval - " + infoAPI.getDeadlockDetectorSleepInterval() + " [ms]");
        out.println();
    }



    @Override
    public void main(String[] args, IApplicationAPI applicationAPI, Guid guid) {
        out = applicationAPI.getStdOut();
        if (args.length == 0) {
            printTitle();
            printHelp();
        } else if (args.length > 3) {
            printTitle();
            printHelp();
            out.println("Invalid argument input");
            out.println();
        } else {
            IFeature feature;
            char featureChar = args[0].toUpperCase().charAt(0);
            if (featureChar == 'S') {
                feature = new SleepFeature();
            } else if (featureChar == 'C') {
                feature = new ConditionFeature();
            } else if (featureChar == 'M') {
                feature = new MonitorFeature();
            } else if (featureChar == 'N') {
                feature = new MonitorNoFeature();
            } else if (featureChar == 'D') {
                feature = new DeadlockDetectFeature();
            } else if (featureChar == 'P') {
                feature = new DeadlockPreventFeature();
            } else if (featureChar == 'T') {
                feature = new DeadlockTestFeature();
            } else if (featureChar == 'I') {
                printTitle();
                printInfo(applicationAPI);
                return;
            } else if (featureChar == 'X') {
                feature = new PiSingleFeature();
            } else if (featureChar == 'Y') {
                feature = new PiMultiFeature();
            } else {
                printTitle();
                printHelp();
                out.format("Unknown feature \"%s\"%n", args[0]);
                out.println();
                return;
            }
            if (args.length != 3) {
                printTitle();
                out.println("Description:");
                out.format("    %s%n", feature.getDescription());
                out.println();
            }
            try {
                feature.run(args, applicationAPI, guid);
            } catch (Exception e) {
                throw new RuntimeException(e);
            }
        }
    }
}
