/*************************************************************************
 *                           Smart Active Node                           *
 *                 copyright (c) since 2007 KIV-FAV-ZCU                  *
 *************************************************************************
 * The license for this code is included in LICENSE.txt.  If you did not *
 * get this file with this distribution, request a copy from:            *
 * txkoutny@kiv.zcu.cz                                                   *
 *************************************************************************/
package san.RendezVous;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Map.Entry;
import san.ident.Guid;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import san.core.Kernel;
import san.core.ProcessApplication;
import san.interpreter.sanint.fiber.call.RendezVousCall;

/**
 * Entry call object for directing of rendez-vous
 * @author Miroslav Hendrych
 */
public class Rendezvous {
    HashMap<Guid, IRendezvousServer> tasks;  // rendez-vous servery
    HashMap<Guid, HashSet<Integer>> serverEntryCalls; // jmena jednotlivych entrycall
    HashMap<Guid, LinkedList<EntryCallParams>> calls;
    HashMap<Integer, EntrySet> entryCalls;
    private int klic = 0;



    private final Kernel kernel;



    public Rendezvous(Kernel kernel) {
        this.kernel = kernel;
        tasks = new HashMap<Guid, IRendezvousServer>();
        entryCalls = new HashMap<Integer, EntrySet>();
        serverEntryCalls = new HashMap<Guid, HashSet<Integer>>();
        calls = new HashMap<Guid, LinkedList<EntryCallParams>>();
    }

    public synchronized String[] getEntryCallNamesByGuid(Guid server) {
        HashSet<Integer> klice = serverEntryCalls.get(klic);
        if (klice == null)
            return null;
        String[] names = new String[klice.size()];
        int i=0;
        for (Iterator<Integer> it = klice.iterator(); it.hasNext(); )
            names[i++] = entryCalls.get(it.next()).getName();
        return names;
    }

    public synchronized Guid registerServer(IRendezvousServer task, Guid guid) {
        if (tasks.containsKey(guid))
            return null;
        tasks.put(guid, task);
        calls.put(guid, new LinkedList<EntryCallParams>());
        serverEntryCalls.put(guid, new HashSet<Integer>());
        return guid;
    }

    public synchronized Guid registerServer(IRendezvousServer task) {
        Guid guid = null;
        do {
            guid = Guid.randomGuid();
        } while (tasks.containsKey(guid));
        return registerServer(task, guid);
    }

    public synchronized Integer registerEntryCall(Guid server, String entryName) {
        if (!tasks.containsKey(server))
            return null;
        klic++;
        EntrySet set = new EntrySet(new Integer(klic), server, entryName);
        entryCalls.put(set.getId(), set);
        serverEntryCalls.get(server).add(set.getId());
        return set.getId();
    }

    public synchronized void unregisterServer(Guid guid) {
        if (tasks.remove(guid) != null) {
            HashSet<Integer> idecka = serverEntryCalls.remove(guid);
            for (Iterator<Integer> it = idecka.iterator(); it.hasNext(); )
                entryCalls.remove(it.next());
            idecka.clear();
            List<EntryCallParams> call = calls.remove(guid);
            for (Iterator<EntryCallParams> it = call.iterator(); it.hasNext();) {
                EntryCallParams ecp = it.next();
                ecp.setResult(null);
            }
        }
    }

    public Object callEntryCallByID(Integer entryCallId, Object parameter) throws InterruptedException {
        EntrySet ecs = getEntryCallSet(entryCallId);
        if (ecs == null)
            return null;

        EntryCallParams ecp = new EntryCallParams(ecs.getName(), parameter);
        List<EntryCallParams> list = calls.get(ecs.getServer());
        synchronized (list) {
            list.add(ecp);
            list.notify();
        }

        synchronized (ecp) {
            while (ecp.isWaitingForResult()) {
                ecp.wait();
            }
        }
        return ecp.getResult();
    }

    public Object callEntryCallByName(Guid guid, String entryName, Object parameter) throws InterruptedException {
        Integer entryCallId = getEntryCallId(guid, entryName);
        return callEntryCallByID(entryCallId, parameter);
    }

    public synchronized Integer getEntryCallId(Guid guid, String entryName) {
        HashSet<Integer> serverEntry = serverEntryCalls.get(guid);
        if (serverEntry == null)
            return null;
        for (Iterator<Integer> it = serverEntry.iterator(); it.hasNext(); ) {
            Integer entryId = it.next();
            EntrySet ecs = entryCalls.get(entryId);
            if (ecs.getName().equals(entryName))
                return entryId;
        }
        return null;
    }

    private synchronized EntrySet getEntryCallSet(Integer entryId) {
        return entryCalls.get(entryId);
    }

    public void accept(Guid guid, san.core.Process process) throws InterruptedException {
        List<EntryCallParams> list = calls.get(guid);
        if (list == null)
            return;

        synchronized (list) {
            if (list.isEmpty()) {
                list.wait();
            }
            EntryCallParams ecp = list.remove(0);
            RendezvousServerRef task = (RendezvousServerRef)tasks.get(guid);
            String name = ecp.getEntryName();
//            ClassManager cman = new ClassManager(task.getCodePkg(), process);
//            Object result = null;
//            try {
//                result = cman.runMethod(task.getReference().getClassInfo().getName(), "entry", "(Ljava/lang/String;Ljava/lang/Object;)Ljava/lang/Object;", new Object[]{name, ecp.getParam()}, task.getReference(), false);
//            } catch (Exception ex) {
//                ex.printStackTrace();
//            }
//            ecp.setResult(result);
            ProcessApplication specialProcess = new ProcessApplication(kernel);
            specialProcess.setMainMethodName("entry");
            specialProcess.setMainMethodDescriptor("(Ljava/lang/String;Ljava/lang/Object;)Ljava/lang/Object;");
            specialProcess.setParent(process.getId());
            specialProcess.setSpecialCall(new RendezVousCall(name, ecp.getParam(), task.getReference()));
            kernel.getScheduler().runSpecialProcess(specialProcess, task.getCodePkg());
            ecp.setResult(specialProcess.getReturnValue());
        }
    }

    public void accept(Guid guid, String[] entryNames, san.core.Process process) throws InterruptedException {
        LinkedList<EntryCallParams> list = calls.get(guid);
        if (list == null)
            return;
        HashSet<String> accEntry = new HashSet<String>(Arrays.asList(entryNames));

        synchronized (list) {
            EntryCallParams ecp = null;
            int index = 0;
            do {
                for (Iterator<EntryCallParams> it = list.listIterator(index); it.hasNext(); ) {
                    EntryCallParams call = it.next();
                    if (accEntry.contains(call.getEntryName())) {
                        ecp = call;
                        break;
                    }
                    index++;
                }
                if (ecp == null)
                    list.wait();
            } while (ecp == null);
            list.remove(index);
            RendezvousServerRef task = (RendezvousServerRef)tasks.get(guid);
            String name = ecp.getEntryName();
//            ClassManager cman = new ClassManager(task.getCodePkg(), process);
//            Object result = null;
//            try {
//                result = cman.runMethod(task.getReference().getClassInfo().getName(), "entry", "(Ljava/lang/String;Ljava/lang/Object;)Ljava/lang/Object;", new Object[]{name, ecp.getParam()}, task.getReference(), false);
//            } catch (Exception ex) {
//                ex.printStackTrace();
//            }
//            ecp.setResult(result);
            ProcessApplication specialProcess = new ProcessApplication(kernel);
            specialProcess.setMainMethodName("entry");
            specialProcess.setMainMethodDescriptor("(Ljava/lang/String;Ljava/lang/Object;)Ljava/lang/Object;");
            specialProcess.setParent(process.getId());
            specialProcess.setSpecialCall(new RendezVousCall(name, ecp.getParam(), task.getReference()));
            kernel.getScheduler().runSpecialProcess(specialProcess, task.getCodePkg());
            ecp.setResult(specialProcess.getReturnValue());
        }
    }

    public synchronized Guid getServerGuidByEntryId(Integer entryId) {
        return entryCalls.get(entryId).getServer();
    }

    public synchronized List<Guid> getServerGuidByEntryName(String name) {
        List<Guid> list = new ArrayList<Guid>();
        for (Iterator<Entry<Integer, EntrySet>> it = entryCalls.entrySet().iterator(); it.hasNext(); ) {
            Entry<Integer, EntrySet> entry = it.next();
            if (entry.getValue().getName().equals(name))
                list.add(entry.getValue().getServer());
        }
        return list;
    }

}
