/*************************************************************************
 *                           Smart Active Node                           * 
 *                 copyright (c) since 2007 KIV-FAV-ZCU                  *
 *************************************************************************
 * The license for this code is included in LICENSE.txt.  If you did not *
 * get this file with this distribution, request a copy from:            *
 * txkoutny@kiv.zcu.cz                                                   *
 *************************************************************************/
package san.applicationRepository;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;

import san.localization.Localization;

/**
 * <b>Title:</b> Smart Active Node<br>
 * <b>Description:</b> This class make SAN package available.<br>
 * @author Michal Rejda
 * @since 13.10.2007
 * @version 1.0
 */
public class SanCodePkg extends CodePackage {

	public static final String MANIFEST_ENTRY = "META-INF/MANIFEST.MF";
	public static final String BINARY_ENTRY = "bin";
	public static final String MANIFEST_PROPERTY_SEPARATOR = ":";
	public static final String BIN_EXTENSION = "class";
	public static final String CODE_PKG_CLOSED_EXCEPTION = Localization
			.getLocalization().getString("M000005");
	private ZipFile aplPkgZipFile = null;

	public SanCodePkg(CodePackageSymLink symLink, File aplPkgFile)
			throws AplPkgException {
		super(symLink, aplPkgFile);
		try {
			aplPkgZipFile = new ZipFile(aplPkgFile);
		}
		catch (ZipException e) {
			throw new AplPkgException(e.getMessage());
		}
		catch (IOException e) {
			throw new AplPkgException(e.getMessage());
		}
	}

    /**
     * Access method to Code Package ZipFile
     * @return path to zip file
     */
    public String getAplPkgZipFilename() {
        return aplPkgZipFile.getName();
    }

	@Override
	public String[] getAllFullClassNamesFromPackage() throws AplPkgException {
		if (aplPkgZipFile == null)
			throw new AplPkgException(CODE_PKG_CLOSED_EXCEPTION);
		
		Collection<String> classes = new ArrayList<String>();
		
		Enumeration<? extends ZipEntry> enumeration = aplPkgZipFile.entries();
		ZipEntry zipEntry;
		String fullBinName;
		while (enumeration.hasMoreElements()) {
			  zipEntry = enumeration.nextElement();
			  if(!zipEntry.isDirectory()) {
				  fullBinName = getFullBinNameFromFullName(zipEntry.getName());
				  if(fullBinName != null)
					  classes.add(fullBinName);
			  }
		}
		
		if(classes.size() <= 0)
			return null;
		else
			return (String[]) classes.toArray(new String[0]);
	}
	
	private String getFullBinNameFromFullName(String fullNameOfFile){
		String preffix = "bin/";
		
		if(!fullNameOfFile.startsWith(preffix))
			return null;
		
		try{
			//deleting prefix
			fullNameOfFile = fullNameOfFile.substring(preffix.length());
			//deleting extension
			fullNameOfFile = fullNameOfFile.substring(0,fullNameOfFile.length() - BIN_EXTENSION.length() - 1);
			//creating dots from path separators
			fullNameOfFile = fullNameOfFile.replaceAll("/", "\\.");
			//return full clss name
			return fullNameOfFile;
		}catch(Exception e) {
			return null;
		}
		
	}

	@Override
	public InputStream getClass(String fullClassName)
			throws FileNotFoundException, AplPkgException {
		if (aplPkgZipFile == null)
			throw new AplPkgException(CODE_PKG_CLOSED_EXCEPTION);
		String fullName = fullClassName.replaceAll("\\.", "/");
		try {
			ZipEntry ze = aplPkgZipFile.getEntry(BINARY_ENTRY + "/"
					+ fullName + "." + BIN_EXTENSION);
			if(ze == null)
				throw new IOException();
			return aplPkgZipFile.getInputStream(ze);
		}
		catch (IOException e) {
			Object[] values = new Object[]{fullClassName};
			throw new FileNotFoundException(Localization.getLocalization()
					.getString("M000052", values));
		}
	}

	public InputStream getClass(String packageName, String className)
			throws FileNotFoundException, AplPkgException {
		if (aplPkgZipFile == null)
			throw new AplPkgException(CODE_PKG_CLOSED_EXCEPTION);
		String fullName = packageName.replaceAll("\\.", "/")
				+ "/" + className;
		try {
			ZipEntry ze = aplPkgZipFile.getEntry(BINARY_ENTRY + "/"
					+ fullName + "." + BIN_EXTENSION);
			return aplPkgZipFile.getInputStream(ze);
		}
		catch (IOException e) {
			Object[] values = new Object[]{className, packageName};
			throw new FileNotFoundException(Localization.getLocalization()
					.getString("M000013", values));
		}
	}

	@Override
	public InputStream getApplicationMainClass() throws FileNotFoundException,
			AplPkgException {
		if (aplPkgZipFile == null)
			throw new AplPkgException(CODE_PKG_CLOSED_EXCEPTION);
		String mainClass = getManifestPropertyValue(APP_MAIN_CLASS_MFP);
		if (mainClass == null) {
			Object[] values = new Object[]{APP_MAIN_CLASS_MFP};
			throw new FileNotFoundException(Localization.getLocalization()
					.getString("M000007", values));
		}
		mainClass = mainClass.replaceAll("\\.", "/");
		try {
			ZipEntry ze = aplPkgZipFile.getEntry(BINARY_ENTRY + "/"
					+ mainClass + "." + BIN_EXTENSION);
			return aplPkgZipFile.getInputStream(ze);
		}
		catch (IOException e) {
			Object[] values = new Object[]{mainClass, aplPkgFile.getName()};
			throw new FileNotFoundException(Localization.getLocalization()
					.getString("M000008", values));
		}
	}

	@Override
	public InputStream getCapsuleMainClass() throws FileNotFoundException,
			AplPkgException {
		if (aplPkgZipFile == null)
			throw new AplPkgException(CODE_PKG_CLOSED_EXCEPTION);
		String mainClass = getManifestPropertyValue(CAPSULE_MAIN_CLASS_MFP);
		if (mainClass == null) {
			Object[] values = new Object[]{CAPSULE_MAIN_CLASS_MFP};
			throw new FileNotFoundException(Localization.getLocalization()
					.getString("M000007", values));
		}
		mainClass = mainClass.replaceAll("\\.", "/");
		try {
			ZipEntry ze = aplPkgZipFile.getEntry(BINARY_ENTRY + "/"
					+ mainClass + "." + getExtension());
			return aplPkgZipFile.getInputStream(ze);
		}
		catch (IOException e) {
			Object[] values = new Object[]{mainClass, aplPkgFile.getName()};
			throw new FileNotFoundException(Localization.getLocalization()
					.getString("M000009", values));
		}
	}

	/**
	 * Return manifest property value.
	 */
	public String getManifestPropertyValue(String property)
			throws AplPkgException {
		if (aplPkgZipFile == null)
			throw new AplPkgException(CODE_PKG_CLOSED_EXCEPTION);
		ZipEntry ze = aplPkgZipFile.getEntry(MANIFEST_ENTRY);
		if (ze == null)
			return null;
		InputStream is = null;
		InputStreamReader isr = null;
		BufferedReader br = null;
		try {
			is = aplPkgZipFile.getInputStream(ze);
			isr = new InputStreamReader(is);
			br = new BufferedReader(isr);
			String text = null;
			String value = null;
			while ((text = br.readLine()) != null) {
				if (text.startsWith(property)) {
					value = text.substring(text
							.indexOf(MANIFEST_PROPERTY_SEPARATOR) + 1);
				}
			}
			if (value != null)
				value = value.trim();
			return value;
		}
		catch (Exception e) {
			try {
				if (br != null)
					br.close();
			}
			catch (Exception ex) {
			}
			try {
				if (isr != null)
					isr.close();
			}
			catch (Exception ex) {
			}
			try {
				if (is != null)
					is.close();
			}
			catch (Exception ex) {
			}
			return null;
		}
	}

	@Override
	public void close() throws IOException {
		super.close();
		aplPkgZipFile.close();
		aplPkgZipFile = null;
	}
}
