/*                           Smart Active Node                           * 
 *                 copyright (c) since 2007 KIV-FAV-ZCU                  *
 *************************************************************************
 * The license for this code is included in LICENSE.txt.  If you did not *
 * get this file with this distribution, request a copy from:            *
 * txkoutny@kiv.zcu.cz                                                   *
 *************************************************************************/
package san.codeDistribution.strategies;

import java.io.IOException;
import java.util.Set;

import org.apache.log4j.Logger;

import san.applicationRepository.AplPkgException;
import san.applicationRepository.CodePackage;
import san.applicationRepository.IRepository;
import san.codeDistribution.CapsuleTrafficLogger;
import san.core.Kernel;
import san.data.CodePkgData;
import san.data.TransmitCapsule;
import san.ident.NetIdentifier;
import san.logging.ILogger;
import san.logging.LoggerFactory;
import san.logging.ILogger.LogType;
import san.outgoingQueue.OutgoingQueue;
import san.routing.IRouting;

/*
 * A strategy that looks for an app always on the source node, from where a casule was send.
 */
public class NaiveCodeDistributionStrategy extends CodeDistributionStrategy {

	private static final transient Logger log = Logger.getLogger(NaiveCodeDistributionStrategy.class);
	
	
	public NaiveCodeDistributionStrategy(IRepository repository,
			IRouting router, Kernel kernel, CapsuleTrafficLogger capsuleTrafficLogger) {

		super(repository, router, kernel, capsuleTrafficLogger);		
	}

	@Override
	public void sendRequestForApp(TransmitCapsule originalCapsule) {
		
		if (log.isTraceEnabled()) {
			log.trace("sendRequestForApp(" + originalCapsule + ")");
		}
	
		assert !originalCapsule.isAplication();

		TransmitCapsule requestCapsule = new TransmitCapsule();
		requestCapsule.setSrc(null);
		requestCapsule.setPrev(null);
		requestCapsule.setDest(getBestAppLocation(originalCapsule));
		requestCapsule.setOver(null);
		requestCapsule.setIsRequestForAplication(true);
		requestCapsule.setHash(originalCapsule.getHash());

		kernel.sendCapsule(requestCapsule);
	}

	@Override
	public boolean shouldCapsuleBeStoredInQueue(TransmitCapsule capsule) {
		
		if (((capsule.isRequestForApplication() || capsule.isAplication() || !capsule
				.isDestinationExecute()) && !router.isThisNode(capsule
				.getDest()))) {

			return false;
		} else if (capsule.isRequestForApplication()) {

			return false;
		}

		return true;
	}

	@Override
	public void dealWithCapsuleAfterQueue(TransmitCapsule capsule) {
		if (log.isTraceEnabled()) {
			log.trace("dealWithCapsuleAfterQueue(" + capsule + ")");
		}
		
		if(shouldCapsuleBeStoredInQueue(capsule)){
			// do nothing
		}
		else{			
			if (capsule.isRequestForApplication() && router.isThisNode(capsule.getDest())) {
				sendApplicationFor(capsule);
			} else {
				capsule.setOver(null);
				capsule.setPrev(null);
				
				kernel.sendCapsule(capsule);
				
				if (log.isDebugEnabled()) {
					log.debug("preposilam kapsuli dal");
				}	
			}
		}
		
	}
	
	@Override
	void sendApplicationFor(TransmitCapsule originalCapsule) {
		
		if (log.isTraceEnabled()) {
			log.trace("sendApplicationFor(" + originalCapsule + ")");
		}

		CodePkgData cpd = repository.getCodePkgData(originalCapsule.getHash());
		
		if (cpd != null) {
		
			TransmitCapsule applicationCapsule = new TransmitCapsule();
			applicationCapsule.setSrc(originalCapsule.getDest());
			applicationCapsule.setPrev(null);
			applicationCapsule.setDest(originalCapsule.getSrc());
			applicationCapsule.setOver(router.getDirectionTo(originalCapsule.getSrc()));
			applicationCapsule.setAplication(true);
			applicationCapsule.setHash(originalCapsule.getHash());
			applicationCapsule.setData(cpd);	
					
			kernel.sendCapsule(applicationCapsule);	

		} else {
			log.debug("App not found on this node");
			
			// od 'app' nebo 'data' came from? !!!!
			Set<NetIdentifier> ni = capsuleTrafficLogger.getAppPlacementInfo(originalCapsule.getHash()).getNodesDataCameFrom();
			
			for (NetIdentifier n : ni) {
				
				if(n.compareTo(originalCapsule.getSrc()) != 0){
				
					TransmitCapsule applicationCapsule = new TransmitCapsule();
					applicationCapsule.setSrc(originalCapsule.getSrc());
					applicationCapsule.setPrev(null);
					applicationCapsule.setDest(n);
					applicationCapsule.setOver(null);
					applicationCapsule.setAplication(false);
					applicationCapsule.setIsRequestForAplication(true);
					applicationCapsule.setHash(originalCapsule.getHash());
	
					kernel.sendCapsule(applicationCapsule);
					return;
				}
			}
			
			log.error("No useful information about app " + originalCapsule.getHash() + " on this node");
		}		
	}

	@Override
	NetIdentifier getBestAppLocation(TransmitCapsule capsuleDataArrivedIn) {
		return capsuleDataArrivedIn.getSrc();
	}
}
