/*************************************************************************
 *                           Smart Active Node                           * 
 *                 copyright (c) since 2007 KIV-FAV-ZCU                  *
 *************************************************************************
 * The license for this code is included in LICENSE.txt.  If you did not *
 * get this file with this distribution, request a copy from:            *
 * txkoutny@kiv.zcu.cz                                                   *
 *************************************************************************/
package san.core;

import bsh.This;
import java.io.File;
import java.io.IOException;

import san.Utility;
import san.applicationRepository.HashCalculation;
import san.applicationRepository.IRepository;
import san.data.CodePkgData;
import san.data.TransmitCapsule;
import san.ident.CodePkgIdent;
import san.incomingQueue.IncomingQueue;
import java.util.Collection;

import org.apache.log4j.Logger;

/**
 * <b>Title:</b> Smart Active Node<br>
 * <b>Description:</b> First: This class read capsules with applications from
 * incoming queue. The applications are being stored. After the storing is the
 * Interpreter manager informed of new application in repository.<br>
 * <br>
 * Second: this class contains method to saving application into the repository.<br>
 * @author Michal Rejda
 * @since 21.2.2007
 * @version 1.1
 */
public class CodeStoreManager {

	private static final transient Logger log = Logger.getLogger(CodeStoreManager.class);
	
	/**
	 * @uml.property name="incomingQueue"
	 * @uml.associationEnd multiplicity="(1 1)"
	 *                     inverse="aplStoreManager:san.incomingQueue.IncomingQueue"
	 */
	private IncomingQueue incomingQueue = null;
	/**
	 * @uml.property name="repository"
	 * @uml.associationEnd multiplicity="(1 1)"
	 *                     inverse="aplStoreManager:san.applicationRepository.IRepository"
	 */
	private IRepository repository = null;
	/**
	 * @uml.property name="codeStoreListeners"
	 * @uml.associationEnd multiplicity="(0 -1)" aggregation="shared"
	 *                     inverse="codeStoreManager:san.core.CodeStoredListener"
	 */
	private Collection<CodeStoredListener> codeStoreListeners = new java.util.LinkedList<CodeStoredListener>();
	private Thread readThread;
	private boolean running;

	public CodeStoreManager(IncomingQueue inQueue, IRepository rep) {
		this.incomingQueue = inQueue;
		this.repository = rep;
		this.running = true;
		// create new thread to checking of new application in incoming queue
		this.readThread = new Thread() {

			public void run() {
				while (running) {
					readNewAppFromCapsule(); // read new application from incoming queue
				}
			}
		};
		readThread.start();
	}

	/**
	 * Read new application from incoming queue.
	 */
	private void readNewAppFromCapsule() {
		if (incomingQueue != null) {
			// getting new capsule with program
			TransmitCapsule capsule = incomingQueue.getProgramCode();
			if (capsule != null && repository != null) {
				// store application
				CodePkgData data = (CodePkgData) capsule.getData();
				repository.saveCodePkg(capsule.getHash(), data.getData(), data.getExtension(), false);
				
				if (log.isDebugEnabled()) {
					log.debug("ulozen program " + capsule.getHash());
				}
				
				//fire listeners
//				fireCodeStoredListeners(capsule.getHash());
				fireCodeStoredListeners(capsule);
			}
		}
	}

	/**
	 * Store application from local file into the applications repository.
	 * @param localPath local path to applications file
	 * @return Return stored application identification
	 */
	public CodePkgIdent storeLocalCodeFromFile(String localPath) throws AplStoreException {
		CodePkgIdent ident = null;
		try {
			ident = new CodePkgIdent(computeFileHash(localPath));
			String extension = localPath
					.substring(localPath.lastIndexOf(".") + 1);
			repository.saveCodePkg(ident, Utility.getBytesFromFile(new File(localPath)), extension, true);
			//fire listeners
			fireCodeStoredListeners(ident);
		}
		catch (IOException e) {
			throw new AplStoreException(e.getMessage());
		}
		return ident;
	}

    /**
     * Computes whirlpool hash of file in localPath
     * @param localPath a path to file in local file system
     * @return 16 byte array containing 128bit hash computed by function
     */
	public static byte[] computeFileHash(String localPath) {
		try {
			// calculating hash of application
			byte[] tmp = new byte[CodePkgIdent.LENGHT];
			/* Whirlpool hash */
			byte[] fileBytes = Utility.getBytesFromFile(new File(localPath));
			HashCalculation hash = new HashCalculation();
			/*
			 * initialization of the hash number and calculation of the hash
			 * code
			 */
			hash.NESSIEinit();
			hash.NESSIEadd(fileBytes, (long) fileBytes.length);
			hash.NESSIEfinalize(tmp);

            return tmp;
		} catch (IOException e) {
            log.error(e);
		}
        return null;
	}
	
    public static void main(String[] args) {
        if (args.length == 1) {
            try {
                String hash = HashCalculation.hashToString(computeFileHash(args[0]));
                String newFileName = args[0].substring(0, args[0].indexOf('.')) + '_' + hash + args[0].substring(args[0].indexOf('.'), args[0].length());

                File oldFile = new File(args[0]);

                oldFile.renameTo(new File(newFileName));

                System.out.println(args[0] + " -> " + newFileName);
            } catch (Exception e) {
                log.error(e);
            }
        } else {
            log.error("Please supply a file name to rename.");
        }
    }

	/**
	 * Fire code stored listeners.
	 */
	private void fireCodeStoredListeners(CodePkgIdent code) {	
		for (CodeStoredListener csl : codeStoreListeners) {
			csl.codeStored(code);
		}
	}
	
	private void fireCodeStoredListeners(TransmitCapsule capsule) {
		for (CodeStoredListener csl : codeStoreListeners) {
			csl.codeStored(capsule);
		}
	}
	
	/**
	 * Add new listener for listen of code storing.
	 */
	public void addCodeStoredListener(CodeStoredListener listener) {
		this.codeStoreListeners.add(listener);
	}
}
