/*************************************************************************
 *                           Smart Active Node                           * 
 *                 copyright (c) since 2007 KIV-FAV-ZCU                  *
 *************************************************************************
 * The license for this code is included in LICENSE.txt.  If you did not *
 * get this file with this distribution, request a copy from:            *
 * txkoutny@kiv.zcu.cz                                                   *
 *************************************************************************/
package san.core;

import java.io.InputStream;
import java.io.PrintWriter;
import java.security.InvalidKeyException;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.util.Collection;

import javax.crypto.BadPaddingException;
import javax.crypto.IllegalBlockSizeException;
import san.RendezVous.IRendezvousServer;
import san.core.fiber.api.IConditionAPI;
import san.core.fiber.api.IInfoGetterAPI;
import san.core.fiber.api.IMonitorAPI;
import san.core.fiber.api.ISleepAPI;
import san.security.Crypto.KeyGenerator;
import san.data.ReceiveDataListener;
import san.data.TransmitData;
import san.ident.CodePkgIdent;
import san.ident.Guid;
import san.ident.NetIdentifier;
import san.ii.Packet;
import san.softstate.ISoftState;

/**
 * <b>Title:</b> Smart Active Node<br> <b>Description:</b> The API witch is used by application process.<br />
 * @author     Michal Rejda
 * @since     21.10.2007
 * @version     1.0
 * @uml.dependency   supplier="san.core.IServerControl"
 * @uml.dependency   supplier="san.softstate.ISoftState"
 */
public interface IApplicationAPI extends ICommonApi, IConditionAPI, IMonitorAPI, ISleepAPI, IInfoGetterAPI {

	/**
	 * Create new capsule. This capsule is afterwards send to the destination.
	 * @throws DestinationNotAvailableException if destination is not jet available
	 */
	public void injectCapsule(CodePkgIdent executedCode,
			NetIdentifier destination, TransmitData payload) throws DestinationNotAvailableException;
	
	public void injectCapsule(CodePkgIdent executedCode,
			NetIdentifier destination, TransmitData payload, boolean executeContinuously) throws DestinationNotAvailableException;

	/**
	 * Return server control API.
	 */
	public IServerControl getServerControl();

		/**
	 * All IOnSendDataListener to listen onSendData event.
	 */
	public void addOnSendDataListener(Guid listener);

	/**
	 * Remove listener. Return true if listener was existed.
	 */
	public boolean removeOnSendDataListener(ReceiveDataListener listener);

	/**
	 * Run application with specific code identification.
	 */
	public Guid runApplication(CodePkgIdent ident, String[] args)
			throws ExecuteException;

	/**
	 * Run application with specific name.
	 */
	public Guid runApplication(String name, String[] args)
			throws ExecuteException;

	/**
	 * Wait for end of application.
	 * @param applId application id
	 */
	public void waitForApplication(Guid applId) throws InterruptedException;

	/**
	 * Wait for end of collection of application.
	 * @param applId application id
	 */
	public void waitForApplication(Collection<Guid> applId)
			throws InterruptedException;

	/**
	 * Return standard input stream.
	 */
	public InputStream getStdIn();

	/**
	 * Return standard output writer.
	 */
	public PrintWriter getStdOut();

	/**
	 * Return application id.
	 */
	public Guid getId();
	
	/**
	 * Return identification of executed code.
	 */
	public CodePkgIdent getExecutedCodeIdent();
	
	/**
	 * Return soft state.
	 */
	public ISoftState getSoftState();
	
	
	public String getNodeInfo();

        /**
         * Log the user to the node
         * @param username user username
         * @param password user password
         * @return if username and password are valid, return true, otherwise false
         */
        public boolean logUser(String username, String password);

        /**
         * Method set owner of parent process. New owner will be same as owner current process.
         */
        public void supressOwnerToParent();

        /**
         * WhoAmI returns name of current user
         * @return WhoAmI returns name of current user
         */
        public String whoAmI();

        /**
         * getParentOwner returns name of parent process user
         * @return getParentOwner returns name of parent process user
         */
        public String getParentOwner();

        /**
         * Method create key generators, which holds private and public keys
         * @return instance of KeyGenerator
         */
        public KeyGenerator createKeyGenerator();

        /**
         * Method encrypt message by private key
         * @param message message, which will be encrypted
         * @param privateKey key, which will be used for encrypting
         * @return String, which represent encrypted string
         */
        public String getEncryptedString(String message, PublicKey publicKey) throws InvalidKeyException, IllegalBlockSizeException, BadPaddingException;

         /**
         * Method decrypt message by public key
         * @param encryptedMessage message, which will be decrypted
         * @param publicKey key, which will be used for decrypting
         * @return String, which represent decrypted string (original string)
         */
        public String getDecryptedString(String encryptedMessage, PrivateKey privateKey) throws InvalidKeyException, IllegalBlockSizeException, BadPaddingException;


  /**
   * This method makes sending from appliaction data to server possible
   */
  public void sendPayload(Packet p);

  /**
   * This method register rendez-vous call with random Guid
   * @param entryClass instance of object with entry method
   * @return registered Guid
   */
  public Guid registerServer(IRendezvousServer entryClass);

  /**
   * This method register rendez-vous call with the given Guid
   * @param entryClass instance of object with entry method
   * @param server Server guid for register
   * @return True if registering is succesfull
   */
  public boolean registerServer(IRendezvousServer entryClass, Guid server);

  /**
   * Register entry call name
   * @param server Registered server Guid
   * @param entryName entry call name for registered
   * @return entry call ID if succesfull or null
   */
  public Integer registerEntryCall(Guid server, String entryName);

  /**
   * Accept entry call on registered server
   * @param server Server Guid
   * @throws java.lang.InterruptedException
   */
  public void makeAccept(Guid server) throws InterruptedException;

  /**
   * Accept group entry call on registered server
   * @param server Server Guid
   * @throws java.lang.InterruptedException
   */
  public void makeAccept(Guid server, String[] entryNames) throws InterruptedException;

  /**
   * Unregister server and terminate and unregister all registered entry call name
   * @param server Server Guid
   */
  public void unregisterServer(Guid server);
}
