package san.core.fiber.condition;



import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import org.apache.log4j.Logger;
import san.core.Process;
import san.core.fiber.log.FiberLoggerFactory;
import san.core.fiber.scheduler.ISchedulerModule;



/**
 *
 * @author Jiri Nohavec
 */
public class ConditionManager implements ISchedulerModule {



    private static final Logger logger = FiberLoggerFactory.getSchedulerLogger();

    private final Map<Integer, Queue<Process>> conditionMap;
    private final List<Process> timeoutProcessList;



    public ConditionManager() {
        conditionMap = new HashMap<Integer, Queue<Process>>();
        timeoutProcessList = new LinkedList<Process>();
    }



    @Override
    public synchronized void checkForStoppedProcesses() throws ConditionException {
        Queue<Process> processQueue;
        for (Integer condition: conditionMap.keySet()) {
            processQueue = conditionMap.get(condition);
            for (Iterator<Process> iterator = processQueue.iterator(); iterator.hasNext(); ) {
                Process process = iterator.next();
                switch (process.getState()) {
                    case FINISHED:
                    case STOPPED:
                        iterator.remove();
                        break;
                    default:
                        if (logger.isDebugEnabled()) {
                            logger.debug(String.format("    CON - waiting on condition %d, %s",
                                                       condition,
                                                       process.getDebugInfo()));
                        }
                        break;
                }
            }
        }
        long currentTime = System.currentTimeMillis();
        for (Iterator<Process> iterator = timeoutProcessList.iterator(); iterator.hasNext(); ) {
            Process process = iterator.next();
            if (process.getTimeToWakeUp() < currentTime) {
                iterator.remove();
                processQueue = conditionMap.get(process.getConditionRequest().getCondition());
                processQueue.remove(process);
                process.getInterpretingContext().replaceReturnValue(false);
                process.setState(Process.State.READY_TO_RUN);
                if (logger.isDebugEnabled()) {
                    logger.debug(String.format("    CON - timeout on condition %d, %s",
                                               process.getConditionRequest().getCondition(),
                                               process.getDebugInfo()));
                }
            }
        }
    }



    private void actionSignalOne(Queue<Process> processQueue) {
        if (!processQueue.isEmpty()) {
            Process process = processQueue.remove();
            if (timeoutProcessList.remove(process)) {
                process.getInterpretingContext().replaceReturnValue(true);
            }
            process.setState(Process.State.READY_TO_RUN);
        }
    }



    private void actionSignalAll(Queue<Process> processQueue) {
        while (!processQueue.isEmpty()) {
            actionSignalOne(processQueue);
        }
    }



    @Override
    public synchronized void checkProcessCall(Process process) throws ConditionException {
        ConditionRequest conditionRequest = process.getConditionRequest();
        if (!conditionRequest.isNewCall()) {
            return;
        }
        int condition = conditionRequest.getCondition();
        ConditionRequest.Type type = conditionRequest.getType();
        Queue<Process> processQueue;
        boolean conditionInitialized;
        boolean unblockProcess = true;
        if (logger.isDebugEnabled()) {
            logger.debug(String.format("    CON - call %s, id %d, %s",
                                       type.toString(),
                                       condition,
                                       process.getDebugInfo()));
        }
        processQueue = conditionMap.get(condition);
        conditionInitialized = (processQueue != null);
        switch (type) {
            case WAIT:
            case WAIT_TIMEOUT:
            case SIGNAL_ONE:
            case SIGNAL_ALL:
            case GET_WAITING_COUNT:
                if (!conditionInitialized) {
                    throw new ConditionException(String.format("Condition %d is not initialized", condition));
                }
                break;
        }
        switch (type) {
            case INITIALIZE:
                if (!conditionInitialized) {
                    conditionMap.put(condition, new LinkedList<Process>());
                }
                process.getInterpretingContext().replaceReturnValue(!conditionInitialized);
                break;
            case DESTROY:
                if (conditionInitialized) {
                    conditionMap.remove(condition);
                    actionSignalAll(processQueue);
                }
                process.getInterpretingContext().replaceReturnValue(conditionInitialized);
                break;
            case WAIT:
                processQueue.add(process);
                unblockProcess = false;
                break;
            case WAIT_TIMEOUT:
                processQueue.add(process);
                timeoutProcessList.add(process);
                unblockProcess = false;
                break;
            case SIGNAL_ONE:
                actionSignalOne(processQueue);
                break;
            case SIGNAL_ALL:
                actionSignalAll(processQueue);
                break;
            case GET_WAITING_COUNT:
                process.getInterpretingContext().replaceReturnValue(processQueue.size());
                break;
        }
        conditionRequest.newCallReceived();
        if (unblockProcess) {
            process.setState(Process.State.RUNNING);
        }
    }
}
