package san.core.fiber.info;



import java.text.SimpleDateFormat;
import java.util.Date;
import san.core.Process;
import san.core.fiber.api.IInfoAPI;
import san.core.fiber.settings.FiberSettings;



/**
 *
 * @author Jiri Nohavec
 */
public class FiberInfo implements IInfoAPI {



    private volatile int applicationsStartedCount;
    private volatile int applicationsFinishedCount;
    private volatile int applicationsFailedCount;
    private volatile int capsulesStartedCount;
    private volatile int capsulesFinishedCount;
    private volatile int capsulesFailedCount;
    private volatile int specialCallStartedCount;
    private volatile int specialCallFinishedCount;
    private volatile int specialCallFailedCount;

    private final String startupTime;
    private final long startupTimeMilis;
    private final int physicalCPUCount;
    private final int virtualCPUCount;
    private final int maxCapsulesRunning;
    private final long schedulerSleepInterval;
    private final String schedulerStrategy;
    private final boolean deadlockDetectorEnabled;
    private final long deadlockDetectorSleepInterval;



    public FiberInfo() {
        FiberSettings fiberSettings = FiberSettings.getInstance();
        applicationsStartedCount = 0;
        applicationsFinishedCount = 0;
        applicationsFailedCount = 0;
        capsulesStartedCount = 0;
        capsulesFinishedCount = 0;
        capsulesFailedCount = 0;
        specialCallStartedCount = 0;
        specialCallFinishedCount = 0;
        specialCallFailedCount = 0;
        startupTime = new SimpleDateFormat("dd.MM.yyyy HH:mm:ss").format(new Date());
        startupTimeMilis = System.currentTimeMillis();
        physicalCPUCount = Runtime.getRuntime().availableProcessors();
        virtualCPUCount = fiberSettings.getVirtualCPUCount();
        maxCapsulesRunning = fiberSettings.getMaxCapsulesRunning();
        schedulerSleepInterval = fiberSettings.getSchedulerSleepInterval();
        schedulerStrategy = fiberSettings.getSchedulerStrategy();
        deadlockDetectorEnabled = fiberSettings.isDeadlockDetectorEnabled();
        deadlockDetectorSleepInterval = fiberSettings.getDeadlockDetectorSleepInterval();
    }



    public void incrementStartedCount(Process process) {
        switch (process.getType()) {
            case APPLICATION:
                applicationsStartedCount++;
                break;
            case CAPSULE:
                capsulesStartedCount++;
                break;
            case SPECIAL_CALL:
                specialCallStartedCount++;
                break;
        }
    }



    public void incrementFinishedCount(Process process) {
        switch (process.getType()) {
            case APPLICATION:
                applicationsFinishedCount++;
                break;
            case CAPSULE:
                capsulesFinishedCount++;
                break;
            case SPECIAL_CALL:
                specialCallFinishedCount++;
                break;
        }
    }



    public void incrementFailedCount(Process process) {
        switch (process.getType()) {
            case APPLICATION:
                applicationsFailedCount++;
                break;
            case CAPSULE:
                capsulesFailedCount++;
                break;
            case SPECIAL_CALL:
                specialCallFailedCount++;
                break;
        }
    }



    @Override
    public int getApplicationsStartedCount() {
        return applicationsStartedCount;
    }



    @Override
    public int getApplicationsFinishedCount() {
        return applicationsFinishedCount;
    }



    @Override
    public int getApplicationsFailedCount() {
        return applicationsFailedCount;
    }



    @Override
    public int getCapsulesStartedCount() {
        return capsulesStartedCount;
    }



    @Override
    public int getCapsulesFinishedCount() {
        return capsulesFinishedCount;
    }



    @Override
    public int getCapsulesFailedCount() {
        return capsulesFailedCount;
    }



    @Override
    public int getSpecialCallStartedCount() {
        return specialCallStartedCount;
    }



    @Override
    public int getSpecialCallFinishedCount() {
        return specialCallFinishedCount;
    }



    @Override
    public int getSpecialCallFailedCount() {
        return specialCallFailedCount;
    }



    @Override
    public String getStartupTime() {
        return startupTime;
    }



    @Override
    @SuppressWarnings("static-access")
    public String getRunTime() {
        long runTime, days, hours, minutes, seconds;
        runTime = System.currentTimeMillis() - startupTimeMilis;
        runTime /= 1000L;
        seconds = runTime % 60L;
        runTime /= 60L;
        minutes = runTime % 60L;
        runTime /= 60L;
        hours = runTime % 24L;
        runTime /= 24L;
        days = runTime;
        if (days == 0L && hours == 0 && minutes == 0) {
            return "".format("%02d seconds", seconds);
        } else if (days == 0L && hours == 0) {
            return "".format("%02d minutes, %02d seconds", minutes, seconds);
        } else if (days == 0L) {
            return "".format("%02d hours, %02d minutes, %02d seconds", hours, minutes, seconds);
        } else {
            return "".format("%d days, %02d hours, %02d minutes, %02d seconds", days, hours, minutes, seconds);
        }
    }



    @Override
    public int getPhysicalCPUCount() {
        return physicalCPUCount;
    }



    @Override
    public int getVirtualCPUCount() {
        return virtualCPUCount;
    }



    @Override
    public int getMaxCapsulesRunning() {
        return maxCapsulesRunning;
    }



    @Override
    public long getSchedulerSleepInterval() {
        return schedulerSleepInterval;
    }



    @Override
    public String getSchedulerStrategy() {
        return schedulerStrategy;
    }



    @Override
    public boolean isDeadlockDetectorEnabled() {
        return deadlockDetectorEnabled;
    }



    @Override
    public long getDeadlockDetectorSleepInterval() {
        return deadlockDetectorSleepInterval;
    }
}
