package san.core.fiber.scheduler;



import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Queue;
import san.core.Process;
import san.core.fiber.info.FiberInfo;
import san.ident.CodePkgIdent;
import san.ident.Guid;



/**
 *
 * @author Jiri Nohavec
 */
public class FiberSchedulerMailbox {



    private final FiberInfo fiberInfo;
    private final Object schedulerSleepCondition;
    private final Object processWaitCondition;
    private final Queue<Process> newProcessQueue;
    private final Queue<Process> intepretingEndedProcessQueue;
    private final Map<Guid, Process> processMap;

    private volatile int capsulesCurrentlyRunning;



    public FiberSchedulerMailbox() {
        fiberInfo = new FiberInfo();
        schedulerSleepCondition = new Object();
        processWaitCondition = new Object();
        newProcessQueue = new LinkedList<Process>();
        intepretingEndedProcessQueue = new LinkedList<Process>();
        processMap = new HashMap<Guid, Process>();
        capsulesCurrentlyRunning = 0;
    }



    public void waitOnSchedulerSleepCondition(long timeout) {
        synchronized (schedulerSleepCondition) {
            try {
                schedulerSleepCondition.wait(timeout);
            } catch (InterruptedException e) {
            }
        }
    }



    public void waitOnProcessWaitCondition() {
        synchronized (processWaitCondition) {
            try {
                processWaitCondition.wait();
            } catch (InterruptedException e) {
            }
        }
    }



    public void notifyAllConditions() {
        synchronized (schedulerSleepCondition) {
            schedulerSleepCondition.notify();
        }
        synchronized (processWaitCondition) {
            processWaitCondition.notifyAll();
        }
    }



    public void sendNewProcess(Process process) {
        synchronized (newProcessQueue) {
            if (!newProcessQueue.contains(process)) {
                newProcessQueue.add(process);
            }
        }
        switch (process.getType()) {
            case APPLICATION:
                synchronized (processMap) {
                    processMap.put(process.getId(), process);
                }
                break;
            case CAPSULE:
                capsulesCurrentlyRunning++;
                break;
        }
        fiberInfo.incrementStartedCount(process);
        synchronized (schedulerSleepCondition) {
            schedulerSleepCondition.notify();
        }
    }



    public void sendFinishedProcess(Process process) {
        switch (process.getType()) {
            case APPLICATION:
                synchronized (processMap) {
                    processMap.remove(process.getId());
                }
                break;
            case CAPSULE:
                capsulesCurrentlyRunning--;
                break;
        }
        switch (process.getState()) {
            case FINISHED:
                fiberInfo.incrementFinishedCount(process);
                break;
            case STOPPED:
                fiberInfo.incrementFailedCount(process);
                break;
        }
        synchronized (processWaitCondition) {
            processWaitCondition.notifyAll();
        }
    }



    public void sendIntepretingEndedProcess(Process process) {
        synchronized (intepretingEndedProcessQueue) {
            if (!intepretingEndedProcessQueue.contains(process)) {
                intepretingEndedProcessQueue.add(process);
            }
        }
        synchronized (schedulerSleepCondition) {
            schedulerSleepCondition.notify();
        }
    }



    public boolean hasNewProcess() {
        synchronized (newProcessQueue) {
            return (!newProcessQueue.isEmpty());
        }
    }



    public Process removeNewProcess() {
        synchronized (newProcessQueue) {
            return newProcessQueue.remove();
        }
    }



    public boolean hasIntepretingEndedProcess() {
        synchronized (intepretingEndedProcessQueue) {
            return (!intepretingEndedProcessQueue.isEmpty());
        }
    }



    public Process removeIntepretingEndedProcess() {
        synchronized (intepretingEndedProcessQueue) {
            return intepretingEndedProcessQueue.remove();
        }
    }



    public Process getProcess(Guid guid) {
        synchronized (processMap) {
            return processMap.get(guid);
        }
    }



    public Collection<Process> getProcess(CodePkgIdent codePkgIdent) {
        Queue<Process> matchingProcessQueue = new LinkedList<Process>();
        synchronized (processMap) {
            for (Process process: processMap.values()) {
                if (process.getCodePackage().getPkgIdentification().equals(codePkgIdent)) {
                    matchingProcessQueue.add(process);
                }
            }
        }
        return matchingProcessQueue;
    }



    public int getProcessCount(CodePkgIdent codePkgIdent) {
        int processCount = 0;
        synchronized (processMap) {
            for (Process process: processMap.values()) {
                if (process.getCodePackage().getPkgIdentification().equals(codePkgIdent)) {
                    processCount++;
                }
            }
        }
        return processCount;
    }



    public int getCapsulesCurrentlyRunning() {
        return capsulesCurrentlyRunning;
    }



    public FiberInfo getFiberInfo() {
        return fiberInfo;
    }
}
