package san.core.fiber.strategy;



import java.util.Iterator;
import java.util.LinkedList;
import java.util.Queue;
import org.apache.log4j.Logger;
import san.core.Process;
import san.core.fiber.log.FiberLoggerFactory;



/**
 *
 * @author Jiri Nohavec
 */
public abstract class ISchedulerStrategy {



    private static final Logger logger = FiberLoggerFactory.getSchedulerLogger();

    protected final int cpuCount;
    protected final Queue<Process> allProcessQueue;
    protected final Queue<Process> selectedProcessQueue;

    private final Process[] previousProcessArray;
    private final Process[] optimalizedProcessArray;
    private final Process[] emptyProcessArray;



    public ISchedulerStrategy(int cpuCount, Queue<Process> allProcessQueue) {
        this.cpuCount = cpuCount;
        this.allProcessQueue = allProcessQueue;
        selectedProcessQueue = new LinkedList<Process>();
        previousProcessArray = new Process[cpuCount];
        optimalizedProcessArray = new Process[cpuCount];
        emptyProcessArray = new Process[cpuCount];
    }



    protected abstract void reschedule();



    private void optimalize() {
        for (Iterator<Process> iterator = selectedProcessQueue.iterator(); iterator.hasNext(); ) {
            Process process = iterator.next();
            for (int i = 0; i < cpuCount; i++) {
                if (previousProcessArray[i] == process) {
                    optimalizedProcessArray[i] = process;
                    iterator.remove();
                    break;
                }
            }
        }
        for (Iterator<Process> iterator = selectedProcessQueue.iterator(); iterator.hasNext(); ) {
            Process process = iterator.next();
            if (optimalizedProcessArray[process.getCpuIndex()] == null) {
                optimalizedProcessArray[process.getCpuIndex()] = process;
                iterator.remove();
            }
        }
        int optimalizedIndex = 0;
        while (!selectedProcessQueue.isEmpty()) {
            while (optimalizedProcessArray[optimalizedIndex] != null) {
                optimalizedIndex++;
            }
            Process process = selectedProcessQueue.remove();
            process.setCpuIndex(optimalizedIndex);
            optimalizedProcessArray[optimalizedIndex] = process;
        }
    }



    public final void rescheduleCall() throws StrategyException {
        selectedProcessQueue.clear();
        reschedule();
        if (selectedProcessQueue.size() > cpuCount) {
            throw new StrategyException(String.format("Strategy selected more processes than we have CPUs"));
        }
        for (Process process: selectedProcessQueue) {
            if (process == null) {
                throw new StrategyException(String.format("Strategy selected null process"));
            }
        }
        System.arraycopy(emptyProcessArray, 0, optimalizedProcessArray, 0, cpuCount);
        optimalize();
        System.arraycopy(optimalizedProcessArray, 0, previousProcessArray, 0, cpuCount);
        if (logger.isDebugEnabled()) {
            logger.debug(String.format("SCH - schedule plan:"));
            for (int i = 0; i < cpuCount; i++) {
                if (optimalizedProcessArray[i] != null) {
                    logger.debug(String.format("        * CPU %d, %s",
                                               i,
                                               optimalizedProcessArray[i].getDebugInfo()));
                } else {
                    logger.debug(String.format("        * CPU %d, idle", i));
                }
            }
        }
    }



    public Process getSelectedProcess(int cpuIndex) throws StrategyException {
        Process process = optimalizedProcessArray[cpuIndex];
        if (process == null) {
            return null;
        }
        if (cpuIndex != process.getCpuIndex()) {
            throw new StrategyException(String.format("Strategy selected wrong CPU index"));
        }
        return process;
    }
}
