package san.interpreter.sanint.fiber;



import org.apache.log4j.Logger;
import san.applicationRepository.SanCodePkg;
import san.core.IApplicationAPI;
import san.core.ICapsuleAPI;
import san.core.Kernel;
import san.core.Process;
import san.core.ProcessApplication;
import san.core.ProcessCapsule;
import san.core.fiber.log.FiberLoggerFactory;
import san.core.fiber.log.FiberUncaughtExceptionHandler;
import san.data.TransmitCapsule;
import san.interpreter.IInterpreter;
import san.interpreter.InterpretManager;
import san.interpreter.sanint.classLoading.ClassManager;
import san.interpreter.sanint.classLoading.Reference;
import san.interpreter.sanint.fiber.call.ISpecialCall;
import san.interpreter.sanint.fiber.call.ReceiveDataCall;
import san.interpreter.sanint.fiber.call.RendezVousCall;



/**
 *
 * @author Jiri Nohavec
 */
public class SanFiberInterpreter implements IInterpreter, Runnable {



    private static final Logger logger = FiberLoggerFactory.getInterpreterLogger();



    public static SanFiberInterpreter[] startInterpreters(InterpretManager interpretManager,
                                                          Kernel kernel,
                                                          int cpuCount) {
        SanFiberInterpreter[] sanFiberInterpreterArray = new SanFiberInterpreter[cpuCount];
        for (int cpuIndex = 0; cpuIndex < cpuCount; cpuIndex++) {
            sanFiberInterpreterArray[cpuIndex] = new SanFiberInterpreter(interpretManager, kernel, cpuIndex);
            Thread thread = new Thread(sanFiberInterpreterArray[cpuIndex]);
            thread.setName(String.format("San fiber interpreter thread %d", cpuIndex));
            thread.setUncaughtExceptionHandler(new FiberUncaughtExceptionHandler(logger));
            thread.start();
            if (logger.isDebugEnabled()) {
                logger.debug(String.format("    INT %d - interpreter thread started", cpuIndex));
            }
        }
        return sanFiberInterpreterArray;
    }



    private final InterpretManager interpretManager;
    private final Kernel kernel;
    private final int cpuIndex;
    private final Object interpreterSleepCondition;

    private volatile Process scheduledProcess;
    private volatile boolean running;

    private boolean releaseProcess;



    public SanFiberInterpreter(InterpretManager interpretManager, Kernel kernel, int cpuIndex) {
        this.interpretManager = interpretManager;
        this.kernel = kernel;
        this.cpuIndex = cpuIndex;
        interpreterSleepCondition = new Object();
        scheduledProcess = null;
        running = true;
        releaseProcess = false;
    }



    public void stopThread() {
        running = false;
        synchronized (interpreterSleepCondition) {
            interpreterSleepCondition.notify();
        }
    }



    public int getCpuIndex() {
        return cpuIndex;
    }



    @Override
    public void runProcess(Process process) {
        synchronized (interpreterSleepCondition) {
            scheduledProcess = process;
            interpreterSleepCondition.notify();
        }
    }



    @Override
    public Process terminate() {
        Process process = scheduledProcess;
        synchronized (interpreterSleepCondition) {
            scheduledProcess = null;
            interpreterSleepCondition.notify();
        }
        return process;
    }



    @Override
    public Process getProcess() {
        synchronized (interpreterSleepCondition) {
            interpreterSleepCondition.notify();
        }
        return scheduledProcess;
    }



    @Override
    public boolean isInterpreting() {
        synchronized (interpreterSleepCondition) {
            interpreterSleepCondition.notify();
        }
        return (scheduledProcess != null);
    }



    private void haltInterpreter(Process process) {
        if (process != null) {
            interpretManager.interpretingEnded(process);
            if (process.getState() != Process.State.RUNNING) {
                releaseProcess = true;
            }
        }
        synchronized (interpreterSleepCondition) {
            try {
                interpreterSleepCondition.wait();
            } catch (InterruptedException e) {
            }
        }
    }



    @Override
    public void run() {
        FiberInterpreterBytecode fiberInterpreterBytecode = new FiberInterpreterBytecode(cpuIndex);
        Process nextProcess;
        Process process = null;
        while (running) {
            if (releaseProcess) {
                releaseProcess = false;
                if (process != null) {
                    process.setNowInterpreted(false);
                    if (logger.isDebugEnabled()) {
                        logger.debug(String.format("    INT %d - leave,    %s",
                                                   cpuIndex,
                                                   process.getDebugInfo()));
                    }
                    process = null;
                }
            }
            synchronized (interpreterSleepCondition) {
                nextProcess = scheduledProcess;
            }
            if (nextProcess != process) {
                if (process != null) {
                    if (process.getState() == Process.State.RUNNING) {
                        process.setState(Process.State.READY_TO_RUN);
                    }
                    process.setNowInterpreted(false);
                    if (logger.isDebugEnabled()) {
                        logger.debug(String.format("    INT %d - leave,    %s",
                                                   cpuIndex,
                                                   process.getDebugInfo()));
                    }
                }
                process = nextProcess;
                if (process == null) {
                    haltInterpreter(process);
                    continue;
                }
                while (process.isNowInterpreted()) {
                    ;
                }
                if (process.getState() != Process.State.READY_TO_RUN) {
                    haltInterpreter(process);
                    continue;
                }
                process.setState(Process.State.RUNNING);
                process.setNowInterpreted(true);
                if (process.getInterpretingContext() != null) {
                    if (logger.isDebugEnabled()) {
                        logger.debug(String.format("    INT %d - continue, %s",
                                                   cpuIndex,
                                                   process.getDebugInfo()));
                    }
                }
            }
            if (process == null || process.getState() != Process.State.RUNNING) {
                haltInterpreter(process);
                continue;
            }
            try {
                if (process.getInterpretingContext() == null) {
                    switch (process.getType()) {
                        case APPLICATION:
                        case CAPSULE:
                            startProcess(process);
                            break;
                        case SPECIAL_CALL:
                            startSpecialProcess(process);
                            break;
                    }
                    if (logger.isDebugEnabled()) {
                        logger.debug(String.format("    INT %d - start,    %s",
                                                   cpuIndex,
                                                   process.getDebugInfo()));
                    }
                } else {
                    for (int i = 0; i < 10; i++) {
                        boolean moreToInterpret = fiberInterpreterBytecode.oneStep(process);
                        if (!moreToInterpret) {
                            process.setState(Process.State.FINISHED);
                            endProcess(process);
                        }
                        if (process.getState() != Process.State.RUNNING) {
                            interpretManager.interpretingEnded(process);
                            break;
                        }
                    }
                    if (logger.isDebugEnabled()) {
                        switch (process.getState()) {
                            case FINISHED:
                                logger.debug(String.format("    INT %d - finished, %s",
                                                           cpuIndex,
                                                           process.getDebugInfo()));
                                break;
                            case STOPPED:
                                logger.debug(String.format("    INT %d - stopped,  %s",
                                                           cpuIndex,
                                                           process.getDebugInfo()));
                                break;
                            case BLOCKED:
                                logger.debug(String.format("    INT %d - blocked,  %s",
                                                           cpuIndex,
                                                           process.getDebugInfo()));
                                break;
                            case ASLEEP:
                                logger.debug(String.format("    INT %d - asleep,   %s",
                                                           cpuIndex,
                                                           process.getDebugInfo()));
                                break;
                        }
                    }
                }
            } catch (Exception e) {
                if (process != null) {
                    process.setState(Process.State.STOPPED);
                    endProcess(process);
                    interpretManager.interpretingEnded(process);
                    if (logger.isDebugEnabled()) {
                        logger.debug(String.format("    INT %d - stopped,  %s",
                                                   cpuIndex,
                                                   process.getDebugInfo()));
                    }
                    logger.warn(String.format("Exception occurred while interpreting in INT %d%nProcess: %s",
                                              cpuIndex,
                                              process.getDebugInfo()), e);
                }
            }
        }
    }



    private void startProcess(Process process) throws Exception {
        SanCodePkg sanCodePackage = (SanCodePkg) process.getCodePackage();
        ClassManager classManager = new ClassManager(sanCodePackage, process);
        String mainClass;
        Reference reference;
        String[] attributes;
        process.setInterpretingContext(new FiberInterpretingContext(classManager));
        classManager.Init();
        if (process instanceof IApplicationAPI) {
            mainClass = sanCodePackage.getManifestPropertyValue(SanCodePkg.APP_MAIN_CLASS_MFP);
            reference = classManager.createReference(mainClass);
            attributes = ((ProcessApplication) process).getArgs();
            classManager.runMethod(mainClass,
                                   process.getMainMethodName(),
                                   process.getMainMethodDescriptor(),
                                   new Object[] {
                                       attributes,
                                       process,
                                       process.getId()
                                   },
                                   reference,
                                   false);
            interpretManager.addRunningApplication(process.getId(), reference);
        } else if (process instanceof ICapsuleAPI) {
            mainClass = sanCodePackage.getManifestPropertyValue(SanCodePkg.CAPSULE_MAIN_CLASS_MFP);
            reference = classManager.createReference(mainClass);
            classManager.runMethod(mainClass,
                                   process.getMainMethodName(),
                                   process.getMainMethodDescriptor(),
                                   new Object[] {
                                       process
                                   },
                                   reference,
                                   false);
        }
    }



    private void startSpecialProcess(Process process) throws Exception {
        ISpecialCall specialCall = process.getSpecialCall();
        if (specialCall.getType() == ISpecialCall.Type.RECEIVE_DATA) {
            ReceiveDataCall receiveDataCall = (ReceiveDataCall) specialCall;
            SanCodePkg sanCodePackage = (SanCodePkg) process.getCodePackage();
            ClassManager classManager = new ClassManager(sanCodePackage, process);
            Reference reference = interpretManager.getRunningApplication(receiveDataCall.getReceiverId());
            if (reference == null) {
                throw new InterpretingException("External reference for special call is null");
            }
            process.setInterpretingContext(new FiberInterpretingContext(classManager));
            classManager.Init();
            classManager.runMethod(reference.getClassInfo().getName(),
                                   process.getMainMethodName(),
                                   process.getMainMethodDescriptor(),
                                   new Object[] {
                                       receiveDataCall.getSenderId(),
                                       receiveDataCall.getTransmitData(),
                                       process
                                   },
                                   reference,
                                   false);
        } else if (specialCall.getType() == ISpecialCall.Type.RENDEZ_VOUS) {
            RendezVousCall randevousCall = (RendezVousCall) specialCall;
            SanCodePkg sanCodePackage = (SanCodePkg) process.getCodePackage();
            ClassManager classManager = new ClassManager(sanCodePackage, process);
            Reference reference = randevousCall.getReference();
            if (reference == null) {
                throw new InterpretingException("External reference for special call is null");
            }
            process.setInterpretingContext(new FiberInterpretingContext(classManager));
            classManager.Init();
            classManager.runMethod(reference.getClassInfo().getName(),
                                   process.getMainMethodName(),
                                   process.getMainMethodDescriptor(),
                                   new Object[] {
                                       randevousCall.getEntryName(),
                                       randevousCall.getEntryParameter()
                                   },
                                   reference,
                                   false);
        } else {
            throw new InterpretingException("Unsupported special call");
        }
    }



    private void endProcess(Process process) {
        switch (process.getType()) {
            case APPLICATION:
                interpretManager.removeRunningApplication(process.getId());
                break;
            case CAPSULE:
                ProcessCapsule processCapsule = (ProcessCapsule) process;
                if (processCapsule.isSendCapsuleOnExit()) {
                    TransmitCapsule transmitCapsule = processCapsule.getCapsule();
                    transmitCapsule.setOver(null);
                    transmitCapsule.setPrev(null);
                    kernel.sendCapsule(transmitCapsule);
                }
                break;
        }
    }
}
