/*************************************************************************
 *                           Smart Active Node                           * 
 *                 copyright (c) since 2007 KIV-FAV-ZCU                  *
 *************************************************************************
 * The license for this code is included in LICENSE.txt.  If you did not *
 * get this file with this distribution, request a copy from:            *
 * txkoutny@kiv.zcu.cz                                                   *
 *************************************************************************/
package san.networkInterfaces;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.util.Collection;
import java.util.Hashtable;
import java.util.Map;
import san.ident.NetIdentifier;
import san.localization.Localization;
import san.logging.LoggerFactory;
import san.logging.ILogger.LogType;
import san.networkInterfaces.IncomingPacket.Head;

/**
 * <b>Title:</b> Smart Active Node<br>
 * <b>Description:</b> This class solve connection to another node on the
 * specific interface.<br>
 * @author Michal Rejda
 * @since 16.2.2007
 * @version 2.0
 */
public class ConnectionsSolverManager {

	public static final long CONNECTION_TIME_OUT = 10000;
	public static final int NEIGHBOURS_TIME_OUT = 1000;
	public static final long CHECK_TIME = 1000;
	/**
	 * @uml.property name="netInterface"
	 * @uml.associationEnd multiplicity="(1 1)"
	 *                     inverse="connectionsSolverManager:san.networkInterfaces.NetInterface"
	 */
	private NetInterface netInterface = null;
	/**
	 * @uml.property name="connectionSolvers"
	 * @uml.associationEnd multiplicity="(1 1)" aggregation="composite"
	 *                     inverse="manager:san.networkInterfaces.ConnectionSolver"
	 *                     qualifier="key:san.ident.NetIdentifier
	 *                     san.networkInterfaces.ConnectionSolver"
	 */
	private Map<NetIdentifier, ConnectionSolver> connectionSolvers;
	private DatagramSocket dataSocket;
	private Thread nighboursThread;

	public ConnectionsSolverManager(NetInterface netInterface) {
		this.netInterface = netInterface;
		this.connectionSolvers = new Hashtable<NetIdentifier, ConnectionSolver>();
		ConnectionTimeOutTimer ct = new ConnectionTimeOutTimer(
				NEIGHBOURS_TIME_OUT);
		ct.start(); // STRT

	}

	/*
	 * periodicky volana z NetInterface.NighborousTimer
	 */
	void findNighbours() {
		byte[] head = new byte[]{0, 0, 0, 0};
		IncomingPacket.setHeadBit(head, Head.FIND_NEIGHBOURS, true);
		try {
			dataSocket = new DatagramSocket();
			//sendNighbours(DatagramSocket dataSocket, InetAddress destinationAddress,
			//	int destinationPort, int localPort, AccessibleMethod thisMethod, byte[] head)
			netInterface.sendNighbours(dataSocket, netInterface.getBroadcastAddress(), 
					netInterface.getBroadcastPort(), 0,	netInterface.getNextBestMethod(null), head);
			Thread nighboursThread = new Thread() {

				public void run() {
					nighboursThread();
				}
			};
			nighboursThread.start();
		}
		catch (SocketException e) {
			LoggerFactory.createLogger().Log(
					Localization.getLocalization().getString("M000036") + " "
							+ e, LogType.ERROR);
		}
	}

	private void nighboursThread() {
		byte[] data = new byte[IncomingPacket.LENGHT];
		DatagramPacket dp = new DatagramPacket(data, data.length);
		IncomingPacket inPacket;
		try {
			do {
				dataSocket.setSoTimeout(NEIGHBOURS_TIME_OUT);
				dataSocket.receive(dp);
				inPacket = new IncomingPacket(dp);
				findNighboursPacket(inPacket);
			} while (true);
		}
		catch (SocketTimeoutException ex) {
		}
		catch (IOException e) {
			LoggerFactory.createLogger().Log(
					Localization.getLocalization().getString("M000037") + " "
							+ e, LogType.ERROR);
		}
		finally {
			dataSocket.close();
			dataSocket = null;
		}
	}

	/**
	 * Attend to incoming packet witch is trying to find neighbours. 
	 * First one or as a response.
	 */
	void findNighboursPacket(IncomingPacket inPacket) {
		if (inPacket.getIdentifier().equals(netInterface.getIdentifier()) == false) {
			NetIdentifier i = inPacket.getIdentifier();
			ConnectionSolver s = new ConnectionSolver(this, netInterface,
					inPacket);
			connectionSolvers.put(i, s);
			s.start();
		}
	}

	/**
	 * Check out time of the connection solvers.
	 */
	private void checkTimeOut() {
		Collection<ConnectionSolver> solvers = connectionSolvers.values();
		long currentTime = System.currentTimeMillis();
		for (ConnectionSolver cs : solvers) {
			if (currentTime - cs.getLastUsed().getTime() > CONNECTION_TIME_OUT) {
				cs.closeConnection();
				LoggerFactory.createLogger().Log(
						Localization.getLocalization().getString("M000038"),
						LogType.DEBUG);
				connectionSolvers.remove(cs.getAddressTo());
			}
		}
	}

	/**
	 * Remove connection solver from manager.
	 */
	void solverEnded(ConnectionSolver solver) {
		this.connectionSolvers.remove(solver.getAddressTo());
	}

	/**
	 * <b>Title:</b> Smart Active Node<br>
	 * <b>Description:</b> Timer to finding time out by finding connection.<br>
	 * @author Michal Rejda
	 * @since 17.2.2007
	 * @version 1.0
	 */
	private class ConnectionTimeOutTimer {

		private boolean runnning;
		private long time;
		private Thread t;

		public ConnectionTimeOutTimer(long time) {
			this.time = time;
			this.t = new Thread() {

				public void run() {
					try {
						while (runnning) {
							Thread.sleep(CHECK_TIME);
							checkTimeOut();
						}
					}
					catch (InterruptedException e) {
						LoggerFactory.createLogger().Log(
								Localization.getLocalization().getString(
										"M000039")
										+ " " + e, LogType.ERROR);
					}
				}
			};
		}

		/**
		 * Start timer.
		 */
		public void start() {
			this.runnning = true;
			t.start();
		}

		/**
		 * Stop timer.
		 */
		public void stop() {
			this.runnning = false;
		}
	}

	/**
	 * Getter of the property <tt>netInterface</tt>
	 * @return Returns the netInterface.
	 * @uml.property name="netInterface"
	 */
	public NetInterface getNetInterface() {
		return netInterface;
	}

	/**
	 * Setter of the property <tt>netInterface</tt>
	 * @param netInterface The netInterface to set.
	 * @uml.property name="netInterface"
	 */
	public void setNetInterface(NetInterface netInterface) {
		this.netInterface = netInterface;
	}
}
