/*************************************************************************
 *                           Smart Active Node                           * 
 *                 copyright (c) since 2007 KIV-FAV-ZCU                  *
 *************************************************************************
 * The license for this code is included in LICENSE.txt.  If you did not *
 * get this file with this distribution, request a copy from:            *
 * txkoutny@kiv.zcu.cz                                                   *
 *************************************************************************/
package san.networkInterfaces;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.MulticastSocket;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.PriorityQueue;
import san.Utility;
import san.ident.NetIdentifier;
import san.localization.Localization;
import san.logging.LoggerFactory;
import san.logging.ILogger.LogType;
import java.util.Collection;

import org.apache.log4j.Logger;

/**
 * <b>Title:</b> Smart Active Node<br>
 * <b>Description:</b> Describe physical network interface. The interface
 * listen on the broadcast (multicast) socket and hand over incoming packets to
 * network interface manager if is set.<br>
 * @author Jan Boháč, Michal Rejda
 * @since 27.11.2006
 * @version 1.0
 */
public class NetInterface extends Thread implements IConnectionLostListener {

	private static final transient Logger log = Logger.getLogger(NetInterface.class);
	
	public static long TIME_BETWEN_FINDING_NIGHBOURS = 5000;
	/**
	 * Identifier of this interface - address and network
	 */
	private NetIdentifier identifier;
	/**
	 * The associated network interface to this interface.
	 */
	private NetworkInterface netInterface;
	/**
	 * Multicast socket is used as broadcast socked.
	 */
	private MulticastSocket broadcastSocket;
	/**
	 * Identifier for broadcast listening.
	 */
	private InetAddress broadcastAddress;
	/**
	 * Port for broadcast listening.
	 */
	private int broadcastPort;
	/**
	 * accessible methods
	 * @uml.property   name="accMethod"
	 * @uml.associationEnd   multiplicity="(0 -1)" inverse="netInterface:san.networkInterfaces.AccessibleMethod"
	 */
	private PriorityQueue<AccessibleMethod> accMethod;
	/**
	 * @uml.property   name="connectionChangedListeners"
	 * @uml.associationEnd   multiplicity="(0 -1)" aggregation="shared" inverse="netInterface:san.networkInterfaces.IConnectionChangedListener"
	 */
	private Collection<IConnectionChangedListener> connectionChangedListeners;
	/**
	 * @uml.property   name="remoteInterfaces"
	 * @uml.associationEnd   readOnly="true" multiplicity="(1 1)" inverse="netInterfaceManager:san.networkInterfaces.RemoteInterface" qualifier="identifier:san.ident.NetIdentifier san.networkInterfaces.RemoteInterface"
	 */
	private Map<NetIdentifier, RemoteInterface> remoteInterfaces;
	/**
	 * Manager of solve connections.
	 * @uml.property   name="conSolvMgr"
	 * @uml.associationEnd   multiplicity="(1 1)" inverse="netIntMgr:san.networkInterfaces.ConnectionsSolverManager"
	 */
	private ConnectionsSolverManager conSolvMgr = null;
	private boolean running; // indicate if the network interface is

	// listening on the broadcast (multicast) socket
	public NetInterface(NetIdentifier identifier,
			NetworkInterface netInterface, InetAddress broadcastAddress,
			int broadcastPort) {
		this.netInterface = netInterface;
		this.conSolvMgr = new ConnectionsSolverManager(this);
		remoteInterfaces = new Hashtable<NetIdentifier, RemoteInterface>();
		new NighboursTimer(TIME_BETWEN_FINDING_NIGHBOURS).start();
		this.netInterface = netInterface;
		this.identifier = identifier;
		this.broadcastAddress = broadcastAddress;
		this.broadcastPort = broadcastPort;
		running = false;
		connectionChangedListeners = new LinkedList<IConnectionChangedListener>();
		accMethod = new PriorityQueue<AccessibleMethod>();
		try {
			broadcastSocket = new MulticastSocket(broadcastPort);
			broadcastSocket.joinGroup(broadcastAddress);
			broadcastSocket.setBroadcast(true);
		}
		catch (Exception e) {
			LoggerFactory.createLogger().Log(e, LogType.ERROR);
			broadcastSocket = null;
		}
	}

	/**
	 * Add the accessible number of transporting to this interface.
	 */
	public void addAccessibleMethod(AccessibleMethod am) {
		accMethod.add(am);
	}

	public void run() {
		running = true;
		if (log.isDebugEnabled()) {
			log.debug("Interface running:\n" + this.getIdentifier());			
		}
		IncomingPacket inPa;
		while (running) {
			byte[] data = new byte[IncomingPacket.LENGHT];
			DatagramPacket dp = new DatagramPacket(data, data.length);
			try {
				broadcastSocket.receive(dp); // recieving the packet
				inPa = createIncomingPacket(dp);
				// System.out.println("Recieve from: " +
				if (inPa.isFindingNeighbouts()) {
					findNighboursPacket(inPa);
				} else {
					LoggerFactory.createLogger()
							.Log(
									Localization.getLocalization().getString(
											"M000042"), LogType.WARNING);
				}
			}
			catch (IOException e) {
				LoggerFactory.createLogger().Log(e, LogType.ERROR);
			}
		}
		running = false;
		LoggerFactory.createLogger().Log(
				Localization.getLocalization().getString("M000043"),
				LogType.INFORMATION);
	}

	/**
	 * Constructs the incoming packet of the datagram packet.
	 */
	private IncomingPacket createIncomingPacket(DatagramPacket dp) {
		return new IncomingPacket(dp);
	}

	/**
	 * Send datagram packet to finding nighbours as broadcast (multicast).
	 * @param dataSocket socket ovet there is the packed send
	 * @param destinationAddress destination address
	 * @param destinationPort destination port
	 * @param localPort this port to comunication
	 * @param thisMethod method to comunication
	 * @param head head of the packet
	 */
	public void sendNighbours(DatagramSocket dataSocket, InetAddress destinationAddress,
			int destinationPort, int localPort, AccessibleMethod thisMethod, byte[] head) {
		byte[] tmp;
		int i = 0;
		int j = 0;
		byte[] data = new byte[IncomingPacket.LENGHT];
		tmp = head;
		for (j = 0; j < tmp.length; j++) { // head
			data[i + j] = tmp[j];
		}
		i += j;
		tmp = this.identifier.getAddress().toByte();
		for (j = 0; j < tmp.length; j++) { // address
			data[i + j] = tmp[j];
		}
		i += j;
		tmp = this.identifier.getNetwork().toByte();
		for (j = 0; j < tmp.length; j++) { // network
			data[i + j] = tmp[j];
		}
		i += j;
		tmp = Utility.intTo2ByteArray(localPort); // port
		for (j = 0; j < tmp.length; j++) {
			data[i + j] = tmp[j];
		}
		i += j;
		tmp = Utility.intTo2ByteArray(thisMethod.number); // method
		for (j = 0; j < tmp.length; j++) {
			data[i + j] = tmp[j];
		}
		try {
			// packet creating
			DatagramPacket packet = new DatagramPacket(data, data.length, destinationAddress,
					destinationPort);
			// sending of packet
			dataSocket.send(packet);
		}
		catch (SocketException e) {
			LoggerFactory.createLogger().Log(e, LogType.ERROR);
		}
		catch (IOException e) {
			LoggerFactory.createLogger().Log(e, LogType.ERROR);
		}
	}

	/**
	 * Return next best number to send as current.
	 */
	public AccessibleMethod getNextBestMethod(AccessibleMethod previous) {
		if (previous == null)
			return accMethod.peek(); // return the best number
		AccessibleMethod element;
		boolean set = false;
		for (Iterator<AccessibleMethod> iter = accMethod.iterator(); iter
				.hasNext();) {
			element = iter.next();
			if (set == true) {
				return element;
			} else {
				if (element.number == previous.number)
					set = true;
			}
		}
		return null;
	}

	private void findNeighbours() {
		conSolvMgr.findNighbours();
	}

	void findNighboursPacket(IncomingPacket inPacket) {
		if (isConnectionWith(inPacket.getIdentifier()) == false) {
			if (isPossibleToConnectWith(inPacket.getIdentifier()) == true) {
				conSolvMgr.findNighboursPacket(inPacket);
			}
		}
	}

	/**
	 * Indicate if is connection with node.
	 */
	public boolean isConnectionWith(NetIdentifier node) {
		return remoteInterfaces.containsKey(node);
	}

	/**
	 * Indicate if is possible to connect with another node. The network part of
	 * address must equal.
	 */
	public boolean isPossibleToConnectWith(NetIdentifier node) {
		return identifier.getNetwork().equals(node.getNetwork());
	}

	/**
	 * Add new remote interface and initialize this interface in Queues.
	 */
	public void addRemoteInterface(RemoteInterface ri) {
		remoteInterfaces.put(ri.getIdentifier(), ri);
		fireConnectionAdded(ri);
	}

	/**
	 * Getter of the property <tt>remoteInterfaces</tt>
	 * @return Returns the remoteInterfaces.
	 * @uml.property name="remoteInterfaces"
	 */
	public Map<NetIdentifier, RemoteInterface> getRemoteInterfaces() 
	
	{
		return remoteInterfaces;
	}

	/**
	 * Return broadcast address (it is a multicast address).
	 */
	public InetAddress getBroadcastAddress() {
		return broadcastAddress;
	}

	/**
	 * Return broadcast port.
	 */
	public int getBroadcastPort() {
		return broadcastPort;
	}

	/**
	 * Return true if the accessible method is available.
	 */
	public boolean isAccessiable(AccessibleMethod as) {
		return accMethod.contains(as);
	}

	/**
	 * Return identifier of the interface.
	 */
	public NetIdentifier getIdentifier() {
		return identifier;
	}

	/**
	 * Return true if the interface is on (if is possible to recieving broadcast
	 * packets).
	 */
	public boolean isRunning() {
		return running;
	}

	public void connectionLost(Object sender, RemoteInterface remoteInterface) {
		removeRemoteInterface(remoteInterface);
		fireConnectionLost(remoteInterface);
	}

	/**
	 * Cancel connection and remove remote interface from this interface.s
	 * @param remoteInterface remote interface
	 */
	private void removeRemoteInterface(RemoteInterface remoteInterface) {
		remoteInterfaces.remove(remoteInterface);
		remoteInterface.cancelConnection();
	}

	/**
	 * Add new connection changed listener.
	 */
	public void addConnectionChangedListener(IConnectionChangedListener listener) {
		connectionChangedListeners.add(listener);
	}

	/**
	 * Remove connection changed listener.
	 */
	public void removeConnectionChangedListener(
			IConnectionChangedListener listener) {
		connectionChangedListeners.remove(listener);
	}

	/**
	 * Fire connection changed listener. Method
	 * {@link IConnectionChangedListener.remoteInterfaceAdded} is called.
	 */
	private void fireConnectionAdded(RemoteInterface remoteInterface) {
		for (IConnectionChangedListener listener : connectionChangedListeners) {
			listener.remoteInterfaceAdded(this,remoteInterface);
		}
	}

	/**
	 * Fire connection changed listener. Method
	 * {@link IConnectionChangedListener.remoteInterfaceLost} is called.
	 */
	private void fireConnectionLost(RemoteInterface remoteInterface) {
		for (IConnectionChangedListener listener : connectionChangedListeners) {
			listener.remoteInterfaceLost(this,remoteInterface);
		}
	}

	/**
	 * <b>Title:</b> Smart Active Node<br>
	 * <b>Description:</b> Timer to finding nighbours.<br>
	 * @author Michal Rejda
	 * @since 17.2.2007
	 * @version 1.0
	 */
	private class NighboursTimer {

		private boolean runnning;
		private long time;
		private Thread t;

		public NighboursTimer(long time) {
			this.time = time;
			this.t = new Thread() {

				public void run() {
					try {
						while (runnning) {
							Thread.sleep(TIME_BETWEN_FINDING_NIGHBOURS);
							findNeighbours();
						}
					}
					catch (InterruptedException e) {
						LoggerFactory.createLogger().Log(
								Localization.getLocalization().getString(
										"M000044"), LogType.ERROR);
					}
				}
			};
		}

		public void start() {
			this.runnning = true;
			t.start();
		}

		public void stop() {
			this.runnning = false;
		}
	}
}
