/*************************************************************************
 *                           Smart Active Node                           * 
 *                 copyright (c) since 2007 KIV-FAV-ZCU                  *
 *************************************************************************
 * The license for this code is included in LICENSE.txt.  If you did not *
 * get this file with this distribution, request a copy from:            *
 * txkoutny@kiv.zcu.cz                                                   *
 *************************************************************************/
package san.routing;

import java.util.Hashtable;
import java.util.Map;

import org.apache.log4j.Logger;

import san.ident.Guid;
import san.ident.NetIdentifier;
import san.networkInterfaces.NetInterface;
import san.networkInterfaces.NetInterfaces;
import san.networkInterfaces.RemoteInterface;
import san.networkInterfaces.IInterfacesChanged;
import san.networkInterfaces.IConnectionChangedListener;

/**
 * <b>Title:</b> Smart Active Node<br>
 * <b>Description:</b> Provide basic routing to another nodes in the same
 * network.<br>
 * @author Jan Boháč, Michal Rejda
 * @since 30.10.2006
 * @version 1.1
 */
public class Router
		implements
			IRouting,
			IInterfacesChanged,
			IConnectionChangedListener {

	private static final transient Logger log = Logger.getLogger(Router.class);
	
	/**
	 * @uml.property name="routingTable"
	 * @uml.associationEnd multiplicity="(1 1)" aggregation="shared"
	 *                     inverse="baseRouter:san.routing.RoutingTable"
	 */
	private RoutingTable routingTable = null;
	// key is identifier of interface on this node and value is interface
	// manager
	private Map<NetIdentifier, NetInterface> interfaces;
	// connections to another nodes, key is identifier of the remote node and
	// value is remote interface
	private Map<NetIdentifier, RemoteInterface> connections;
	private Map<Guid, RoutingRecord> defaultGateways;

	public Router(NetInterfaces netInterfaces) {
		netInterfaces.addInterfacesChangedListener(this);
		routingTable = new RoutingTable(this);
		interfaces = new Hashtable<NetIdentifier, NetInterface>();
		connections = new Hashtable<NetIdentifier, RemoteInterface>();
		defaultGateways = new Hashtable<Guid, RoutingRecord>();
	}

	/**
	 * @see san.routing.IRouting#isThisNode(san.ident.NetIdentifier)
	 */
	public boolean isThisNode(NetIdentifier ident) {
		return interfaces.containsKey(ident);
	}

	/**
	 * Return true if the out interface exists.
	 */
	public boolean existOutInterface(NetIdentifier ident) {
		return interfaces.containsKey(ident);
	}

	/**
	 * Return direction to the destination. This direction is nearest neighbour
	 * in the direction. It's a gateway for this destination.
	 */
	public NetIdentifier getDirectionTo(NetIdentifier destination) {
		RoutingRecord rr = routingTable.getRoute(destination);
		NetIdentifier result = null;
		
		if (rr != null)
			return rr.getGateway();
		else
			result = defaultGateways.get(destination.getNetwork()).getGateway();
			if(result == null){
				defaultGateways.get(defaultGateways.keySet().iterator().next()).getGateway() ;
			}
			
			log.debug("pouzivam defaulni cestu: " + result);
			return result;
	}

	/**
	 * Return local network identifier in determined direction. It's a interface
	 * identifier of interface over there lead the way to the destination.
	 */
	public NetIdentifier getLocalNetIdentInDirection(NetIdentifier destination) {
		RoutingRecord rr = routingTable.getRoute(destination);
		NetIdentifier result = null;
		
		if (rr != null)
			return rr.getNetInterface();
		else
			result = defaultGateways.get(destination.getNetwork()).getNetInterface();
		if(result == null){
			defaultGateways.get(defaultGateways.keySet().iterator().next()).getNetInterface();
		}		
		return result;			
	}

	/**
	 * Getter of the property <tt>routingTable</tt>
	 * @return Returns the routingTable.
	 * @uml.property name="routingTable"
	 */
	public RoutingTable getRoutingTable() {
		return routingTable;
	}

	/**
	 * Setter of the property <tt>routingTable</tt>
	 * @param routingTable The routingTable to set.
	 * @uml.property name="routingTable"
	 */
	public void setRoutingTable(RoutingTable routingTable) {
		this.routingTable = routingTable;
	}

	public void interfaceAdded(NetInterface newInterface) {
		interfaces.put(newInterface.getIdentifier(), newInterface);
		newInterface.addConnectionChangedListener(this);
	}

	public void interfaceRemoved(NetInterface removedInterface) {
		interfaces.remove(removedInterface.getIdentifier());
		removedInterface.removeConnectionChangedListener(this);
	}

	public void remoteInterfaceAdded(NetInterface localInterface,
			RemoteInterface remoteInterface) {
		connections.put(remoteInterface.getIdentifier(), remoteInterface);
		// creating routing record
		RoutingRecord rr = new RoutingRecord();
		rr.setDescription("Neighbour");
		rr.setDestination(remoteInterface.getIdentifier());
		rr.setGateway(remoteInterface.getIdentifier());
		rr.setNetInterface(localInterface.getIdentifier());
		rr.setMetric(remoteInterface.getMetric());
		// adding routing record into table
		routingTable.add(rr);
		routingTable.addDirectNeigbour(rr);
		
//		if(defaultGateway == null){
//			defaultGateway = remoteInterface.getIdentifier();
//		}
	}

	public void remoteInterfaceLost(NetInterface localInterface,
			RemoteInterface remoteInterface) {
		connections.remove(remoteInterface.getIdentifier());
		routingTable.remove(remoteInterface.getIdentifier());
		routingTable.removeDirectNeigbour(remoteInterface.getIdentifier());
	}	

	public RoutingRecord getDefaultGateway(Guid net) {
		return defaultGateways.get(net);
	}

	public void setDefaultGateway(Guid net, NetIdentifier defaultGatewayDestination,
			NetIdentifier localInterfaceForDefaultGateway) {
		
		RoutingRecord defaultGateway = new RoutingRecord();		
		defaultGateway.setDescription("default gateway");
		defaultGateway.setGateway(defaultGatewayDestination);
		defaultGateway.setNetInterface(localInterfaceForDefaultGateway);
		defaultGateway.setMetric(-1);	
		
		defaultGateways.put(net, defaultGateway);
		
	}

	public Map<Guid, RoutingRecord> getDefaultGateways() {
		return defaultGateways;
	}
}
