/*************************************************************************
 *                           Smart Active Node                           * 
 *                 copyright (c) since 2007 KIV-FAV-ZCU                  *
 *************************************************************************
 * The license for this code is included in LICENSE.txt.  If you did not *
 * get this file with this distribution, request a copy from:            *
 * txkoutny@kiv.zcu.cz                                                   *
 *************************************************************************/
package san.routing;

import java.util.Map;
import san.ident.NetIdentifier;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Collection;
import java.util.SortedSet;
import java.util.TreeSet;

/**
 * <b>Title:</b> Smart Active Node<br>
 * <b>Description:</b> Represents routing table.<br>
 * @author Michal Rejda
 * @since 30.10.2007
 * @version 1.0
 */
public class RoutingTable implements IRoutingTableControl {

	/**
	 * @uml.property name="baseRouter"
	 * @uml.associationEnd multiplicity="(1 1)"
	 *                     inverse="routingTable:san.routing.Router"
	 */
	private Router baseRouter = null;
	/**
	 * @uml.property name="routingRecords"
	 * @uml.associationEnd multiplicity="(0 -1)" ordering="true"
	 *                     aggregation="composite"
	 *                     inverse="routingTable:san.routing.RoutingRecord"
	 *                     qualifier="destination:san.ident.NetIdentifier
	 *                     san.routing.RoutingRecord"
	 */
	private Map<NetIdentifier, SortedSet<RoutingRecord>> routingRecordsMap;
	/**
	 * @uml.property name="directNeihbours"
	 * @uml.associationEnd multiplicity="(0 -1)" aggregation="composite"
	 *                     inverse="routingTable:san.routing.RoutingRecord"
	 */
	private Collection<RoutingRecord> directNeihbours;

	public RoutingTable(Router baseRouter) {
		this.baseRouter = baseRouter;
		this.routingRecordsMap = new HashMap<NetIdentifier, SortedSet<RoutingRecord>>();
		this.directNeihbours = new ArrayList<RoutingRecord>();
	}

	/**
	 * Add routing record in routing records map.
	 */
	public void add(RoutingRecord record) {
		if (record == null)
			return;
		record.setAge(System.currentTimeMillis());
		SortedSet<RoutingRecord> records = routingRecordsMap.get(record
				.getDestination());
		if (records == null) {
                        records = new TreeSet<RoutingRecord>(new MetricComparator());
                        records.add(record);
                        routingRecordsMap.put(record.getDestination(), records);
		}
		/*RoutingRecord r;
		for (Iterator<RoutingRecord> recIter = records.iterator(); recIter
				.hasNext();) {
			r = recIter.next();
			if (r.getGateway().equals(record.getGateway())) {
				if (r.getMetric() > record.getMetric()) {
					recIter.remove();
				} else {
					record = null;
				}
				break;
			}
		}
		if (record != null)*/
                
                else {
			records.add(record);
                }
	}

	/**
	 * Add routing record in routing records map.
	 */
	public void add(String description, NetIdentifier destination,
			NetIdentifier gateway, int metric, NetIdentifier netInterface) {
		RoutingRecord rr = new RoutingRecord();
		rr.setDescription(description);
		rr.setDestination(destination);
		rr.setGateway(gateway);
		rr.setMetric(metric);
		rr.setNetInterface(netInterface);
		add(rr);
	}

	/**
	 * Add routing record in routing records map.
	 */
	public void add(String description, NetIdentifier destination,
			NetIdentifier gateway, int metric) {
		NetIdentifier netInterface = getRoute(gateway).getNetInterface();
		add(description, destination, gateway, metric, netInterface);
	}

	/**
	 * Add direct neighbour to the table.
	 */
	public void addDirectNeigbour(RoutingRecord rr) {
		directNeihbours.add(rr);
		add(rr);
	}

	/**
	 * Remove routing record from table.
	 */
	public void remove(RoutingRecord record) {
		if (record == null)
			return;
		SortedSet<RoutingRecord> records = routingRecordsMap.get(record
				.getDestination());
		records.remove(record);
		if (records.size() <= 0)
			routingRecordsMap.remove(record.getDestination());
	}

	/**
	 * Remove all routing records with specific destination.
	 */
	public void remove(NetIdentifier destination) {
		if (destination == null)
			return;
		routingRecordsMap.remove(destination);
	}

	/**
	 * Remove routing record with specific destination and gateway.
	 */
	public void remove(NetIdentifier destination, NetIdentifier gateway) {
		if (destination == null || gateway == null)
			return;
		SortedSet<RoutingRecord> records = routingRecordsMap.get(destination);
		RoutingRecord record;
		for (Iterator<RoutingRecord> iterator = records.iterator(); iterator
				.hasNext();) {
			record = (RoutingRecord) iterator.next();
			if (record.getGateway().equals(gateway))
				iterator.remove();
		}
		if (records.size() <= 0)
			routingRecordsMap.remove(destination);
	}

	/**
	 * Remove all records which have gateway same as method parameter.
	 */
	public void removeDirectNeigbour(NetIdentifier neighbour) {
		Collection<SortedSet<RoutingRecord>> recordsCollection = routingRecordsMap
				.values();
		SortedSet<RoutingRecord> records;
		RoutingRecord record;
		for (Iterator<SortedSet<RoutingRecord>> reColIter = recordsCollection
				.iterator(); reColIter.hasNext();) {
			records = reColIter.next();
			for (Iterator<RoutingRecord> recordIter = records.iterator(); recordIter
					.hasNext();) {
				record = recordIter.next();
				if (record.getGateway().equals(neighbour))
					// remove record where gateway is the neighbour
					recordIter.remove();
			}
			if (records.size() <= 0)
				// remove empty collection of routing records
				reColIter.remove();
		}
	}

	/**
	 * Return all routes to specific destination. Return null if any aren't
	 * exist.
	 */
	public SortedSet<RoutingRecord> getRoutes(NetIdentifier destination) {
		if (destination == null)
			return null;
		SortedSet<RoutingRecord> records = routingRecordsMap.get(destination);
		if (records.size() <= 0)
			return null;
		return records;
	}

	/**
	 * Return first best route to destination.
	 */
	public RoutingRecord getRoute(NetIdentifier destination) {
		if (destination == null)
			return null;
		SortedSet<RoutingRecord> records = routingRecordsMap.get(destination);
		if (records == null || records.size() <= 0)
			return null;
		return records.first();
	}

	/**
	 * Return collection of neigbours or null if is empty. 
	 */
	public Collection<RoutingRecord> getDirectNeighbours() {
		if(directNeihbours.size() > 0)
			return directNeihbours;
		else
			return null;
	}
	
	public Map<NetIdentifier, SortedSet<RoutingRecord>> getRoutingRecords() {
		return routingRecordsMap;
	}

	/**
	 * Update metric for specific destination end gateway.
	 */
	public void updateMetric(NetIdentifier destination, NetIdentifier gateway,
			int newMetric) {
		if (destination == null || gateway == null)
			return;
		SortedSet<RoutingRecord> records = routingRecordsMap.get(destination);
		RoutingRecord record;
		for (Iterator<RoutingRecord> iterator = records.iterator(); iterator
				.hasNext();) {
			record = (RoutingRecord) iterator.next();
			if (record.getGateway().equals(gateway))
				record.setMetric(newMetric);
		}
	}

	/**
	 * Getter of the property <tt>baseRouter</tt>
	 * @return Returns the baseRouter.
	 * @uml.property name="baseRouter"
	 */
	public Router getBaseRouter() {
		return baseRouter;
	}

	/**
	 * Setter of the property <tt>baseRouter</tt>
	 * @param baseRouter The baseRouter to set.
	 * @uml.property name="baseRouter"
	 */
	public void setBaseRouter(Router baseRouter) {
		this.baseRouter = baseRouter;
	}

	/**
	 * <b>Title:</b> Smart Active Node<br>
	 * <b>Description:</b> This class provides compare of two RouterRecords on
	 * the metrics.<br>
	 * @author Michal Rejda
	 * @since 31.10.2007
	 * @version 1.0
	 */
	private class MetricComparator implements Comparator<RoutingRecord> {

		public int compare(RoutingRecord o1, RoutingRecord o2) {
			return o1.getMetric() - o2.getMetric();
		}
	}
}
