/*************************************************************************
 *                           Smart Active Node                           * 
 *                 copyright (c) since 2007 KIV-FAV-ZCU                  *
 *************************************************************************
 * The license for this code is included in LICENSE.txt.  If you did not *
 * get this file with this distribution, request a copy from:            *
 * txkoutny@kiv.zcu.cz                                                   *
 *************************************************************************/
package san.softstate;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.swing.Timer;

import san.ident.CodePkgIdent;
import san.ident.Guid;

/**
 * <b>Title:</b> Smart Active Node<br>
 * <b>Description:</b> This class is soft state of the server. Into this class
 * is possible to store user data.<br>
 * @author Michal Rejda
 * @since 8.2.2008
 * @version 1.0
 */
public class SoftState implements ISoftState {

	/**
	 * Default interval between cleaning soft state.
	 */
	public static final int DEFAULT_CLEANER_INTERVAL = 1000;
	private Timer cleanerTimer;
	private Map<Guid, TimedObject> valuesAccordingToGuids;
	private Map<CodePkgIdent, Map<Guid, TimedObject>> valuesAccordingToCodeIdent;

	public SoftState() {
		//initialization of soft state storages
		valuesAccordingToGuids = new HashMap<Guid, TimedObject>();
		valuesAccordingToCodeIdent = new HashMap<CodePkgIdent, Map<Guid,TimedObject>>();
		
		//initialization of the cleaner timer
		initCleanerTimer();
	}

	private void initCleanerTimer() {
		cleanerTimer = new Timer(DEFAULT_CLEANER_INTERVAL,
				new ActionListener() {

					public void actionPerformed(ActionEvent e) {
						cleanSoftState();
					}
				});
		cleanerTimer.setRepeats(true);
		cleanerTimer.start();
	}

	private void cleanSoftState() {

		long localTime = System.currentTimeMillis();
		
		synchronized(valuesAccordingToGuids) {
			//clean values according to guids
			Collection<TimedObject> values = valuesAccordingToGuids.values();
			TimedObject value;
			for (Iterator<TimedObject> iterator = values.iterator(); iterator.hasNext();) {
				value = iterator.next();
				if(value.getExpires() < localTime)
					iterator.remove();
			}
		}
		
		synchronized(valuesAccordingToGuids) {
			//clean values according to code identification
			Collection<Map<Guid,TimedObject>> maps = valuesAccordingToCodeIdent.values();
			Map<Guid,TimedObject> map;
			Collection<TimedObject> values;
			TimedObject value;
			for (Iterator<Map<Guid,TimedObject>> mapsIterator = maps.iterator(); mapsIterator.hasNext();) {
				map = mapsIterator.next();
				values = map.values();
				for (Iterator<TimedObject> iterator = values.iterator(); iterator.hasNext();) {
					value = iterator.next();
					if(value.getExpires() < localTime)
						iterator.remove();
				}	
			}
		}
	}
	
	/**
	 * @see san.softstate.ISoftState#add(java.lang.Object, int)
	 */
	public Guid add(Object value, int time){
		//generating random guid
		Guid guid = Guid.randomGuid();
		//store value
		add(guid, value, time);
		// return guid
		return guid;
	}
	
	/**
	 * @see san.softstate.ISoftState#add(san.ident.Guid, java.lang.Object, int)
	 */
	public void add(Guid guid, Object value, int time){
		synchronized(valuesAccordingToGuids) {
			TimedObject timedObject = new TimedObject(value,System.currentTimeMillis() + time);
			valuesAccordingToGuids.put(guid, timedObject);
		}
	}
	
	/**
	 * @see san.softstate.ISoftState#add(san.ident.CodePkgIdent, san.ident.Guid, java.lang.Object, int)
	 */
	public void add(CodePkgIdent codeIdent, Guid guid, Object value, int time) {
		TimedObject timedObject = new TimedObject(value,System.currentTimeMillis() + time);
		
		synchronized(valuesAccordingToCodeIdent) {
			Map<Guid, TimedObject> valuesAccordingToGuids = valuesAccordingToCodeIdent.get(codeIdent);
			if(valuesAccordingToGuids == null) {
				valuesAccordingToGuids = new HashMap<Guid, TimedObject>();
				valuesAccordingToCodeIdent.put(codeIdent, valuesAccordingToGuids);
			}
			
			valuesAccordingToGuids.put(guid, timedObject);
		}
	}
	
	/**
	 * @see san.softstate.ISoftState#get(san.ident.Guid)
	 */
	public Object get(Guid guid){
		synchronized(valuesAccordingToGuids) {
			TimedObject timedObject = valuesAccordingToGuids.get(guid);
			if(timedObject != null){
				return timedObject.getObject();
			}
			else {
				return null;
			}
		}
	}
	
	/**
	 * @see san.softstate.ISoftState#get(san.ident.CodePkgIdent, san.ident.Guid)
	 */
	public Object get(CodePkgIdent codeIdent, Guid guid){
		synchronized(valuesAccordingToCodeIdent) {
			Map<Guid, TimedObject> valuesAccordingToGuids = valuesAccordingToCodeIdent.get(codeIdent);
			if(valuesAccordingToGuids != null) {
				TimedObject timedObject = valuesAccordingToGuids.get(guid);
				if(timedObject != null){
					return timedObject.getObject();
				}
				else {
					return null;
				}
			} else {
				return null;
			}
		}
	}
}
