import java.util.ArrayList;
import java.util.List;



interface FiberThread {



    public void executeOneInstruction();



    public boolean isAlive();
}



interface Interpreter {



    public void runThread(FiberThread thread);



    public void terminateThread();



    public void start();



    public void stopThread();
}



class SimpleFiberTask implements FiberThread {



    private final int id;

    private int i;
    private double d;



    public SimpleFiberTask(int id) {
        this.id = id;
        i = 0;
        d = 0.0;
    }



    public void executeOneInstruction() {
        if (i == 0) {
            System.out.format("--- Task %d started%n", id);
        }
        d += Math.sin(Math.random() * 55.55) * Math.cos(Math.random() * -33.33) * 0.44;
        if (i % 1000000 == 0) {
            System.out.format("Task %d in iteration %d: %f%n", id, i, d);
        }
        i++;
        if (i >= 100000000) {
            System.out.format("--- Task %d finished%n", id);
        }
    }



    public boolean isAlive() {
        return (i < 100000000);
    }
}



class SimpleInterpreter extends Thread implements Interpreter {



    private final int id;

    private FiberThread scheduledThread;
    private boolean stop;



    public SimpleInterpreter(int id) {
        this.id = id;
        scheduledThread = null;
        stop = false;
    }



    public void runThread(FiberThread thread) {
        scheduledThread = thread;
    }



    public void terminateThread() {
        scheduledThread = null;
    }



    public void stopThread() {
        stop = true;
    }



    public void run() {
        System.out.format("--- Interpreter %d started%n", id);
        while (!stop) {
            FiberThread thread = scheduledThread;
            if (thread != null && thread.isAlive()) {
                thread.executeOneInstruction();
            }
        }
        System.out.format("--- Interpreter %d ended%n", id);
    }
}



class MultiFiberScheduler extends Thread {



    private static final long TIME_QUANTUM = 5000L;
    private static final int CPU_COUNT = 2;

    private final List<FiberThread> threadList;
    private final List<Interpreter> interpreterList;

    private int lastIndex;



    public MultiFiberScheduler(List<FiberThread> threadList) {
        this.threadList = threadList;
        interpreterList = new ArrayList<Interpreter>();
        for (int i = 0; i < CPU_COUNT; i++) {
            interpreterList.add(new SimpleInterpreter(i));
        }
        for (int i = 0; i < CPU_COUNT; i++) {
            interpreterList.get(i).start();
        }
        lastIndex = 0;
    }



    private FiberThread selectThreadToSchedule() {
        if (threadList.isEmpty()) {
            return null;
        }
        if (lastIndex >= threadList.size()) {
            lastIndex = 0;
        }
        if (!threadList.get(lastIndex).isAlive()) {
            threadList.remove(lastIndex);
            return null;
        }
        return threadList.get(lastIndex++);
    }



    public void run() {
        System.out.println("--- Scheduler started");
        FiberThread selectedThread;
        while (!threadList.isEmpty()) {
            for (int i = 0; i < CPU_COUNT && i < threadList.size(); i++) {
                selectedThread = selectThreadToSchedule();
                if (selectedThread != null && selectedThread.isAlive()) {
                    interpreterList.get(i).runThread(selectedThread);
                }
            }
            try {
                Thread.sleep(TIME_QUANTUM);
            } catch (InterruptedException e) {
            }
            for (int i = 0; i < CPU_COUNT; i++) {
                interpreterList.get(i).terminateThread();
            }
        }
        for (int i = 0; i < CPU_COUNT; i++) {
            interpreterList.get(i).stopThread();
        }
        System.out.println("--- Scheduler ended");
    }
}



public class MultiFiberSchedulerTest {



    public static void main(String[] args) {
        List<FiberThread> threadList = new ArrayList<FiberThread>();
        threadList.add(new SimpleFiberTask(0));
        threadList.add(new SimpleFiberTask(1));
        threadList.add(new SimpleFiberTask(2));
        threadList.add(new SimpleFiberTask(3));
        MultiFiberScheduler mfs = new MultiFiberScheduler(threadList);
        mfs.start();
        try {
            mfs.join();
        } catch (InterruptedException e) {
        }
    }
}
