/*************************************************************************
 *                           Smart Active Node                           * 
 *                 copyright (c) since 2007 KIV-FAV-ZCU                  *
 *************************************************************************
 * The license for this code is included in LICENSE.txt.  If you did not *
 * get this file with this distribution, request a copy from:            *
 * txkoutny@kiv.zcu.cz                                                   *
 *************************************************************************/
package san.applicationRepository;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;

import org.apache.log4j.Logger;

import san.ident.CodePkgIdent;

/**
 * <b>Title:</b> Smart Active Node<br>
 * <b>Description:</b> Representation of application package.<br>
 * @author Michal Rejda
 * @since 10.10.2007
 * @version 1.0
 */
public abstract class CodePackage {

	private static final transient Logger log = Logger
			.getLogger(CodePackage.class);
	
	public static final int BUFFER_LENGHT = 1024;
	public static final String PACKAGE_NAME_MFP = "Package-Name";
	public static final String APP_MAIN_CLASS_MFP = "Application-Main-Class";
	public static final String CAPSULE_MAIN_CLASS_MFP = "Capsule-Main-Class";
	protected File aplPkgFile;
	/**
	 * @uml.property name="symLink"
	 * @uml.associationEnd multiplicity="(1 1)" aggregation="shared"
	 *                     inverse="codePackage:san.applicationRepository.CodePackageSymLink"
	 */
	protected CodePackageSymLink symLink = null;

	public CodePackage(CodePackageSymLink symLink, File aplPkgFile)
			throws AplPkgException {
		this.symLink = symLink;
		this.aplPkgFile = aplPkgFile;
	}

	/**
	 * Return applications main class.
	 * @throws FileNotFoundException Raise this exception when application main
	 *             class is not found.
	 */
	public abstract InputStream getApplicationMainClass()
			throws FileNotFoundException, AplPkgException;

	/**
	 * Return capsule main class.
	 * @throws FileNotFoundException Raise this exception when capsule main
	 *             class is not found.
	 */
	public abstract InputStream getCapsuleMainClass()
			throws FileNotFoundException, AplPkgException;

	/**
	 * Return all full class names from the package.
	 */
	public abstract String[] getAllFullClassNamesFromPackage()
			throws AplPkgException;

	/**
	 * Return the class.
	 * @param fullClassName Full name of the class. Package name and class name
	 *            together.
	 * @throws FileNotFoundException Raise this exception when class is not
	 *             found.
	 */
	public abstract InputStream getClass(String fullClassName)
			throws FileNotFoundException, AplPkgException;

	/**
	 * Return the class.
	 * @param packageName Name of the package where is the class defined.
	 * @param className Name of the class.
	 * @throws FileNotFoundException Raise this exception when class is not
	 *             found.
	 */
	public abstract InputStream getClass(String packageName, String className)
			throws FileNotFoundException, AplPkgException;

	/**
	 * Close the package (all open streams will be closed).
	 */
	public void close() throws IOException {
		// decrement code package number of opens
		synchronized (symLink) {
			symLink.setNumberOfOpens(symLink.getNumberOfOpens() - 1);
			log.debug("package closed " + symLink.getPkgIdentification() + ", num of opens for " + symLink.getNumberOfOpens());
		}
	}

	/**
	 * Get complete package as byte array.
	 */
	public byte[] getPkgBytes() {
		try {
			ByteArrayOutputStream bab = new ByteArrayOutputStream();
			FileInputStream fis = new FileInputStream(aplPkgFile);
			byte[] buff = new byte[BUFFER_LENGHT];
			int red = 0;
			while ((red = fis.read(buff)) > 0) {
				bab.write(buff, 0, red);
			}
			fis.close();
			return bab.toByteArray();
		}
		catch (Exception e) {
			return null;
		}
	}

	/**
	 * Return extension of the file. It identifies application package type.
	 */
	public String getExtension() {
		String extensiion = aplPkgFile.getName();
		extensiion = extensiion.substring(extensiion.lastIndexOf(".") + 1);
		return extensiion;
	}

	/**
	 * Return manifest property value.
	 */
	public abstract String getManifestPropertyValue(String property)
			throws AplPkgException;

	/**
	 * @uml.property name="pkgName"
	 */
	private String pkgName = "";

	/**
	 * Getter of the property <tt>pkgName</tt>
	 * @return Returns the pkgName.
	 * @uml.property name="pkgName"
	 */
	public String getPkgName() {
		return pkgName;
	}

	/**
	 * Setter of the property <tt>pkgName</tt>
	 * @param pkgName The pkgName to set.
	 * @uml.property name="pkgName"
	 */
	public void setPkgName(String pkgName) {
		this.pkgName = pkgName;
	}

	/**
	 * @uml.property name="since"
	 */
	private long since;

	/**
	 * Getter of the property <tt>since</tt>
	 * @return Returns the since.
	 * @uml.property name="since"
	 */
	public long getSince() {
		return since;
	}

	/**
	 * Setter of the property <tt>since</tt>
	 * @param since The since to set.
	 * @uml.property name="since"
	 */
	public void setSince(long since) {
		this.since = since;
	}

	/**
	 * @uml.property name="pkgIdentification"
	 */
	private CodePkgIdent pkgIdentification;

	/**
	 * Getter of the property <tt>pkgIdentification</tt>
	 * @return Returns the pkgIdentification.
	 * @uml.property name="pkgIdentification"
	 */
	public CodePkgIdent getPkgIdentification() {
		return pkgIdentification;
	}

	/**
	 * Setter of the property <tt>pkgIdentification</tt>
	 * @param pkgIdentification The pkgIdentification to set.
	 * @uml.property name="pkgIdentification"
	 */
	public void setPkgIdentification(CodePkgIdent pkgIdentification) {
		this.pkgIdentification = pkgIdentification;
	}
}
