/*************************************************************************
 *                           Smart Active Node                           * 
 *                 copyright (c) since 2007 KIV-FAV-ZCU                  *
 *************************************************************************
 * The license for this code is included in LICENSE.txt.  If you did not *
 * get this file with this distribution, request a copy from:            *
 * txkoutny@kiv.zcu.cz                                                   *
 *************************************************************************/
package san.console;

import java.awt.Font;
import java.awt.Insets;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.io.PrintWriter;
import java.io.Writer;
import java.util.Vector;

import javax.swing.JFrame;
import javax.swing.JScrollPane;
import javax.swing.JTextPane;
import javax.swing.text.BadLocationException;
import javax.swing.text.DefaultStyledDocument;

import san.localization.Localization;

/**
 * <b>Title:</b> Smart Active Node<br>
 * <b>Description:</b> Smart active node graphic console.<br>
 * @author Michal Rejda
 * @since 28.10.2007
 * @version 1.0
 */
public class JConsole extends JScrollPane implements KeyListener, IConsole {

	private OutputStream outPipe;
	private InputStream in;
	private int cmdStart = 0;
	private Vector<String> history = new Vector<String>();
	private String startedLine;
	private int histLine = 0;
	private JTextPane text;
	private DefaultStyledDocument doc;

	public JConsole() {
		text = new JTextPane(doc = new DefaultStyledDocument()) {

			public void cut() {
				if (text.getCaretPosition() < cmdStart) {
					super.copy();
				} else {
					super.cut();
				}
			}

			public void paste() {
				forceCaretMoveToEnd();
				super.paste();
			}
		};
		Font font = new Font("Monospaced", Font.PLAIN, 12);
		text.setText("");
		text.setFont(font);
		text.setMargin(new Insets(5, 5, 5, 5));
		text.addKeyListener(this);
		setViewportView(text);
		outPipe = new PipedOutputStream();
		try {
			in = new PipedInputStream((PipedOutputStream) outPipe);
		}
		catch (IOException e) {
			// not a exception
		}
		text.requestFocus();
	}

	public void keyPressed(KeyEvent e) {
		type(e);
	}

	public void keyTyped(KeyEvent e) {
		type(e);
	}

	public void keyReleased(KeyEvent e) {
		type(e);
	}

	private synchronized void type(KeyEvent e) {
		switch (e.getKeyCode()) {
			case (KeyEvent.VK_ENTER ) :
				if (e.getID() == KeyEvent.KEY_PRESSED) {
					enter();
					resetCommandStart();
					text.setCaretPosition(cmdStart);
				}
				e.consume();
				text.repaint();
				break;
			case (KeyEvent.VK_UP ) :
				if (e.getID() == KeyEvent.KEY_PRESSED) {
					historyUp();
				}
				e.consume();
				break;
			case (KeyEvent.VK_DOWN	) :
				if (e.getID() == KeyEvent.KEY_PRESSED) {
					historyDown();
				}
				e.consume();
				break;
			case (KeyEvent.VK_LEFT	) :
			case (KeyEvent.VK_RIGHT ) :
			case (KeyEvent.VK_BACK_SPACE ) :
			case (KeyEvent.VK_DELETE ) :
				if (text.getCaretPosition() <= cmdStart) {
					e.consume();
				}
				break;
			case (KeyEvent.VK_HOME ) :
				text.setCaretPosition(cmdStart);
				e.consume();
				break;
			case (KeyEvent.VK_U ) : // clear line
				if ((e.getModifiers() & InputEvent.CTRL_MASK) > 0) {
					replaceRange("", cmdStart, text.getDocument().getLength());
					histLine = 0;
					e.consume();
				}
				break;
			case (KeyEvent.VK_ALT ) :
			case (KeyEvent.VK_CAPS_LOCK ) :
			case (KeyEvent.VK_CONTROL ) :
			case (KeyEvent.VK_META ) :
			case (KeyEvent.VK_SHIFT ) :
			case (KeyEvent.VK_PRINTSCREEN ) :
			case (KeyEvent.VK_SCROLL_LOCK ) :
			case (KeyEvent.VK_PAUSE ) :
			case (KeyEvent.VK_INSERT ) :
			case (KeyEvent.VK_TAB) :
			case (KeyEvent.VK_F1) :
			case (KeyEvent.VK_F2) :
			case (KeyEvent.VK_F3) :
			case (KeyEvent.VK_F4) :
			case (KeyEvent.VK_F5) :
			case (KeyEvent.VK_F6) :
			case (KeyEvent.VK_F7) :
			case (KeyEvent.VK_F8) :
			case (KeyEvent.VK_F9) :
			case (KeyEvent.VK_F10) :
			case (KeyEvent.VK_F11) :
			case (KeyEvent.VK_F12) :
			case (KeyEvent.VK_ESCAPE ) :
				// only modifier pressed
				break;
			// Control-C
			case (KeyEvent.VK_C ) :
				if (text.getSelectedText() == null) {
					if (((e.getModifiers() & InputEvent.CTRL_MASK) > 0)
							&& (e.getID() == KeyEvent.KEY_PRESSED)) {
						append("^C");
					}
					e.consume();
				}
				break;
			default :
				if ((e.getModifiers() & (InputEvent.CTRL_MASK
						| InputEvent.ALT_MASK | InputEvent.META_MASK)) == 0) {
					// plain character
					forceCaretMoveToEnd();
				}
				break;
		}
	}

	private void resetCommandStart() {
		cmdStart = text.getDocument().getLength();
	}

	private void append(String string) {
		int slen = text.getDocument().getLength();
		text.select(slen, slen);
		text.replaceSelection(string);
	}

	private String replaceRange(Object s, int start, int end) {
		String st = s.toString();
		text.select(start, end);
		text.replaceSelection(st);
		return st;
	}

	private void forceCaretMoveToEnd() {
		if (text.getCaretPosition() < cmdStart) {
			// move caret first!
			text.setCaretPosition(text.getDocument().getLength());
		}
		text.repaint();
	}

	private void enter() {
		String s = getCmd();
		if (s.length() == 0) // special hack for empty return!
			s = ";\n";
		else {
			history.addElement(s);
			s = s + "\n";
		}
		append("\n");
		histLine = 0;
		acceptLine(s);
		text.repaint();
	}

	private String getCmd() {
		String s = "";
		try {
			s = text.getText(cmdStart, text.getDocument().getLength()
					- cmdStart);
		}
		catch (BadLocationException e) {
			// should not happen
		}
		return s;
	}

	private void historyUp() {
		if (history.size() == 0)
			return;
		if (histLine == 0) // save current line
			startedLine = getCmd();
		if (histLine < history.size()) {
			histLine++;
			showHistoryLine();
		}
	}

	private void historyDown() {
		if (histLine == 0)
			return;
		histLine--;
		showHistoryLine();
	}

	private void showHistoryLine() {
		String showline;
		if (histLine == 0)
			showline = startedLine;
		else
			showline = (String) history.elementAt(history.size() - histLine);
		replaceRange(showline, cmdStart, text.getDocument().getLength());
		text.setCaretPosition(text.getDocument().getLength());
		text.repaint();
	}

	String ZEROS = "000";

	private void acceptLine(String line) {
		StringBuffer buf = new StringBuffer();
		int lineLength = line.length();
		for (int i = 0; i < lineLength; i++) {
			char c = line.charAt(i);
			if (c > 127) {
				// patch for unicode characters
				String val = Integer.toString(c, 16);
				val = ZEROS.substring(0, 4 - val.length()) + val;
				buf.append("\\u" + val);
			} else {
				buf.append(c);
			}
		}
		line = buf.toString();
		try {
			outPipe.write(line.getBytes());
			outPipe.flush();
		}
		catch (IOException e) {
			outPipe = null;
			throw new RuntimeException(Localization.getLocalization().getString("M000014"));
		}
		text.repaint();
	}

	protected void println(String o) {
		print(String.valueOf(o) + "\n");
		text.repaint();
	}

	protected void print(final String o) {
		append(String.valueOf(o));
		resetCommandStart();
		text.setCaretPosition(cmdStart);
	}

	public void setFont(Font font) {
		super.setFont(font);
		if (text != null)
			text.setFont(font);
	}
	
	@Override
	public void grabFocus() {
		super.grabFocus();
		text.grabFocus();
	}
	
	/**
	 * @see san.console.IConsole#getIn()
	 */
	public InputStream getIn() {
		return in;
	}
	
	/**
	 * @see san.console.IConsole#getOut()
	 */
	public PrintWriter getOut() {
		return new PrintWriter((new ConsoleWriter()));
	}

	public static void main(String[] args) {
		JFrame jf = new JFrame();//for console testing
		jf.setContentPane(new JConsole());
		jf.setVisible(true);
	}
	
	/** 
	 * <b>Title:</b> Smart Active Node<br> 
	 * <b>Description:</b> This class provides writing into console.<br>
	 * @author Michal Rejda
	 * @since 7.11.2007
	 * @version 1.0
	 */
	private class ConsoleWriter extends Writer {
		
		@Override
		public Writer append(char c) throws IOException {
			print(Character.toString(c));
			return this;
		}
		
		@Override
		public Writer append(CharSequence csq) throws IOException {
			print(csq.toString());
			return this;
		}
		
		@Override
		public Writer append(CharSequence csq, int start, int end) {
			print(csq.subSequence(start, end).toString());
			return this;
		}
		
		@Override
		public void write(char[] cbuf) throws IOException {
			write(cbuf, 0, cbuf.length);
		}
		
		@Override
		public void write(int c) throws IOException {
			print(Integer.toString(c));
		}
		
		@Override
		public void write(String str) throws IOException {
			print(str);
		}
		
		@Override
		public void write(String str, int off, int len) throws IOException {
			print(str.substring(off, off+len));
		}
		
		@Override
		public void write(char[] cbuf, int off, int len) throws IOException {
			for (int i = off; i < len+off || i < cbuf.length; i++) {
				print(Character.toString(cbuf[i]));
			}
		}
		
		@Override
		public void flush() throws IOException {
			//nothing to do
		}
		
		@Override
		public void close() throws IOException {
			//nothing to do
		}
	}
}
