/*************************************************************************
 *                           Smart Active Node                           * 
 *                 copyright (c) since 2007 KIV-FAV-ZCU                  *
 *************************************************************************
 * The license for this code is included in LICENSE.txt.  If you did not *
 * get this file with this distribution, request a copy from:            *
 * txkoutny@kiv.zcu.cz                                                   *
 *************************************************************************/
package san.core;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import san.core.fiber.api.IInfoAPI;
import san.security.SecurityMonitor;
import san.applicationRepository.CodePackage;
import san.ident.Guid;
import san.console.IConsole;
import san.core.fiber.api.IConditionAPI;
import san.core.fiber.api.IInfoGetterAPI;
import san.core.fiber.api.IMonitorAPI;
import san.core.fiber.api.ISleepAPI;
import san.core.fiber.condition.ConditionRequest;
import san.core.fiber.monitor.MonitorRequest;
import san.interpreter.IInterpretingContext;
import san.interpreter.sanint.fiber.call.ISpecialCall;

/**
 * <b>Title:</b> Smart Active Node<br>
 * <b>Description:</b> Each process parent class.<br>
 * @author Michal Rejda
 * @since 20.10.2007
 * @version 1.0
 */
public abstract class Process implements IConditionAPI, IMonitorAPI, ISleepAPI, IInfoGetterAPI {



    /**
	 * <b>Title:</b> Smart Active Node<br>
	 * <b>Description:</b> Priority of the process.<br>
	 * @author Michal Rejda
	 * @since 26.10.2007
	 * @version 1.0
	 */
	public enum PriorityClass {
		LOW,
        NORMAL,
        HIGH,
        REAL_TIME
	}



    /**
	 * <b>Title:</b> Smart Active Node<br>
	 * <b>Description:</b> States of the process.<br>
	 * @author Michal Rejda
	 * @since 25.10.2007
	 * @version 1.0
	 */
	public enum State {
		RUNNING,
        READY_TO_RUN,
        STOPPED,
        ASLEEP,
        FINISHED,
        BLOCKED
	}



    public enum BlockedReason {
        JOIN,
        MONITOR,
        CONDITION
    }



    public enum Type {
        APPLICATION,
        CAPSULE,
        SPECIAL_CALL
    }



    /**
	 * This constant is for advantage of priority when this process is running
	 * now. This constant is used by scheduler by process planing.
	 */
	protected static final long ADVANTAGE_FOR_RUNNING = 10;
	/**
	 * @uml.property name="state"
	 */
	protected volatile State state = State.READY_TO_RUN;
	/**
	 * @uml.property name="priorityClass"
	 */
	protected PriorityClass priorityClass = PriorityClass.NORMAL;
	/**
	 * @uml.property name="codePackage"
	 * @uml.associationEnd multiplicity="(1 1)" aggregation="composite"
	 *                     inverse="executable:san.applicationRepository.CodePackage"
	 */
	protected CodePackage codePackage = null;
	/**
	 * @uml.property name="interpretingContext"
	 * @uml.associationEnd multiplicity="(1 1)" aggregation="composite"
	 *                     inverse="executable:san.interpreter.IInterpretingContext"
	 */
	protected IInterpretingContext executionContext = null;
	/**
	 * @uml.property name="kernel"
	 * @uml.associationEnd readOnly="true" inverse="process:san.core.Kernel"
	 */
	protected Kernel kernel;
	/**
	 * @uml.property name="id"
	 */
	protected Guid id = null;
	/**
	 * @uml.property name="parent"
	 */
	private Guid parent = null;
	/**
	 * @uml.property name="console"
	 */
	protected IConsole console = null;

    private String Owner = "anonymous";



    private final Type type;
    private final List<Guid> waitingForGuidList;
    private final ConditionRequest conditionRequest;
    private final MonitorRequest monitorRequest;

    private volatile String mainMethodName;
    private volatile String mainMethodDescriptor;
    private volatile Object returnValue;
    private volatile boolean returnValueSet;
    private volatile boolean nowInterpreted;
    private volatile int cpuIndex;
    private volatile ISpecialCall specialCall;
    private volatile long sleepTime;
    private volatile BlockedReason blockedReason;



    public Process(Kernel kernel, Type type, String mainMethodName, String mainMethodDescriptor) {
		this.kernel = kernel;
        this.type = type;
        waitingForGuidList = new LinkedList<Guid>();
        conditionRequest = new ConditionRequest();
        monitorRequest = new MonitorRequest();
        this.mainMethodName = mainMethodName;
        this.mainMethodDescriptor = mainMethodDescriptor;
        returnValue = null;
        returnValueSet = false;
        nowInterpreted = false;
        cpuIndex = 0;
        specialCall = null;
        sleepTime = 0L;
        blockedReason = null;
	}

	/**
	 * @uml.property name="id"
	 */
	public Guid getId() {
		return id;
	}

	/**
	 * Getter of the property <tt>codePackage</tt>
	 * @return Returns the codePackage.
	 * @uml.property name="codePackage"
	 */
	public CodePackage getCodePackage() {
		return codePackage;
	}

	/**
	 * Setter of the property <tt>codePackage</tt>
	 * @param codePackage The codePackage to set.
	 * @uml.property name="codePackage"
	 */
	public void setCodePackage(CodePackage codePackage) {
		this.codePackage = codePackage;
	}

	/**
	 * Getter of the property <tt>interpretingContext</tt>
	 * @return Returns the executionContext.
	 * @uml.property name="interpretingContext"
	 */
	public IInterpretingContext getInterpretingContext() {
		return executionContext;
	}

	/**
	 * Setter of the property <tt>interpretingContext</tt>
	 * @param interpretingContext The executionContext to set.
	 */
	public void setInterpretingContext(IInterpretingContext interpretingContext) {
		executionContext = interpretingContext;
	}

	/**
	 * Getter of the property <tt>kernel</tt>
	 * @return Returns the kernel.
	 * @uml.property name="kernel"
	 */
	Kernel getKernel() {
		return kernel;
	}

	/**
	 * Getter of the property <tt>state</tt>
	 * @return Returns the state.
	 * @uml.property name="state"
	 */
	public State getState() {
		return state;
	}

	/**
	 * Setter of the property <tt>state</tt>
	 * @param state The state to set.
	 * @uml.property name="state"
	 */
	public void setState(State state) {
		this.state = state;
	}

	/**
	 * @uml.property name="startInterpretingAt"
	 */
	private long startInterpretingAt;

	/**
	 * Getter of the property <tt>startInterpretingAt</tt>
	 * @return Returns the startInterpretingAt.
	 * @uml.property name="startInterpretingAt"
	 */
	public long getStartInterpretingAt() {
		return startInterpretingAt;
	}

	/**
	 * @uml.property name="endedInterpretingAt"
	 */
	private long endedInterpretingAt;

	/**
	 * Getter of the property <tt>endedInterpretingAt</tt>
	 * @return Returns the endedInterpretingAt.
	 * @uml.property name="endedInterpretingAt"
	 */
	public long getEndedInterpretingAt() {
		return endedInterpretingAt;
	}

	/**
	 * Setter of the property <tt>endedInterpretingAt</tt>
	 * @param endedInterpretingAt The endedInterpretingAt to set.
	 * @uml.property name="endedInterpretingAt"
	 */
	public void setEndedInterpretingAt(long endedInterpretingAt) {
		this.endedInterpretingAt = endedInterpretingAt;
	}

	/**
	 * Setter of the property <tt>startInterpretingAt</tt>
	 * @param startInterpretingAt The startInterpretingAt to set.
	 * @uml.property name="startInterpretingAt"
	 */
	public void setStartInterpretingAt(long startInterpretingAt) {
		this.startInterpretingAt = startInterpretingAt;
	}

	/**
	 * @uml.property name="lastUsedQuantum"
	 */
	private long lastUsedQuantum;

	/**
	 * Getter of the property <tt>lastUsedQuantum</tt>
	 * @return Returns the lastUsedQuantum.
	 * @uml.property name="lastUsedQuantum"
	 */
	public long getLastUsedQuantum() {
		return lastUsedQuantum;
	}

	/**
	 * Setter of the property <tt>lastUsedQuantum</tt>
	 * @param lastUsedQuantum The lastUsedQuantum to set.
	 * @uml.property name="lastUsedQuantum"
	 */
	public void setLastUsedQuantum(long lastUsedQuantum) {
		this.lastUsedQuantum = lastUsedQuantum;
	}

	/**
	 * @uml.property name="basePriority"
	 */
	private long basePriority;

	/**
	 * Getter of the property <tt>basePriority</tt>
	 * @return Returns the basePriority.
	 * @uml.property name="basePriority"
	 */
	public long getBasePriority() {
		return basePriority;
	}

	/**
	 * Setter of the property <tt>basePriority</tt>
	 * @param basePriority The basePriority to set.
	 * @uml.property name="basePriority"
	 */
	public void setBasePriority(long basePriority) {
		this.basePriority = basePriority;
	}

	/**
	 * @uml.property name="nicePriority"
	 */
	private long nicePriority;

	/**
	 * Getter of the property <tt>nicePriority</tt>
	 * @return Returns the nicePriority.
	 * @uml.property name="nicePriority"
	 */
	public long getNicePriority() {
		return nicePriority;
	}

	/**
	 * Setter of the property <tt>nicePriority</tt>
	 * @param nicePriority The nicePriority to set.
	 * @uml.property name="nicePriority"
	 */
	public void setNicePriority(long nicePriority) {
		this.nicePriority = nicePriority;
	}

	/**
	 * Return priority number. Bigger number is smaller priority. It is SUM of
	 * base priority, last used quantum, negative nice priority and negative
	 * ADVANTAGE_FOR_RUNNING constant if is running..
	 */
	public long getPriorityNumber() {
		if (state == State.RUNNING)
			return basePriority + lastUsedQuantum - nicePriority
					- ADVANTAGE_FOR_RUNNING;
		else
			return basePriority + lastUsedQuantum - nicePriority;
	}

	/**
	 * Setter of the property <tt>id</tt>
	 * @param id The id to set.
	 * @uml.property name="id"
	 */
	public void setId(Guid id) {
		this.id = id;
	}

	/**
	 * Getter of the property <tt>priorityClass</tt>
	 * @return Returns the priorityClass.
	 * @uml.property name="priorityClass"
	 */
	public PriorityClass getPriorityClass() {
		return priorityClass;
	}

	/**
	 * Setter of the property <tt>priorityClass</tt>
	 * @param priorityClass The priorityClass to set.
	 * @uml.property name="priorityClass"
	 */
	public void setPriorityClass(PriorityClass priorityClass) {
		this.priorityClass = priorityClass;
	}

	/**
	 * Getter of the property <tt>parent</tt>
	 * @return Returns the parent.
	 * @uml.property name="parent"
	 */
	public Guid getParent() {
		return parent;
	}

	/**
	 * Setter of the property <tt>parent</tt>
	 * @param parent The parent to set.
	 * @uml.property name="parent"
	 */
	public void setParent(Guid parent) {
		this.parent = parent;
	}

	/**
	 * @return the console
	 * @uml.property name="console"
	 */
	public IConsole getConsole() {
		return console;
	}

	/**
	 * @param console the console to set
	 * @uml.property name="console"
	 */
	public void setConsole(IConsole console) {
		this.console = console;
	}

        public void setOwner(String owner) {

            SecurityMonitor monitor = san.security.SecurityMonitor.getInstance();

            if (monitor.isUserValid(owner)) {
                this.Owner = owner;
            }
        }

        public String getOwner() {
            return this.Owner;
        }

        public void supressOwnerToParent()
        {
            this.kernel.getScheduler().getProcess(this.getParent()).setOwner(this.getOwner());
        }



    public Type getType() {
        if (specialCall != null) {
            return Type.SPECIAL_CALL;
        }
        return type;
    }



    public String getMainMethodName() {
        return mainMethodName;
    }



    public void setMainMethodName(String mainMethodName) {
        this.mainMethodName = mainMethodName;
    }



    public String getMainMethodDescriptor() {
        return mainMethodDescriptor;
    }



    public void setMainMethodDescriptor(String mainMethodDescriptor) {
        this.mainMethodDescriptor = mainMethodDescriptor;
    }



    public Object getReturnValue() {
        return returnValue;
    }



    public boolean isReturnValueSet() {
        return returnValueSet;
    }



    public void setReturnValue(Object returnValue) {
        this.returnValue = returnValue;
        returnValueSet = true;
    }



    public boolean isNowInterpreted() {
        return nowInterpreted;
    }



    public void setNowInterpreted(boolean nowInterpreted) {
        this.nowInterpreted = nowInterpreted;
    }



    public int getCpuIndex() {
        return cpuIndex;
    }



    public void setCpuIndex(int cpuIndex) {
        this.cpuIndex = cpuIndex;
    }



    public ISpecialCall getSpecialCall() {
        return specialCall;
    }



    public void setSpecialCall(ISpecialCall specialCall) {
        this.specialCall = specialCall;
    }



    public long getSleepTime() {
        return sleepTime;
    }



    public long getTimeToWakeUp() {
        return (endedInterpretingAt + sleepTime);
    }



    public void setBlockedState(BlockedReason blockedReason) {
        this.blockedReason = blockedReason;
        state = State.BLOCKED;
    }



    public BlockedReason getBlockedReason() {
		return blockedReason;
	}



    public Iterator<Guid> waitingForGuidListIterator() {
        return waitingForGuidList.iterator();
    }



    public int waitingForGuidListSize() {
        return waitingForGuidList.size();
    }



    public boolean isWaitingForGuidListEmpty() {
        return waitingForGuidList.isEmpty();
    }



    public void addGuidWaitingFor(Guid guidWaitingFor) {
        waitingForGuidList.add(guidWaitingFor);
    }



    public ConditionRequest getConditionRequest() {
        return conditionRequest;
    }



    public MonitorRequest getMonitorRequest() {
        return monitorRequest;
    }



    @Override
    public void sleep(long timeout) throws IllegalArgumentException {
        if (timeout < 0L) {
            throw new IllegalArgumentException("Sleep timeout value can not be negative");
        }
        sleepTime = timeout;
        state = State.ASLEEP;
    }



    @Override
    public void yield() {
        sleepTime = 0L;
        state = State.ASLEEP;
    }



    @Override
    public boolean conditionInitialize(int conditionId) {
        conditionRequest.setCall(conditionId, ConditionRequest.Type.INITIALIZE);
        setBlockedState(BlockedReason.CONDITION);
        return false;
    }



    @Override
    public boolean conditionDestroy(int conditionId) {
        conditionRequest.setCall(conditionId, ConditionRequest.Type.DESTROY);
        setBlockedState(BlockedReason.CONDITION);
        return false;
    }



    @Override
    public void conditionWait(int conditionId) {
        conditionRequest.setCall(conditionId, ConditionRequest.Type.WAIT);
        setBlockedState(BlockedReason.CONDITION);
    }



    @Override
    public boolean conditionWait(int conditionId, long timeout) throws IllegalArgumentException {
        if (timeout < 0L) {
            throw new IllegalArgumentException("Sleep timeout value is negative");
        }
        sleepTime = timeout;
        conditionRequest.setCall(conditionId, ConditionRequest.Type.WAIT_TIMEOUT);
        setBlockedState(BlockedReason.CONDITION);
        return false;
    }



    @Override
    public void conditionSignalOne(int conditionId) {
        conditionRequest.setCall(conditionId, ConditionRequest.Type.SIGNAL_ONE);
        setBlockedState(BlockedReason.CONDITION);
    }



    @Override
    public void conditionSignalAll(int conditionId) {
        conditionRequest.setCall(conditionId, ConditionRequest.Type.SIGNAL_ALL);
        setBlockedState(BlockedReason.CONDITION);
    }



    @Override
    public int conditionGetWaitingCount(int conditionId) {
        conditionRequest.setCall(conditionId, ConditionRequest.Type.GET_WAITING_COUNT);
        setBlockedState(BlockedReason.CONDITION);
        return (-1);
    }



    @Override
    public void monitorEnter(Object monitor) {
        monitorRequest.setCall(monitor, MonitorRequest.Type.ENTER);
        setBlockedState(BlockedReason.MONITOR);
    }



    @Override
    public void monitorExit(Object monitor) {
        monitorRequest.setCall(monitor, MonitorRequest.Type.EXIT);
        setBlockedState(BlockedReason.MONITOR);
    }



    @Override
    public boolean monitorTest(Object monitor) {
        monitorRequest.setCall(monitor, MonitorRequest.Type.TEST);
        setBlockedState(BlockedReason.MONITOR);
        return false;
    }



    @Override
    public boolean monitorTestAndEnter(Object monitor) {
        monitorRequest.setCall(monitor, MonitorRequest.Type.TEST_AND_ENTER);
        setBlockedState(BlockedReason.MONITOR);
        return false;
    }



    public Object[] monitorGetSequence() {
        return monitorRequest.getMonitorSequenceArray();
    }



    @Override
    public void monitorSetSequence(Object[] monitorSequenceArray) {
        monitorRequest.setMonitorSequenceArray(monitorSequenceArray);
    }



    @Override
    public IInfoAPI getInfo() {
        return kernel.getScheduler().getInfo();
    }



    public String getDebugInfo() {
        String stateString;
        String processName;
        if (state == State.BLOCKED) {
            stateString = String.format("%s (%s)", state.toString(), blockedReason.toString());
        } else if (state == State.ASLEEP) {
            stateString = String.format("%s (%d [ms])", state.toString(), sleepTime);
        } else {
            stateString = state.toString();
        }
        if (specialCall != null) {
            processName = String.format("%s (call %s)",
                                        codePackage.getPkgName(),
                                        specialCall.getType().toString());
        } else {
            processName = codePackage.getPkgName();
        }
        if (this instanceof ProcessApplication) {
            return String.format("Application %s, %s, %s, CPU %d, %s",
                                 processName,
                                 id.toString(),
                                 stateString,
                                 cpuIndex,
                                 java.util.Arrays.toString(((ProcessApplication) this).getArgs()));
        } else if (this instanceof ProcessCapsule) {
            return String.format("Capsule %s, %s, CPU %d",
                                 processName,
                                 stateString,
                                 cpuIndex);
        } else {
            return "Unknown process";
        }
    }
}
