package san.core.fiber.monitor;



import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import org.apache.log4j.Logger;
import san.core.Process;
import san.core.fiber.log.FiberLoggerFactory;
import san.core.fiber.monitor.deadlock.DeadlockDetector;
import san.core.fiber.scheduler.ISchedulerModule;



/**
 *
 * @author Jiri Nohavec
 */
public class MonitorManager implements ISchedulerModule {



    private static final Logger logger = FiberLoggerFactory.getSchedulerLogger();

    private final Map<Object, MonitorWrap> monitorWrapMap;
    private final List<Process> sequenceProcessList;
    private final List<Object> monitorToRemoveList;
    private final DeadlockDetector deadlockDetector;



    public MonitorManager() {
        monitorWrapMap = new HashMap<Object, MonitorWrap>();
        sequenceProcessList = new LinkedList<Process>();
        monitorToRemoveList = new ArrayList<Object>();
        deadlockDetector = new DeadlockDetector(monitorWrapMap);
    }



    @Override
    public synchronized void checkForStoppedProcesses() throws MonitorException {
        MonitorWrap monitorWrap;
        boolean isFirst;
        boolean unblockNext;
        monitorToRemoveList.clear();
        for (Object monitor: monitorWrapMap.keySet()) {
            monitorWrap = monitorWrapMap.get(monitor);
            isFirst = true;
            unblockNext = false;
            for (Iterator<Process> iterator = monitorWrap.getProcessIterator(); iterator.hasNext(); ) {
                Process process = iterator.next();
                switch (process.getState()) {
                    case FINISHED:
                    case STOPPED:
                        iterator.remove();
                        if (isFirst) {
                            monitorWrap.resetLockCount();
                            unblockNext = true;
                            if (logger.isDebugEnabled()) {
                                logger.debug(String.format("    MON - process removed from monitor %s, %s",
                                                           monitor.toString(),
                                                           process.getDebugInfo()));
                            }
                        }
                        break;
                    default:
                        if (!isFirst) {
                            if (logger.isDebugEnabled()) {
                                logger.debug(String.format("    MON - waiting on monitor %s, %s",
                                                           monitor,
                                                           process.getDebugInfo()));
                            }
                        }
                        break;
                }
                isFirst = false;
            }
            if (monitorWrap.isProcessQueueEmpty()) {
                monitorToRemoveList.add(monitor);
            } else if (unblockNext) {
                monitorWrap.getHeadProcess().setState(Process.State.READY_TO_RUN);
                monitorWrap.incrementLockCount();
            }
        }
        for (Object monitor: monitorToRemoveList) {
            monitorWrapMap.remove(monitor);
        }
        monitorToRemoveList.clear();
        checkForBlockedProcessesWithSequence();
        deadlockDetector.detect();
    }



    private boolean isMonitorSequenceFree(Process process) {
        MonitorWrap sequenceMonitorWrap;
        if (process.monitorGetSequence() != null) {
            for (Object sequenceMonitor: process.monitorGetSequence()) {
                sequenceMonitorWrap = monitorWrapMap.get(sequenceMonitor);
                if (sequenceMonitorWrap != null
                        && !sequenceMonitorWrap.isProcessQueueEmpty()
                        && !sequenceMonitorWrap.isHeadProcess(process)) {
                    return false;
                }
            }
        }
        return true;
    }



    private void checkForBlockedProcessesWithSequence() throws MonitorException {
        for (Iterator<Process> iterator = sequenceProcessList.iterator(); iterator.hasNext(); ) {
            Process process = iterator.next();
            if (isMonitorSequenceFree(process)) {
                Object monitor = process.getMonitorRequest().getMonitor();
                MonitorWrap monitorWrap = monitorWrapMap.get(monitor);
                if (monitorWrap == null) {
                    monitorWrap = new MonitorWrap(monitor);
                    monitorWrapMap.put(monitor, monitorWrap);
                }
                if (actionEnterWithoutSequence(process, monitorWrap)) {
                    process.setState(Process.State.READY_TO_RUN);
                    iterator.remove();
                }
            }
            switch (process.getState()) {
                case FINISHED:
                case STOPPED:
                    iterator.remove();
                    break;
                default:
                    if (logger.isDebugEnabled()) {
                        logger.debug(String.format("    MON - waiting on monitor sequence, %s",
                                                   process.getDebugInfo()));
                    }
                    break;
            }
        }
    }



    private boolean actionEnter(Process process,
                                MonitorWrap monitorWrap) throws MonitorException {
        if (process.monitorGetSequence() != null) {
            for (Object sequenceMonitor: process.monitorGetSequence()) {
                if (sequenceMonitor.equals(monitorWrap.getMonitor())) {
                    return actionEnterWithSequence(process, monitorWrap);
                }
            }
        }
        return actionEnterWithoutSequence(process, monitorWrap);
    }



    private boolean actionEnterWithoutSequence(Process process,
                                               MonitorWrap monitorWrap) throws MonitorException {
        if ((!monitorWrap.isHeadProcess(process) && monitorWrap.containsProcess(process))
                || (monitorWrap.isProcessQueueEmpty() && monitorWrap.getLockCount() != 0)
                || monitorWrap.getLockCount() < 0) {
            throw new MonitorException(String.format("Failure when entering monitor %s, %s",
                                                     monitorWrap.getMonitor().toString(),
                                                     process.getDebugInfo()));
        }
        if (monitorWrap.isProcessQueueEmpty()) {
            monitorWrap.addProcess(process);
            monitorWrap.incrementLockCount();
        } else if (monitorWrap.getHeadProcess().equals(process)) {
            monitorWrap.incrementLockCount();
        } else {
            monitorWrap.addProcess(process);
            return false;
        }
        return true;
    }



    private boolean actionEnterWithSequence(Process process,
                                            MonitorWrap monitorWrap) throws MonitorException {
        if (isMonitorSequenceFree(process)) {
            return actionEnterWithoutSequence(process, monitorWrap);
        } else {
            sequenceProcessList.add(process);
            return false;
        }
    }



    private void actionExit(Process process, MonitorWrap monitorWrap) throws MonitorException {
        if (!monitorWrap.isHeadProcess(process)
                || monitorWrap.isProcessQueueEmpty()
                || monitorWrap.getLockCount() <= 0) {
            throw new MonitorException(String.format("Failure when leaving monitor %s, %s",
                                                     monitorWrap.getMonitor().toString(),
                                                     process.getDebugInfo()));
        }
        monitorWrap.decrementLockCount();
        if (monitorWrap.getLockCount() == 0) {
            monitorWrap.removeHeadProcess();
            if (!monitorWrap.isProcessQueueEmpty()) {
                monitorWrap.getHeadProcess().setState(Process.State.READY_TO_RUN);
                monitorWrap.incrementLockCount();
            }
        }
    }



    private void actionTest(Process process, MonitorWrap monitorWrap) throws MonitorException {
        if (monitorWrap.isHeadProcess(process) || monitorWrap.isProcessQueueEmpty()) {
            process.getInterpretingContext().replaceReturnValue(true);
        } else {
            process.getInterpretingContext().replaceReturnValue(false);
        }
    }



    private boolean actionTestAndEnter(Process process, MonitorWrap monitorWrap) throws MonitorException {
        if (monitorWrap.isHeadProcess(process) || monitorWrap.isProcessQueueEmpty()) {
            process.getInterpretingContext().replaceReturnValue(true);
            return actionEnter(process, monitorWrap);
        } else {
            process.getInterpretingContext().replaceReturnValue(false);
            return true;
        }
    }



    @Override
    public synchronized void checkProcessCall(Process process) throws MonitorException {
        MonitorRequest monitorRequest = process.getMonitorRequest();
        if (!monitorRequest.isNewCall()) {
            return;
        }
        Object monitor = monitorRequest.getMonitor();
        MonitorRequest.Type type = monitorRequest.getType();
        MonitorWrap monitorWrap;
        boolean unblockProcess = true;
        if (logger.isDebugEnabled()) {
            logger.debug(String.format("    MON - call %s, monitor %s, %s",
                                       type.toString(),
                                       monitor.toString(),
                                       process.getDebugInfo()));
        }
        if (!monitorWrapMap.containsKey(monitor)) {
            monitorWrap = new MonitorWrap(monitor);
            monitorWrapMap.put(monitor, monitorWrap);
        } else {
            monitorWrap = monitorWrapMap.get(monitor);
        }
        switch (type) {
            case ENTER:
                unblockProcess = actionEnter(process, monitorWrap);
                break;
            case EXIT:
                actionExit(process, monitorWrap);
                break;
            case TEST:
                actionTest(process, monitorWrap);
                break;
            case TEST_AND_ENTER:
                unblockProcess = actionTestAndEnter(process, monitorWrap);
                break;
        }
        if (monitorWrap.isProcessQueueEmpty()) {
            monitorWrapMap.remove(monitor);
        }
        monitorRequest.newCallReceived();
        if (unblockProcess) {
            process.setState(Process.State.RUNNING);
        }
    }
}
