package san.core.fiber.monitor.deadlock;



import java.util.Map;
import org.apache.log4j.Logger;
import san.core.Process;
import san.core.fiber.log.FiberLoggerFactory;
import san.core.fiber.monitor.MonitorWrap;
import san.core.fiber.settings.FiberSettings;



/**
 *
 * @author Jiri Nohavec
 */
public class DeadlockDetector {



    private static final Logger logger = FiberLoggerFactory.getSchedulerLogger();

    private final boolean DETECTION_ENABLED;
    private final long DETECTION_SLEEP_INTERVAL;

    private final Map<Object, MonitorWrap> monitorWrapMap;
    private final RagGraph ragGraph;

    private long lastDetectionTime;



    public DeadlockDetector(Map<Object, MonitorWrap> monitorWrapMap) {
        DETECTION_ENABLED = FiberSettings.getInstance().isDeadlockDetectorEnabled();
        DETECTION_SLEEP_INTERVAL = FiberSettings.getInstance().getDeadlockDetectorSleepInterval();
        this.monitorWrapMap = monitorWrapMap;
        ragGraph = new RagGraph();
        lastDetectionTime = System.currentTimeMillis();
    }



    private boolean isDetectionTime() {
        long currentTime = System.currentTimeMillis();
        if (lastDetectionTime + DETECTION_SLEEP_INTERVAL < currentTime) {
            lastDetectionTime = currentTime;
            return true;
        } else {
            return false;
        }
    }



    public void detect() {
        if (!DETECTION_ENABLED || !isDetectionTime()) {
            return;
        }
        boolean isFirst;
        for (Object monitor: monitorWrapMap.keySet()) {
            isFirst = true;
            for (Process process: monitorWrapMap.get(monitor).getProcessIterable()) {
                if (isFirst) {
                    isFirst = false;
                    ragGraph.addEdge(monitor, process);
                } else {
                    ragGraph.addEdge(process, monitor);
                }
            }
        }
        if (ragGraph.hasCycle()) {
            if (logger.isDebugEnabled()) {
                logger.debug(String.format("    MON - deadlock scan found cycle:"));
            }
            IRagNodeCycle bestRagNodeCycle = null;
            for (IRagNodeCycle ragNodeCycle: ragGraph.getRagNodeCycleListIterable()) {
                if (ragNodeCycle.getObject() instanceof Process) {
                    if (bestRagNodeCycle == null) {
                        bestRagNodeCycle = ragNodeCycle;
                    } else if (bestRagNodeCycle.waitingForThisListSize() > ragNodeCycle.waitingForThisListSize()) {
                        bestRagNodeCycle = ragNodeCycle;
                    }
                    if (logger.isDebugEnabled()) {
                        Process process = ((Process) ragNodeCycle.getObject());
                        logger.debug(String.format("        + process, %s", process.getDebugInfo()));
                    }
                } else {
                    if (logger.isDebugEnabled()) {
                        logger.debug(String.format("        + monitor, %s", ragNodeCycle.getObject().toString()));
                    }
                }
            }
            if (bestRagNodeCycle != null) {
                Process process = ((Process) bestRagNodeCycle.getObject());
                process.setState(Process.State.STOPPED);
                logger.warn(String.format("    MON - terminated process to solve deadlock, %s",
                                          process.getDebugInfo()));
            }
        } else {
            if (logger.isDebugEnabled()) {
                logger.debug(String.format("    MON - deadlock scan OK"));
            }
        }
        ragGraph.clear();
    }
}
