package san.core.fiber.scheduler;



import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Queue;
import org.apache.log4j.Logger;
import san.applicationRepository.AplPkgException;
import san.applicationRepository.BshCodePkg;
import san.applicationRepository.CodePackage;
import san.applicationRepository.IRepository;
import san.codeDistribution.strategies.CodeDistributionStrategy;
import san.core.ExecuteException;
import san.core.IScheduler;
import san.core.Kernel;
import san.core.Process;
import san.core.ProcessApplication;
import san.core.ProcessCapsule;
import san.core.fiber.api.IInfoAPI;
import san.core.fiber.condition.ConditionManager;
import san.core.fiber.info.FiberInfoHttpServer;
import san.core.fiber.log.FiberLoggerFactory;
import san.core.fiber.log.FiberUncaughtExceptionHandler;
import san.core.fiber.monitor.MonitorManager;
import san.core.fiber.settings.FiberSettings;
import san.core.fiber.strategy.ISchedulerStrategy;
import san.core.fiber.strategy.StrategyManager;
import san.data.TransmitCapsule;
import san.ident.CodePkgIdent;
import san.ident.Guid;
import san.incomingQueue.IncomingQueue;
import san.interpreter.IInterpreter;
import san.interpreter.InterpretManager;



/**
 *
 * @author Jiri Nohavec
 */
public class FiberScheduler implements IScheduler {



    private static final Logger logger = FiberLoggerFactory.getSchedulerLogger();

    private static final long LOW_PRIORITY_QUANTUM = 10L;
    private static final long NORMAL_PRIORITY_QUANTUM = 100L;
    private static final long HIGH_PRIORITY_QUANTUM = 200L;
    private static final long REAL_TIME_PRIORITY_QUANTUM = 500L;

    private final int CPU_COUNT;
    private final int MAX_CAPSULES_RUNNING;
    private final int MAX_SAME_CODE_THREADS_RUNNING;
    private final long SCHEDULER_INTERVAL_SLEEP;

    private final Kernel kernel;
    private final IRepository repository;
    private final IncomingQueue incomingQueue;
    private final InterpretManager interpretManager;
    private final CodeDistributionStrategy codeDistributionStrategy;
    private final FiberSchedulerMailbox fiberSchedulerMailbox;
    private final Queue<Process> allProcessQueue;
    private final Queue<Process> runningProcessQueue;
    private final Queue<Process> beanShellProcessQueue;
    private final Queue<TransmitCapsule> noCodeCapsuleQueue;
    private final Queue<TransmitCapsule> haveCodeCapsuleQueue;
    private final ISchedulerStrategy schedulerStrategy;
    private final ISchedulerModule conditionManager;
    private final ISchedulerModule monitorManager;
    private final Thread schedulerThread;
    private final Thread capsulesPreparationThread;
    private final FiberInfoHttpServer fiberInfoHttpServer;

    private volatile boolean running;

    private long lastScheduleTime;



    public FiberScheduler(Kernel kernel,
                          IRepository repository,
                          IncomingQueue incomingQueue,
                          InterpretManager interpretManager,
                          CodeDistributionStrategy codeDistributionStrategy) {
        CPU_COUNT = FiberSettings.getInstance().getVirtualCPUCount();
        MAX_CAPSULES_RUNNING = FiberSettings.getInstance().getMaxCapsulesRunning();
        MAX_SAME_CODE_THREADS_RUNNING = 500;
        SCHEDULER_INTERVAL_SLEEP = FiberSettings.getInstance().getSchedulerSleepInterval();
        this.kernel = kernel;
        this.repository = repository;
        this.incomingQueue = incomingQueue;
        this.interpretManager = interpretManager;
        this.codeDistributionStrategy = codeDistributionStrategy;
        fiberSchedulerMailbox = new FiberSchedulerMailbox();
        allProcessQueue = new LinkedList<Process>();
        runningProcessQueue = new LinkedList<Process>();
        beanShellProcessQueue = new LinkedList<Process>();
        noCodeCapsuleQueue = new LinkedList<TransmitCapsule>();
        haveCodeCapsuleQueue = new LinkedList<TransmitCapsule>();
        schedulerStrategy = StrategyManager.getStrategy(FiberSettings.getInstance().getSchedulerStrategy(),
                                                        CPU_COUNT,
                                                        allProcessQueue);
        conditionManager = new ConditionManager();
        monitorManager = new MonitorManager();
        schedulerThread = new Thread() {
            @Override
            public void run() {
                while (running) {
                    schedule();
                }
            }
        };
        schedulerThread.setName("Fiber scheduler thread");
        schedulerThread.setUncaughtExceptionHandler(new FiberUncaughtExceptionHandler(logger));
        capsulesPreparationThread = new Thread() {
            @Override
            public void run() {
                while (running) {
                    prepareCapsule();
                }
            }
        };
        capsulesPreparationThread.setName("Capsules preparation thread");
        capsulesPreparationThread.setUncaughtExceptionHandler(new FiberUncaughtExceptionHandler(logger));
        fiberInfoHttpServer = new FiberInfoHttpServer(fiberSchedulerMailbox.getFiberInfo());
        running = true;
        startThread();
    }



    private void startThread() {
        schedulerThread.start();
        capsulesPreparationThread.start();
        fiberInfoHttpServer.startServer();
    }



    public void stopThread() {
        running = false;
        fiberSchedulerMailbox.notifyAllConditions();
        fiberInfoHttpServer.stopServer();
    }



    @Override
    public Process getProcess(Guid guid) {
        return fiberSchedulerMailbox.getProcess(guid);
    }



    @Override
    public Collection<Process> getProcess(CodePkgIdent codePkgIdent) {
        return fiberSchedulerMailbox.getProcess(codePkgIdent);
    }



    public int getProcessCount(CodePkgIdent codePkgIdent) {
        return fiberSchedulerMailbox.getProcessCount(codePkgIdent);
    }



    private void checkSleepingProcess(Process process, long currentTime) {
        if (process.getSleepTime() == 0L) {
            process.setLastUsedQuantum(SCHEDULER_INTERVAL_SLEEP);
            process.setState(Process.State.READY_TO_RUN);
            if (logger.isDebugEnabled()) {
                logger.debug(String.format("SCH - yield sleep, %s", process.getDebugInfo()));
            }
        } else if (process.getTimeToWakeUp() < currentTime) {
            process.setState(Process.State.READY_TO_RUN);
            if (logger.isDebugEnabled()) {
                logger.debug(String.format("SCH - sleep finished, %s", process.getDebugInfo()));
            }
        } else {
            if (logger.isDebugEnabled()) {
                logger.debug(String.format("SCH - still sleeping for %d [ms], %s",
                                           process.getTimeToWakeUp() - currentTime,
                                           process.getDebugInfo()));
            }
        }
    }



    private void checkBlockedProcess(Process process) {
        if (process.getBlockedReason() == Process.BlockedReason.JOIN) {
            for (Iterator<Guid> iterator = process.waitingForGuidListIterator(); iterator.hasNext(); ) {
                if (getProcess(iterator.next()) == null) {
                    iterator.remove();
                }
            }
            if (process.isWaitingForGuidListEmpty()) {
                process.setState(Process.State.READY_TO_RUN);
                if (logger.isDebugEnabled()) {
                    logger.debug(String.format("SCH - join finished, %s", process.getDebugInfo()));
                }
            } else {
                if (logger.isDebugEnabled()) {
                    logger.debug(String.format("SCH - still waiting to join for %d apps, %s",
                                               process.waitingForGuidListSize(),
                                               process.getDebugInfo()));
                }
            }
        }
    }



    private void schedule() {
        do {
            fiberSchedulerMailbox.waitOnSchedulerSleepCondition(SCHEDULER_INTERVAL_SLEEP);
        } while (running
                && allProcessQueue.isEmpty()
                && !fiberSchedulerMailbox.hasNewProcess()
                && !fiberSchedulerMailbox.hasIntepretingEndedProcess());
        if (!running) {
            return;
        }
        long currentTime = System.currentTimeMillis();
        if (logger.isDebugEnabled()) {
            logger.debug(String.format("SCH - new round --------------------------------------------------"));
        }
        for (Process process: runningProcessQueue) {
            process.setLastUsedQuantum(currentTime - process.getStartInterpretingAt());
        }
        while (fiberSchedulerMailbox.hasNewProcess()) {
            Process process = fiberSchedulerMailbox.removeNewProcess();
            if (process.getCodePackage() instanceof BshCodePkg) {
                process.setStartInterpretingAt(currentTime);
                beanShellProcessQueue.add(process);
                interpretManager.getBshInterpreter().runProcess(process);
                continue;
            }
            allProcessQueue.add(process);
        }
        while (fiberSchedulerMailbox.hasIntepretingEndedProcess()) {
            Process process = fiberSchedulerMailbox.removeIntepretingEndedProcess();
            process.setEndedInterpretingAt(currentTime);
            if (beanShellProcessQueue.contains(process)) {
                beanShellProcessQueue.remove(process);
                fiberSchedulerMailbox.sendFinishedProcess(process);
                kernel.getRepository().dontProtectAppFromDeleteAnyLonger(process.getCodePackage()
                                                                         .getPkgIdentification());
                kernel.getRepository().tryToSendNextRequestForApp();
                continue;
            }
            if (allProcessQueue.contains(process)) {
                runningProcessQueue.remove(process);
            }
        }
        for (Iterator<Process> iterator = allProcessQueue.iterator(); iterator.hasNext(); ) {
            Process process = iterator.next();
            if (process.getState() != Process.State.RUNNING) {
                process.setLastUsedQuantum(process.getLastUsedQuantum() >>> 1);
            }
            switch (process.getState()) {
                case FINISHED:
                case STOPPED:
                    iterator.remove();
                    runningProcessQueue.remove(process);
                    fiberSchedulerMailbox.sendFinishedProcess(process);
                    kernel.getRepository().dontProtectAppFromDeleteAnyLonger(process.getCodePackage()
                                                                             .getPkgIdentification());
                    kernel.getRepository().tryToSendNextRequestForApp();
                    break;
                case ASLEEP:
                    checkSleepingProcess(process, currentTime);
                    break;
                case BLOCKED:
                    checkBlockedProcess(process);
                    break;
            }
        }
        try {
            conditionManager.checkForStoppedProcesses();
        } catch (Exception e) {
            logger.warn("SCH - Error when running condition check", e);
        }
        try {
            monitorManager.checkForStoppedProcesses();
        } catch (Exception e) {
            logger.warn("SCH - Error when running monitor check", e);
        }
        schedulerStrategy.rescheduleCall();
        for (int cpuIndex = 0; cpuIndex < CPU_COUNT; cpuIndex++) {
            executeProcess(schedulerStrategy.getSelectedProcess(cpuIndex), cpuIndex, currentTime);
        }
        if (runningProcessQueue.size() > CPU_COUNT) {
            throw new SchedulerException("More running processes than available CPUs");
        }
        lastScheduleTime = currentTime;
    }



    private void executeProcess(Process process, int cpuIndex, long currentTime) {
        IInterpreter interpreter = interpretManager.getSanFiberInterpreter(cpuIndex);
        if (process == interpreter.getProcess()) {
            return;
        }
        if (interpreter.isInterpreting()) {
            Process terminateProcess = interpreter.terminate();
            terminateProcess.setEndedInterpretingAt(currentTime);
            if (allProcessQueue.contains(terminateProcess)) {
                runningProcessQueue.remove(terminateProcess);
            }
        }
        if (process == null) {
            return;
        }
        if (!runningProcessQueue.contains(process)) {
            runningProcessQueue.add(process);
            process.setStartInterpretingAt(currentTime);
        }
        interpreter.runProcess(process);
        if (logger.isDebugEnabled()) {
            logger.debug(String.format("SCH - scheduled CPU %d, %s", cpuIndex, process.getDebugInfo()));
        }
    }



    @Override
    public void intepretingEnded(Process process) {
        if (process.getState() == Process.State.BLOCKED) {
            switch (process.getBlockedReason()) {
                case CONDITION:
                    conditionManager.checkProcessCall(process);
                    break;
                case MONITOR:
                    monitorManager.checkProcessCall(process);
                    break;
            }
        }
        if (process.getState() == Process.State.RUNNING) {
            return;
        }
        fiberSchedulerMailbox.sendIntepretingEndedProcess(process);
    }



    @Override
    public void waitForApplication(Guid guid,
                                   Collection<Guid> guidWaitingForCollection) throws InterruptedException {
        Process waitingProcess = getProcess(guid);
        if (waitingProcess == null) {
            return;
        }
        if (waitingProcess.getCodePackage() instanceof BshCodePkg) {
            boolean waiting = true;
            while (waiting) {
                waiting = false;
                for (Guid guidWaitingFor: guidWaitingForCollection) {
                    if (getProcess(guidWaitingFor) != null) {
                        waiting = true;
                        break;
                    }
                }
                if (waiting) {
                    fiberSchedulerMailbox.waitOnProcessWaitCondition();
                }
            }
            return;
        }
        for (Guid guidWaitingFor: guidWaitingForCollection) {
            if (getProcess(guidWaitingFor) != null) {
                waitingProcess.addGuidWaitingFor(guidWaitingFor);
            }
        }
        if (waitingProcess.isWaitingForGuidListEmpty()) {
            return;
        }
        waitingProcess.setBlockedState(Process.BlockedReason.JOIN);
        fiberSchedulerMailbox.sendIntepretingEndedProcess(waitingProcess);
    }



    @Override
    public Guid runApplication(CodePkgIdent codePkgIdent,
                               String[] attributeArray,
                               Process.PriorityClass priorityClass,
                               Guid parentGuid,
                               CodePackage codePackage) throws ExecuteException {
        if (fiberSchedulerMailbox.getProcessCount(codePkgIdent) >= MAX_SAME_CODE_THREADS_RUNNING) {
            throw new ExecuteException("Can not start new process, thread limit count for this code reached");
        }
        ProcessApplication processApplication = new ProcessApplication(kernel);
        if (priorityClass == null) {
            priorityClass = Process.PriorityClass.NORMAL;
        }
        processApplication.setPriorityClass(priorityClass);
        switch (priorityClass) {
            case LOW:
                processApplication.setBasePriority(LOW_PRIORITY_QUANTUM);
                break;
            case HIGH:
                processApplication.setBasePriority(HIGH_PRIORITY_QUANTUM);
                break;
            case REAL_TIME:
                processApplication.setBasePriority(REAL_TIME_PRIORITY_QUANTUM);
                break;
            case NORMAL:
                processApplication.setBasePriority(NORMAL_PRIORITY_QUANTUM);
                break;
            default:
                throw new ExecuteException("Unknown priority class selected");
        }
        processApplication.setId(Guid.randomGuid());
        processApplication.setArgs(attributeArray);
        processApplication.setCodePackage(codePackage);
        processApplication.setParent(parentGuid);
        if (parentGuid != null) {
            Process parentProcess = getProcess(parentGuid);
            if (parentProcess != null) {
                processApplication.setOwner(parentProcess.getOwner());
                processApplication.setConsole(parentProcess.getConsole());
            }
        }
        processApplication.setState(Process.State.READY_TO_RUN);
        fiberSchedulerMailbox.sendNewProcess(processApplication);
        return processApplication.getId();
    }



    @Override
    public Guid runApplication(CodePkgIdent codePkgIdent,
                               String[] attributeArray,
                               Process.PriorityClass priorityClass,
                               Guid parentGuid) throws ExecuteException {
        try {
            CodePackage codePackage = repository.getCodePkg(codePkgIdent);
            return runApplication(codePkgIdent, attributeArray, priorityClass, parentGuid, codePackage);
        } catch (AplPkgException e) {
            throw new ExecuteException(e.getMessage());
        }
    }



    @Override
    public Guid runSpecialProcess(ProcessApplication processApplication, CodePackage codePackage) {
        processApplication.setPriorityClass(Process.PriorityClass.HIGH);
        processApplication.setBasePriority(HIGH_PRIORITY_QUANTUM);
        processApplication.setId(Guid.randomGuid());
        processApplication.setCodePackage(codePackage);
        processApplication.setState(Process.State.READY_TO_RUN);
        fiberSchedulerMailbox.sendNewProcess(processApplication);
        return processApplication.getId();
    }



    private void prepareCapsule() {
        TransmitCapsule transmitCapsule = null;
        CodePackage codePackage = null;
        synchronized (haveCodeCapsuleQueue) {
            if (!haveCodeCapsuleQueue.isEmpty()) {
                transmitCapsule = haveCodeCapsuleQueue.remove();
            }
        }
        if (transmitCapsule == null) {
            transmitCapsule = incomingQueue.getCapsule();
        }
        if (transmitCapsule == null) {
            return;
        }
        try {
            codePackage = repository.getCodePkg(transmitCapsule.getHash());
        } catch (AplPkgException e) {
            logger.warn("SCH - Can not get code package for capsule", e);
        }
        if (codePackage == null) {
            synchronized (noCodeCapsuleQueue) {
                noCodeCapsuleQueue.add(transmitCapsule);
            }
            kernel.getRepository().tryToSendNextRequestForApp(transmitCapsule);
            return;
        }
        ProcessCapsule processCapsule = new ProcessCapsule(kernel);
        processCapsule.setPriorityClass(Process.PriorityClass.HIGH);
        processCapsule.setBasePriority(HIGH_PRIORITY_QUANTUM);
        processCapsule.setCapsule(transmitCapsule);
        processCapsule.setCodePackage(codePackage);
        processCapsule.setState(Process.State.READY_TO_RUN);
        while (fiberSchedulerMailbox.getCapsulesCurrentlyRunning() >= MAX_CAPSULES_RUNNING) {
            fiberSchedulerMailbox.waitOnProcessWaitCondition();
            if (!running) {
                return;
            }
        }
        fiberSchedulerMailbox.sendNewProcess(processCapsule);
    }



    @Override
    public void codeExist(CodePkgIdent codePkgIdent) {
        Queue<TransmitCapsule> justGotCodeCapsuleQueue = new LinkedList<TransmitCapsule>();
        boolean changed;
        synchronized (noCodeCapsuleQueue) {
            for (Iterator<TransmitCapsule> iterator = noCodeCapsuleQueue.iterator(); iterator.hasNext(); ) {
                TransmitCapsule transmitCapsule = iterator.next();
                if (transmitCapsule.getHash().equals(codePkgIdent)) {
                    iterator.remove();
                    justGotCodeCapsuleQueue.add(transmitCapsule);
                }
            }
        }
        synchronized (haveCodeCapsuleQueue) {
            changed = haveCodeCapsuleQueue.addAll(justGotCodeCapsuleQueue);
        }
        if (changed && capsulesPreparationThread.getState() != Thread.State.RUNNABLE) {
            capsulesPreparationThread.interrupt();
        }
    }



    @Override
    public IInfoAPI getInfo() {
        return fiberSchedulerMailbox.getFiberInfo();
    }
}
