/*************************************************************************
 *                           Smart Active Node                           * 
 *                      copyright (C) 2007 KIV-FAV-ZCU                   *
 *************************************************************************
 * The licence for this code is included in LICENSE.txt.  If you did not *
 * get this file with this distribution, request a copy from:            *
 * txkoutny@kiv.zcu.cz                                                   *
 *************************************************************************/

package san.interpreter.BSH;

import java.io.InputStreamReader;
import java.util.LinkedList;
import java.util.Queue;

import org.apache.log4j.Logger;

import bsh.Interpreter;
import san.applicationRepository.BshCodePkg;
import san.core.IApplicationAPI;
import san.core.ICapsuleAPI;
import san.core.Kernel;
import san.core.Process;
import san.core.ProcessApplication;
import san.core.ProcessCapsule;
import san.data.TransmitCapsule;
import san.ident.Guid;
import san.interpreter.IInterpreter;
import san.interpreter.InterpretManager;
import san.localization.Localization;
import san.logging.LoggerFactory;
import san.logging.ILogger.LogType;

/**
 * <b>Title:</b> Smart Active Node<br>
 * <b>Description:</b> Interpret the application.<br>
 * @author Michal Rejda
 * @since 21.2.2007
 * @version 1.1
 */
public class BshInterpreter extends Thread implements IInterpreter {

    private static int threadId = 0;

    private static final transient Logger log = Logger.getLogger(BshInterpreter.class);
	private InterpretManager interpretManager;
	private boolean running;
	private Queue<TransmitCapsule> data;
	private String[] attributes;
	private Kernel kernel;
	private Guid runningAplIdent;
	private boolean startByUser;
	private Process executed = null;
	
	public BshInterpreter(InterpretManager interpretManager,Kernel kernel) {
		this.interpretManager = interpretManager;
		this.kernel = kernel;
		this.data = new LinkedList<TransmitCapsule>();
        this.setName(String.format("BSH interpreter thread %d", threadId));
        threadId++;
	}

	/** In this method is the application started and running. */
	public void run() {
		this.running = true;
		try{
			BshCodePkg codePackage = (BshCodePkg) executed.getCodePackage();
			Interpreter i = new Interpreter();
			//set context of the transmitCapsule
			i.setExitOnEOF(false);
			if(executed instanceof ProcessApplication) {
				executed.setState(san.core.Process.State.RUNNING);
				ProcessApplication pa = (ProcessApplication) executed;
				String mainClass = codePackage.getManifestPropertyValue(BshCodePkg.APP_MAIN_CLASS_MFP);
				//interpreting program
				InputStreamReader isr = new InputStreamReader(codePackage.getApplicationMainClass());
				i.eval(isr);
				i.set("applicationAPI", (IApplicationAPI) pa);
                i.set("Guid", pa.getId());
				i.set("args", pa.getArgs());
				i.eval(mainClass + " main = new "+mainClass+"();");	//add main mehod params
				i.eval("main.main(args,applicationAPI,Guid);");	//run the program
				executed.setState(san.core.Process.State.FINISHED);
				isr.close();
			} else if (executed instanceof ProcessCapsule) {
				
				executed.setState(san.core.Process.State.RUNNING);
				ProcessCapsule pc = (ProcessCapsule) executed;
				String mainClass = codePackage.getManifestPropertyValue(BshCodePkg.CAPSULE_MAIN_CLASS_MFP);
				//interpreting program
				InputStreamReader isr = new InputStreamReader(codePackage.getCapsuleMainClass());
				i.eval(isr);
				i.set("capsuleAPI", (ICapsuleAPI) pc);
				i.eval(mainClass + " main = new "+mainClass+"();");	//add main mehod param
				i.eval("main.main(capsuleAPI);");	//run the program
				executed.setState(san.core.Process.State.FINISHED);
				isr.close();
				
				if(pc.isSendCapsuleOnExit()) {
										
					TransmitCapsule tc = pc.getCapsule();
					tc.setOver(null);
					tc.setPrev(null);
//					tc.setSrc(null);
					
					if (log.isDebugEnabled()) {
						log.debug("send capsule on exit (" + tc + ")");
					}
					
					kernel.sendCapsule(tc);
				}
				
			}
		}
		catch(Exception e) {
			LoggerFactory.createLogger().Log(Localization.getLocalization().getString("M000031")+" "+e,LogType.ERROR);
			e.printStackTrace();
		} finally {
			//application was ended
			this.running = false;
			//we say to application manager of end this application.
			interpretManager.interpretingEnded(this);
		}
	}
	
	public Process getProcess() {
		return executed;
	}
	
	public boolean isInterpreting() {
		return running;
	}
	
	public void runProcess(Process executable) {
		this.executed = executable;
		if(executable instanceof ProcessApplication){
			attributes = ((ProcessApplication) executable).getArgs();
		} else if(executable instanceof ProcessCapsule){
			data.add(((ProcessCapsule) executable).getCapsule());
		}
		this.start();
	}
	
	public Process terminate() {
		return null;
	}
	
	public synchronized void addData(TransmitCapsule data) {
		this.data.add(data);
		notifyAll();
	}
	
	/**
	 * @return the runningAplIdent
	 */
	public Guid getRunningAplIdent() {
		return runningAplIdent;
	}

	/**
	 * @param runningAplIdent the runningAplIdent to set
	 */
	public void setRunningAplIdent(Guid runningAplIdent) {
		this.runningAplIdent = runningAplIdent;
	}
	
	/**
	 * @param startByUser the startByUser to set
	 */
	public void setStartByUser(boolean startByUser) {
		this.startByUser = startByUser;
	}
}
