package san.interpreter.sanint.fiber;



import org.apache.log4j.Logger;
import san.core.Process;
import san.core.fiber.log.FiberLoggerFactory;
import san.interpreter.sanint.classLoading.Array;
import san.interpreter.sanint.classLoading.ClassManager;
import san.interpreter.sanint.classLoading.HardClass;
import san.interpreter.sanint.classLoading.Reference;
import san.interpreter.sanint.classLoading.ReferenceConverter;
import san.interpreter.sanint.classLoading.SoftReference;
import san.interpreter.sanint.fiber.monitor.MonitorValidator;
import san.interpreter.sanint.fiber.stack.CallStackLevel;
import san.interpreter.sanint.interpreter.Const;
import san.interpreter.sanint.interpreter.InterpreterStack;
import san.interpreter.sanint.interpreter.LocalVars;
import san.interpreter.sanint.interpreter.References;



/**
 *
 * @author Jiri Nohavec
 */
public class FiberInterpreterBytecode {



    private static final Logger logger = FiberLoggerFactory.getInterpreterLogger();

    private static final boolean LOG_EVERY_INSTRUCTION = false;

    private final int cpuIndex;
    private final ReferenceConverter referenceConverter;

    private Process lastProcess;
    private ClassManager classManager;
    private CallStackLevel callStackLevel;
    private byte[] bytecode;
    private HardClass currentClass;
    private References references;
    private InterpreterStack stack;
    private LocalVars localVars;
    private int returnStackPointer;
    private int index;



    public FiberInterpreterBytecode(int cpuIndex) {
        this.cpuIndex = cpuIndex;
        referenceConverter = new ReferenceConverter();
        lastProcess = null;
        callStackLevel = null;
    }



    private void setReturnValue(Process process) throws Exception {
        String mainMethodDescriptor = process.getMainMethodDescriptor();
        String returnValueDescriptor = mainMethodDescriptor.substring(mainMethodDescriptor.lastIndexOf(')') + 1);
        switch (returnValueDescriptor.charAt(0)) {
            case 'I':
                process.setReturnValue(stack.readInt());
                break;
            case 'F':
                process.setReturnValue(stack.readFloat());
                break;
            case 'C':
                process.setReturnValue((char) stack.readShort());
                break;
            case 'S':
                process.setReturnValue(stack.readShort());
                break;
            case 'Z':
                process.setReturnValue((stack.readInt() != 0));
                break;
            case 'B':
                process.setReturnValue(stack.readByte());
                break;
            case 'J':
                process.setReturnValue(stack.readLong());
                break;
            case 'D':
                process.setReturnValue(stack.readDouble());
                break;
            case '[':
            case 'L':
                Reference reference = references.getReference(stack.readRef());
                process.setReturnValue(referenceConverter.toJavaObject(reference, null));
                break;
            case 'V':
            default:
                process.setReturnValue(null);
                break;
        }
    }



    public boolean oneStep(Process process) throws Exception {
        boolean execute = true;
        if (process.getInterpretingContext().isCallStackEmpty()) {
            return false;
        }
        if (process != lastProcess || callStackLevel != process.getInterpretingContext().peekCall()) {
            lastProcess = process;
            classManager = process.getInterpretingContext().getClassManager();
            callStackLevel = process.getInterpretingContext().peekCall();
            if (callStackLevel.isAuxiliaryCall()) {
                process.getInterpretingContext().popCall();
                callStackLevel.runAuxiliaryCall();
                if (process.getInterpretingContext().isCallStackEmpty()) {
                    setReturnValue(process);
                    return false;
                }
                return true;
            } else {
                bytecode = callStackLevel.getBytecode();
                currentClass = callStackLevel.getCurentClass();
                references = callStackLevel.getReferences();
                stack = callStackLevel.getStack();
                localVars = callStackLevel.getLocalVars();
                returnStackPointer = callStackLevel.getReturnStackPointer();
                index = callStackLevel.getProgramCounter();
            }
        }
        if (index < bytecode.length) {
            int x = bytecode[index] & 0xff;
            if (LOG_EVERY_INSTRUCTION && logger.isDebugEnabled()) {
                logger.debug(String.format("        FIB %d - %s, level %03d, index %03d, %s",
                                           cpuIndex,
                                           process.getDebugInfo(),
                                           process.getInterpretingContext().getCallStackSize(),
                                           index,
                                           Const.getName(x)));
            }

            switch ((x)) {

			case Const.AALOAD:	int aaloadIndex = stack.readInt();
					Reference aaloadRef = references.getReference(stack.readInt());
					((Array)aaloadRef).getVariable(currentClass, aaloadIndex, stack, references);
					break;
							// aaload | arrayref, index › value dloads onto the stack
							// a reference from an array

			case Const.AASTORE:
						int aastoreValue = stack.readInt();
						int aastoreIndex = stack.readInt();
						Reference aastoreRef = references.getReference(stack.readInt());
						((Array)aastoreRef).setVariable(currentClass, aastoreIndex, stack, references, aastoreValue);
					break;
							// aastore | arrayref, index, value › stores into a
							// reference to an array

			case Const.ACONST_NULL:	stack.writeRef(references.putReference(new SoftReference(null)));
					break; 	// aconst_null | › null pushes a null reference onto the
							// stack

			case Const.ALOAD:	stack.writeRef(localVars.GetLocalReference(bytecode[++index]));
					break; 	// aload | › objectref loads a reference onto the stack
							// from a local variable #index

			case Const.ALOAD_0:	stack.writeRef(localVars.GetLocalReference(0));
					break; 	// aload_0 | › objectref loads a reference onto the
							// stack from local variable 0

			case Const.ALOAD_1: 	stack.writeRef(localVars.GetLocalReference(1));
					break; 	// aload_1 | › objectref loads a reference onto the
							// stack from local variable 1

			case Const.ALOAD_2:	stack.writeRef(localVars.GetLocalReference(2));
					break; 	// aload_2 | › objectref loads a reference onto the
							// stack from local variable 2

			case Const.ALOAD_3:	stack.writeRef(localVars.GetLocalReference(3));
					break; 	// aload_3 | › objectref loads a reference onto the
							// stack from local variable 3

			case Const.ANEWARRAY:	int anewarrayIndex = (((int)bytecode[++index] & 0xFF) << 8) | (bytecode[++index] & 0xFF);
						Reference anewArray = classManager.createReferenceArray(currentClass, stack, anewarrayIndex, 1);
						stack.write(references.putReference(anewArray));
					break;
							// anewarray | indexbyte1, indexbyte2 count › arrayref
							// creates a new array of references of length count and
							// index (indexbyte1 << 8 + indexbyte2) in the constant
							// pool

			case Const.ARETURN:		int aretvalue = stack.readInt();
							stack.setStackPointer(returnStackPointer);
							stack.write(aretvalue);
							execute = false;

					break; 	// areturn | objectref › [empty] returns a reference
							// from a method

			case Const.ARRAYLENGTH:	Reference arrayLengthRef = references.getReference(stack.readInt());
						int arrayLength = ((Array)arrayLengthRef).getLength();
						stack.write(arrayLength);
					break;
							// arraylength | arrayref › length gets the length of an
							// array

			case Const.ASTORE:	localVars.AddLocalReference(bytecode[++index], stack.readRef());
					break; 	// astore | objectref › stores a reference into a local
							// variable #index

			case Const.ASTORE_0:	localVars.AddLocalReference(0, stack.readRef());
					break; 	// astore_0 | objectref › stores a reference into local
							// variable 0

			case Const.ASTORE_1:	localVars.AddLocalReference(1, stack.readRef());
					break; 	// astore_1 | objectref › stores a reference into local
							// variable 1

			case Const.ASTORE_2:	localVars.AddLocalReference(2, stack.readRef());
					break; 	// astore_2 | objectref › stores a reference into local
							// variable 2

			case Const.ASTORE_3:	localVars.AddLocalReference(3, stack.readRef());
					break; 	// astore_3 | objectref › stores a reference into local
							// variable 3

			case Const.ATHROW:	int throwref = stack.readRef();
						stack.clearStack();
						stack.writeRef(throwref);
					break;
							// athrow | objectref › [empty], objref throws an error
							// or exception (notice that the rest of the stack is
							// cleared, leaving only a reference to the Throwable)

			case Const.BALOAD:	int baloadIndex = stack.readInt();
						Reference baloadRef = references.getReference(stack.readInt());
						((Array)baloadRef).getVariable(currentClass, baloadIndex, stack, references);
					break;
							// baload | arrayref, index › value loads a byte or
							// Boolean value from an array

			case Const.BASTORE:	int bastoreValue = stack.readByte();
						int bastoreIndex = stack.readInt();
						Reference bastoreRef = references.getReference(stack.readInt());
						((Array)bastoreRef).setVariable(currentClass, bastoreIndex, stack, references, bastoreValue);
					break;
							// bastore | arrayref, index, value › stores a byte or
							// Boolean value into an array

			case Const.BIPUSH:	stack.write((int)bytecode[++index]);
					break; 	// bipush byte | › value pushes a byte onto the stack as
							// an integer value

			case Const.CALOAD:	int caloadIndex = stack.readInt();
						Reference caloadRef = references.getReference(stack.readInt());
						((Array)caloadRef).getVariable(currentClass, caloadIndex, stack, references);
					break;
							// caload | arrayref, index › value loads a char from an
							// array
			case Const.CASTORE:	int castoreValue = stack.readInt();
						int castoreIndex = stack.readInt();
						Reference castoreRef = references.getReference(stack.readInt());
						((Array)castoreRef).setVariable(currentClass, castoreIndex, stack, references, castoreValue);
					break;
							// castore | arrayref, index, value › stores a char into
							// an array

			case Const.CHECKCAST:	int chkcastindex = (bytecode[++index] << 8) | bytecode[++index];
						Reference chkCastRef = references.getReference(stack.readInt());
						if (chkCastRef != null)
						{
							boolean chkCastRes = chkCastRef.checkCast(currentClass, /*index*/chkcastindex); // IMHO by se to melo testovat na tenhle index
						}
						//if(chkCastRes)
							stack.write(references.putReference(chkCastRef));
						//else
							//throw new ClassCastException();
					;
							//TODO dodelat
				break; 	// checkcast indexbyte1, indexbyte2
							// | objectref › objectref checks whether an objectref is of a
							// certain type, the class reference of which is in the constant
							// pool at index (indexbyte1 << 8 + indexbyte2)

			case Const.D2F:	float flval = new Double(stack.readDouble()).floatValue();
						stack.write(flval);
					break; 	// d2f | value › result converts a double to a float

			case Const.D2I:	int intval = new Double(stack.readDouble()).intValue();
						stack.write(intval);
					break; 	// d2i | value › result converts a double to an int

			case Const.D2L:	long longval = new Double(stack.readDouble()).longValue();
						stack.write(longval);
					break; // d2l | value › result converts a double to a long

			case Const.DADD:	stack.write(stack.readDouble() + stack.readDouble());
					break; // dadd | value1, value2 › result adds two doubles

			case Const.DALOAD:	int daloadIndex = stack.readInt();
						Reference daloadRef = references.getReference(stack.readInt());
						((Array)daloadRef).getVariable(currentClass, daloadIndex, stack, references);
				break;
						// daload | arrayref, index › value loads a double from
						// an array

			case Const.DASTORE:	long dastoreValue = stack.readLong();
						int dastoreIndex = stack.readInt();
						Reference dastoreRef = references.getReference(stack.readInt());
						((Array)dastoreRef).setVariable(currentClass, dastoreIndex, stack, references, dastoreValue);
				break;
						// dastore | arrayref, index, value › stores a double
						// into an array

			case Const.DCMPL:	double dcmplval2 = stack.readDouble();
						double dcmplval1 = stack.readDouble();

						if(Double.isNaN(dcmplval1) || Double.isNaN(dcmplval2))
						{
							stack.write(-1);
							index--;
                            break;
						}

						if(dcmplval1 == dcmplval2)
							stack.write(0);
						else if(dcmplval1 > dcmplval2)
							stack.write(1);
						else stack.write(-1);
					break; // dcmpl | value1, value2 › result compares two doubles

			case Const.DCMPG:	double dcmpgval2 = stack.readDouble();
						double dcmpgval1 = stack.readDouble();

						if(Double.isNaN(dcmpgval1) || Double.isNaN(dcmpgval2))
						{
							stack.write(1);
							index--;
                            break;
						}

						if(dcmpgval1 == dcmpgval2)
							stack.write(0);
						else if(dcmpgval1 > dcmpgval2)
							stack.write(1);
						else stack.write(-1);
					break; // dcmpg | value1, value2 › result compares two doubles


			case Const.DCONST_0:	stack.write(0.0d);
					break; 	// dconst_0 | › 0.0 pushes the constant 0.0 onto the
							// stack

			case Const.DCONST_1:	stack.write(1.0d);
					break; 	// dconst_1 | › 1.0 pushes the constant 1.0 onto the
							// stack

			case Const.DDIV:	double ddivval2 = stack.readDouble();
						double ddivval1 = stack.readDouble();

						stack.write(ddivval1 / ddivval2);
					break; 	// ddiv | value1, value2 › result divides two doubles

			case Const.DLOAD:	double ddval = localVars.GetLocalDouble(bytecode[++index]);
						stack.write(ddval);
					break; 	// dload index | › value loads a double value from a
							// local variable #index

			case Const.DLOAD_0:	double d0val = localVars.GetLocalDouble(0);
						stack.write(d0val);
					break; 	// dload_0 | › value loads a double from local variable 0

			case Const.DLOAD_1:	double d1val = localVars.GetLocalDouble(1);
						stack.write(d1val);
					break; 	// dload_1 | › value loads a double from local variable 1

			case Const.DLOAD_2:	double d2val = localVars.GetLocalDouble(2);
						stack.write(d2val);
					break; 	// dload_2 | › value loads a double from local variable 2

			case Const.DLOAD_3:	double d3val = localVars.GetLocalDouble(3);
						stack.write(d3val);
					break; 	// dload_3 | › value loads a double from local variable 3

			case Const.DMUL:	stack.write(stack.readDouble() * stack.readDouble());
					break; 	// dmul | value1, value2 › result multiplies two doubles

			case Const.DNEG:	stack.write((-1) * stack.readDouble());		//TODO NEGATE DOUBLE
					break; 	// dneg | value › result negates a double

			case Const.DREM: 	double dremval2 = stack.readDouble();
						double dremval1 = stack.readDouble();

						stack.write(dremval1 % dremval2);
					break; 	// drem | value1, value2 › result gets the remainder
							// from a division between two doubles

			case Const.DRETURN:	double dretvalue = stack.readDouble();
						//double ret = new Double(dretvalue).doubleValue();
//						if (log.isTraceEnabled()) {
//							//System.out.println("Metoda vraci double: " + dretvalue);
//							log.trace("Metoda vraci double: " + dretvalue);
//						}

						stack.setStackPointer(returnStackPointer);
						stack.write(dretvalue);
						execute = false;
					break; 	// dreturn | value › [empty] returns a double from a
							// method

			case Const.DSTORE:	localVars.AddLocalDouble( bytecode[++index], stack.readDouble());
					break; 	// dstore index | value › stores a double value into a
							// local variable #index

			case Const.DSTORE_0:	localVars.AddLocalDouble(0, stack.readDouble());
					break; 	// dstore_0 | value › stores a double into local
							// variable 0

			case Const.DSTORE_1:	localVars.AddLocalDouble(1, stack.readDouble());
					break; 	// dstore_1 | value › stores a double into local
							// variable 1

			case Const.DSTORE_2:	localVars.AddLocalDouble(2, stack.readDouble());
					break; 	// dstore_2 | value › stores a double into local
							// variable 2

			case Const.DSTORE_3:	localVars.AddLocalDouble(3, stack.readDouble());
					break; 	// dstore_3 | value › stores a double into local
							// variable 3

			case Const.DSUB:	stack.write(stack.readDouble() - stack.readDouble());
					break; 	// dsub | value1, value2 › result subtracts a double
							// from another

			case Const.DUP:	int dupvalue = stack.readInt();	//TODO cteni bez zmeny position
						stack.write(dupvalue);
						stack.write(dupvalue);
					break; 	// dup | value › value, value duplicates the value on
							// top of the stack

			case Const.DUP_X1:	int dupvalue1 = stack.readInt();
						int dupvalue2 = stack.readInt();
						stack.write(dupvalue1);
						stack.write(dupvalue2);
						stack.write(dupvalue1);
					break; 	// dup_x1 | value2, value1 › value1, value2, value1
							// inserts a copy of the top value into the stack two
							// values from the top

			case Const.DUP_X2:	int dupxvalue3 = stack.readInt();
						int dupxvalue2 = stack.readInt();
						int dupxvalue1 = stack.readInt();

						stack.write(dupxvalue1);
						stack.write(dupxvalue2);
						stack.write(dupxvalue3);
						stack.write(dupxvalue1);
				break; 		// dup_x2 | value3, value2, value1 › value1, value3,
							// value2, value1 inserts a copy of the top value into
							// the stack two (if value2 is double or long it takes
							// up the entry of value3, too) or three values (if
							// value2 is neither double nor long) from the top

			case Const.DUP2:	long dup2value = stack.readLong();
						stack.write(dup2value);
						stack.write(dup2value);
				break; 		// dup2 | {value2, value1} › {value2, value1}, {value2,
							// value1} duplicate top two stack words (two values, if
							// value1 is not double nor long; a single value, if
							// value1 is double or long)

			case Const.DUP2_X1:	int dup2x1value3 = stack.readInt();
						long dup2x1value12 = stack.readLong();

						stack.write(dup2x1value12);
						stack.write(dup2x1value3);
						stack.write(dup2x1value12);
					break; 	// dup2_x1 | value3, {value2, value1} › {value2,
							// value1}, value3, {value2, value1} duplicate two words
							// and insert beneath third word (see explanation above)

			case Const.DUP2_X2: 	long dup2x2value43 = stack.readLong();
						long dup2x2value21 = stack.readLong();

						stack.write(dup2x2value21);
						stack.write(dup2x2value43);
						stack.write(dup2x2value21);
				break; 		// dup2_x2 | {value4, value3}, {value2, value1} ›
							// {value2, value1}, {value4, value3}, {value2, value1}
							// duplicate two words and insert beneath fourth word


			case Const.F2D:	stack.write(new Float(stack.readFloat()).doubleValue());
					break; 	// f2d | value › result converts a float to a double

			case Const.F2I:	stack.write(new Float(stack.readFloat()).intValue());
					break; 	// f2i | value › result converts a float to an int

			case Const.F2L:	stack.write(new Float(stack.readFloat()).longValue());
					break; 	// f2l | value › result converts a float to a long

			case Const.FADD:	stack.write(stack.readFloat() + stack.readFloat());
					break; 	// fadd | value1, value2 › result adds two floats

			case Const.FALOAD:	int faloadIndex = stack.readInt();
						Reference faloadRef = references.getReference(stack.readInt());
						((Array)faloadRef).getVariable(currentClass, faloadIndex, stack, references);
				break;
						// faload | arrayref, index › value loads a float from
						// an array

			case Const.FASTORE:	int fastoreValue = stack.readInt();
						int fastoreIndex = stack.readInt();
						Reference fastoreRef = references.getReference(stack.readInt());
						((Array)fastoreRef).setVariable(currentClass, fastoreIndex, stack, references, fastoreValue);
				break;
						// fastore | arreyref, index, value › stores a float in
						// an array

			case Const.FCMPL:	float fcmplval2 = stack.readFloat();
						float fcmplval1 = stack.readFloat();

						if(Double.isNaN(fcmplval1) || Double.isNaN(fcmplval2))
						{
							stack.write(-1);
							index--;
                            break;
						}

						if(fcmplval1 == fcmplval2)
							stack.write(0);
						else if(fcmplval1 > fcmplval2)
							stack.write(1);
						else stack.write(-1);
					break; // fcmpl | value1, value2 › result compares two floats

			case Const.FCMPG:	float fcmpgval2 = stack.readFloat();
						float fcmpgval1 = stack.readFloat();

						if(Double.isNaN(fcmpgval1) || Double.isNaN(fcmpgval2))
						{
							stack.write(1);
							index--;
                            break;
						}

						if(fcmpgval1 == fcmpgval2)
							stack.write(0);
						else if(fcmpgval1 > fcmpgval2)
							stack.write(1);
						else stack.write(-1);
					break; // fcmpg | value1, value2 › result compares two floats

			case Const.FCONST_0:	stack.write((float)0.0);
					break; 	// fconst_0 | › 0.0f pushes 0.0f on the stack

			case Const.FCONST_1:	stack.write((float)1.0);
					break; 	// fconst_1 | › 1.0f pushes 1.0f on the stack

			case Const.FCONST_2:	stack.write((float)2.0);
					break; 	// fconst_2 | › 2.0f pushes 2.0f on the stack

			case Const.FDIV:	float fdivval2 = stack.readFloat();
						float fdivval1 = stack.readFloat();
						stack.write(fdivval1 / fdivval2);

					break; 	// fdiv | value1, value2 › result divides two floats

			case Const.FLOAD:	stack.write(localVars.GetLocalFloat(bytecode[++index]));
					break; 	// fload index | › value loads a float value from a
							// local variable #index

			case Const.FLOAD_0:	stack.write(localVars.GetLocalFloat(0));
					break; 	// fload_0 | › value loads a float value from local
							// variable 0

			case Const.FLOAD_1:	stack.write(localVars.GetLocalFloat(1));
					break; 	// fload_1 | › value loads a float value from local
							// variable 1

			case Const.FLOAD_2:	stack.write(localVars.GetLocalFloat(2));
					break; 	// fload_2 | › value loads a float value from local
							// variable 2

			case Const.FLOAD_3:	stack.write(localVars.GetLocalFloat(3));
					break; 	// fload_3 | › value loads a float value from local
							// variable 3

			case Const.FMUL:	stack.write(stack.readFloat() * stack.readFloat());
					break; 	// fmul | value1, value2 › result multiplies two floats

			case Const.FNEG:	stack.write((-1) * stack.readFloat());	//TODO negate
					break; // fneg | value › result negates a float

			case Const.FREM:	float fremval2 = stack.readFloat();
						float fremval1 = stack.readFloat();

						stack.write(fremval1 % fremval2);
					break; 	// frem | value1, value2 › result gets the remainder
							// from a division between two floats

			case Const.FRETURN:	float fretvalue = stack.readFloat();
//						if (log.isTraceEnabled()) {
//							//System.out.println("Metoda vraci float: " + fretvalue);
//							log.trace("Metoda vraci float: " + fretvalue);
//						}
						stack.setStackPointer(returnStackPointer);
						stack.write(fretvalue);
						execute = false;
					break; // freturn | value › [empty] returns a float

			case Const.FSTORE:	localVars.AddLocalFloat(bytecode[++index], stack.readFloat());
					break; 	// fstore index | value › stores a float value into a
							// local variable #index

			case Const.FSTORE_0:	localVars.AddLocalFloat(0, stack.readFloat());
					break; 	// fstore_0 | value › stores a float value into local
							// variable 0

			case Const.FSTORE_1:	localVars.AddLocalFloat(1, stack.readFloat());
					break; 	// fstore_1 | value › stores a float value into local
							// variable 1

			case Const.FSTORE_2:	localVars.AddLocalFloat(2, stack.readFloat());
					break; 	// fstore_2 | value › stores a float value into local
							// variable 2

			case Const.FSTORE_3:	localVars.AddLocalFloat(3, stack.readFloat());
					break; 	// fstore_3 | value › stores a float value into local
							// variable 3

			case Const.FSUB:	float fsubvalue2 = stack.readFloat();
						float fsubvalue1 = stack.readFloat();

						stack.write(fsubvalue1 - fsubvalue2);
					break; 	// fsub | value1, value2 › result subtracts two floats

			case Const.GETFIELD:
						int getfieldindex = (((int)bytecode[++index] & 0xFF) << 8) | (bytecode[++index] & 0xFF);
						Reference getFieldRef = references.getReference(stack.readInt());
 						getFieldRef.getVariable(currentClass, getfieldindex, stack, references);

					break;
							// getfield | index1, index2 objectref › value gets a
							// field value of an object objectref, where the field
							// is identified by field reference in the constant pool
							// 	index (index1 << 8 + index2)

			case Const.GETSTATIC:	int getstaticindex = ((bytecode[++index] & 0xFF) << 8) | (bytecode[++index] & 0xFF);
						classManager.getStaticVariable(currentClass, getstaticindex, stack, references);
						//Reference getstaticRef = references.getReference(0);
						//getstaticRef.getVariable(currentClass, getstaticindex, stack, references);
				break;
						// getstatic | index1, index2 › value gets a static
						// field value of a class, where the field is identified
						// by field reference in the constant pool index (index1
						// << 8 + index2)

			case Const.GOTO:	int gotoindex = index;
						int gotoval = (short)(((bytecode[++index] & 0xFF) << 8) | (bytecode[++index] & 0xFF)) ;
						index = gotoindex + gotoval - 1;
				break; // goto branchbyte1, branchbyte2 | [no change] goes to
						// another instruction at branchoffset (signed short
						// constructed from unsigned bytes branchbyte1 << 8 +
						// branchbyte2)

			case Const.GOTO_W:	int gotowval = (bytecode[++index] << 24) | bytecode[++index] << 16 | (bytecode[++index] << 8) | bytecode[++index];
						index += gotowval - 1;
				break; 	// goto_w branchbyte1, branchbyte2, branchbyte3,
						// branchbyte4 | [no change] goes to another instruction
						// at branchoffset (signed int constructed from unsigned
						// bytes branchbyte1 << 24 + branchbyte2 << 16 +
						// branchbyte3 << 8 + branchbyte4)


			case Const.I2B:	stack.write(new Integer(stack.readInt()).byteValue());
					break; 	// i2b | value › result converts an int into a byte

			case Const.I2C:	stack.write(new Integer(stack.readInt()).byteValue());	//TODO tady je to naky zmatecny
					break; 	// i2c | value › result converts an int into a character

			case Const.I2D:	stack.write(new Integer(stack.readInt()).doubleValue());
					break;	 // i2d | value › result converts an int into a double

			case Const.I2F:	stack.write(new Integer(stack.readInt()).floatValue());
					break; 	// i2f | value › result converts an int into a float

			case Const.I2L:	stack.write(new Integer(stack.readInt()).longValue());
					break; 	// i2l | value › result converts an int into a long

			case Const.I2S:	stack.write(new Integer(stack.readInt()).shortValue());
					break; 	// i2s | value › result converts an int into a short

			case Const.IADD:
					stack.write(stack.readInt() + stack.readInt());
					break; 	// iadd | value1, value2 › result adds two ints together

			case Const.IALOAD:	int ialoadIndex = stack.readInt();
						Reference ialoadRef = references.getReference(stack.readInt());
						((Array)ialoadRef).getVariable(currentClass, ialoadIndex, stack, references);
				break;
						// iaload | arrayref, index › value loads an int from an
						// array
			case Const.IAND:	int iandvalue2 = stack.readInt();
						int iandvalue1 = stack.readInt();

						stack.write(iandvalue1 & iandvalue2);
					break; // iand | value1, value2 › result performs a logical and
						// on two integers

			case Const.IASTORE:	int iastoreValue = stack.readInt();
						int iastoreIndex = stack.readInt();
						Reference iastoreRef = references.getReference(stack.readInt());
						((Array)iastoreRef).setVariable(currentClass, iastoreIndex, stack, references, iastoreValue);
				break; // iastore | arrayref, index, value › stores an int into
						// an array

			case Const.ICONST_M1:	stack.write((-1));
					break; 	// iconst_m1 | › -1 loads the int value -1 onto the
							// stack

			case Const.ICONST_0:	stack.write(0);
					break; 	// iconst_0 | › 0 loads the int value 0 onto the stack\

			case Const.ICONST_1:	stack.write(1);
					break; 	// iconst_1 | › 1 loads the int value 1 onto the stack

			case Const.ICONST_2:	stack.write(2);
					break; 	// iconst_2 | › 2 loads the int value 2 onto the stack

			case Const.ICONST_3:	stack.write(3);
					break; 	// iconst_3 | › 3 loads the int value 3 onto the stack

			case Const.ICONST_4:	stack.write(4);
					break; 	// iconst_4 | › 4 loads the int value 4 onto the stack

			case Const.ICONST_5:	stack.write(5);
					break; 	// iconst_5 | › 5 loads the int value 5 onto the stack

			case Const.IDIV:	int idivvalue2 = stack.readInt();
						int idivvalue1 = stack.readInt();

						stack.write(idivvalue1 / idivvalue2);
					break; 	// idiv | value1, value2 › result divides two integers

			case Const.IF_ACMPEQ:	int ifx_acmindex = index;
						int if_acmindex = (((int)bytecode[++index]) << 8) | (bytecode[++index] & 0xFF);

						int if_index2 = stack.readInt();
						int if_index1 = stack.readInt();

						//if(this.GetReference(if_index1).equals((Reference)this.GetReference(if_index2))) //TODO odkomentovat
							//index = ifx_acmindex + if_acmindex - 1;
					break;
							// if_acmpeq | branchbyte1, branchbyte2 value1, value2 ›
							// if references are equal, branch to instruction at
							// branchoffset (signed short constructed from unsigned
							// bytes branchbyte1 << 8 + branchbyte2)

			case Const.IF_ACMPNE:	int ifx_acmneindex = index;
						int if_acmneindex = (((int)bytecode[++index]) << 8) | (bytecode[++index] & 0xFF);

						int if_neindex2 = stack.readInt();
						int if_neindex1 = stack.readInt();

						//if(this.GetReference(if_neindex1).equals((Reference)this.GetReference(if_neindex2)))	//TODO odkomentovat
							//index = ifx_acmneindex + if_acmneindex - 1;
					break;
							// if_acmpne | branchbyte1, branchbyte2 value1, value2 ›
							// if references are not equal, branch to instruction at
							// branchoffset (signed short constructed from unsigned
							// bytes branchbyte1 << 8 + branchbyte2)

			case Const.IF_ICMPEQ:	int ifx_icmpeqindex = index;
						int if_icmpeqindex = (((int)bytecode[++index]) << 8) | (bytecode[++index] & 0xFF);

						int if_icmpeq2 = stack.readInt();
						int if_icmpeq1 = stack.readInt();
						if(if_icmpeq1 == if_icmpeq2)
							index = ifx_icmpeqindex + if_icmpeqindex - 1;
				break;
						// if_icmpeq | branchbyte1, branchbyte2 value1, value2 ›
						// if ints are equal, branch to instruction at
						// branchoffset (signed short constructed from unsigned
						// bytes branchbyte1 << 8 + branchbyte2)

			case Const.IF_ICMPNE:	int ifx_icmpneindex = index;
						int if_icmpneindex = (((int)bytecode[++index]) << 8) | (bytecode[++index] & 0xFF);

						int if_icmpne2 = stack.readInt();
						int if_icmpne1 = stack.readInt();

						if(if_icmpne1 != if_icmpne2)
							index = ifx_icmpneindex + if_icmpneindex - 1;
					break;

						// if_icmpne | branchbyte1, branchbyte2 value1, value2 ›
						// if ints are not equal, branch to instruction at
						// branchoffset (signed short constructed from unsigned
						// bytes branchbyte1 << 8 + branchbyte2)

			case Const.IF_ICMPLT:	int ifx_icmpltindex = index;
						int if_icmpltindex = (((short)bytecode[++index]) << 8) | (bytecode[++index] & 0xFF);

						int if_icmplt2 = stack.readInt();
						int if_icmplt1 = stack.readInt();

						if(if_icmplt1 < if_icmplt2)
							index = ifx_icmpltindex + if_icmpltindex - 1;

					break; 	// if_icmplt | branchbyte1, branchbyte2 value1, value2 ›
							// if value1 is less than value2, branch to instruction
							// at branchoffset (signed short constructed from
							// unsigned bytes branchbyte1 << 8 + branchbyte2)

			case Const.IF_ICMPGE:	int ifx_icmpgeindex = index;
						int if_icmpgeindex = (((int)bytecode[++index]) << 8) | (bytecode[++index] & 0xFF);

						int if_icmpge2 = stack.readInt();
						int if_icmpge1 = stack.readInt();

						if(if_icmpge1 >= if_icmpge2)
							index = ifx_icmpgeindex + if_icmpgeindex - 1;
						if (index < 0)
							System.out.println();
				break;
						// if_icmpge | branchbyte1, branchbyte2 value1, value2 ›
						// if value1 is greater than or equal to value2, branch
						// to instruction at branchoffset (signed short
						// constructed from unsigned bytes branchbyte1 << 8 +
						// branchbyte2)

			case Const.IF_ICMPGT:	int ifx_icmpgtindex = index;
						int if_icmpgtindex = (((int)bytecode[++index]) << 8) | (bytecode[++index] & 0xFF);

						int if_icmpgt2 = stack.readInt();
						int if_icmpgt1 = stack.readInt();

						if(if_icmpgt1 > if_icmpgt2)
							index = ifx_icmpgtindex + if_icmpgtindex - 1;
				break;
						// if_icmpgt | branchbyte1, branchbyte2 value1, value2 ›
						// if value1 is greater than value2, branch to
						// instruction at branchoffset (signed short constructed
						// from unsigned bytes branchbyte1 << 8 + branchbyte2)

			case Const.IF_ICMPLE:	int ifx_icmpleindex = index;
						int if_icmpleindex = (((int)bytecode[++index]) << 8) | (bytecode[++index] & 0xFF);

						int if_icmple2 = stack.readInt();
						int if_icmple1 = stack.readInt();

						if(if_icmple1 <= if_icmple2)
							index = ifx_icmpleindex + if_icmpleindex - 1;
				break;
						// if_icmple | branchbyte1, branchbyte2 value1, value2 ›
						// if value1 is less than or equal to value2, branch to
						// instruction at branchoffset (signed short constructed
						// from unsigned bytes branchbyte1 << 8 + branchbyte2)

			case Const.IFEQ:	int ifx_eqindex = index;
						int if_eqindex = (((int)bytecode[++index]) << 8) | (bytecode[++index] & 0xFF);

						int if_eqvalue = stack.readInt();
						if(if_eqvalue == 0)
							index = ifx_eqindex + if_eqindex - 1;

				break;
						// ifeq | branchbyte1, branchbyte2 value › if value is
						// 0, branch to instruction at branchoffset (signed
						// short constructed from unsigned bytes branchbyte1 <<
						// 8 + branchbyte2)

			case Const.IFNE:	int ifx_neindex = index;
						int if_neindex = (((int)bytecode[++index]) << 8) | (bytecode[++index] & 0xFF);

						int if_nevalue = stack.readInt();
						if(if_nevalue != 0)
							index = ifx_neindex + if_neindex - 1;
				break;
						// ifne | branchbyte1, branchbyte2 value › if value is
						// not 0, branch to instruction at branchoffset (signed
						// short constructed from unsigned bytes branchbyte1 <<
						// 8 + branchbyte2)

			case Const.IFLT:	int ifx_ltindex = index;
						int if_ltindex = (((int)bytecode[++index]) << 8) | (bytecode[++index] & 0xFF);

						int if_ltvalue = stack.readInt();
						if(if_ltvalue < 0)
							index = ifx_ltindex + if_ltindex - 1;
				break;
						// iflt | branchbyte1, branchbyte2 value › if value is
						// less than 0, branch to instruction at branchoffset
						// (signed short constructed from unsigned bytes
						// branchbyte1 << 8 + branchbyte2)

			case Const.IFGE:	int ifx_geindex = index;
						int if_geindex = (((int)bytecode[++index]) << 8) | (bytecode[++index] & 0xFF);

						int if_gevalue = stack.readInt();
						if(if_gevalue >= 0)
							index = ifx_geindex + if_geindex - 1;
				break;
						// ifge | branchbyte1, branchbyte2 value › if value is
						// greater than or equal to 0, branch to instruction at
						// branchoffset (signed short constructed from unsigned
						// bytes branchbyte1 << 8 + branchbyte2)

			case Const.IFGT:	int ifx_gtindex = index;
						int if_gtindex = (((int)bytecode[++index]) << 8) | (bytecode[++index] & 0xFF);

						int if_gtvalue = stack.readInt();
						if(if_gtvalue > 0)
							index = ifx_gtindex + if_gtindex - 1;

				break; 	// ifgt | branchbyte1, branchbyte2 value › if value is
						// greater than 0, branch to instruction at branchoffset
						// (signed short constructed from unsigned bytes
						// branchbyte1 << 8 + branchbyte2)

			case Const.IFLE:	int ifx_leindex = index;
						int if_leindex = (((int)bytecode[++index]) << 8) | (bytecode[++index] & 0xFF);

						int if_levalue = stack.readInt();
						if(if_levalue <= 0)
							index = ifx_leindex + if_leindex - 1;
				break;
						// ifle | branchbyte1, branchbyte2 value › if value is
						// less than or equal to 0, branch to instruction at
						// branchoffset (signed short constructed from unsigned
						// bytes branchbyte1 << 8 + branchbyte2)

			case Const.IFNONNULL:	int ifx_nnindex = index;
						int if_nnindex = (((int)bytecode[++index]) << 8) | (bytecode[++index] & 0xFF);
						int if_nnvalue = stack.readRef();
						Reference toTest = references.getReference(if_nnvalue);
						if(! ((toTest == null) || toTest.isNull()))
							index = ifx_nnindex + if_nnindex - 1;
				break;
						// ifnonnull | branchbyte1, branchbyte2 value › if value
						// is not null, branch to instruction at branchoffset
						// (signed short constructed from unsigned bytes
						// branchbyte1 << 8 + branchbyte2)

			case Const.IFNULL:	int ifx_inindex = index;
						//int if_inindex = (short)((bytecode[++index] << 8) | bytecode[++index]);
						int if_inindex = (((int)bytecode[++index]) << 8) | (bytecode[++index] & 0xFF);

						int if_invalue = stack.readRef();
						Reference if_iniRef = references.getReference(if_invalue);
						if(if_iniRef == null || if_iniRef.isNull())
							index = ifx_inindex + if_inindex - 1;
				break;
						// ifnull branchbyte1, branchbyte2 | value › if value is
						// null, branch to instruction at branchoffset (signed
						// short constructed from unsigned bytes branchbyte1 <<
						// 8 + branchbyte2)

			case Const.IINC:	int iincindex = bytecode[++index] & 0xFF;
						int iinclocalvalue = localVars.GetLocalInt(iincindex);
						localVars.AddLocalInt(iincindex, iinclocalvalue + (bytecode[++index]));
					break; 	// iinc index, const | [No change] increment local
							// variable #index by signed byte const

			case Const.ILOAD:	stack.write(localVars.GetLocalInt(bytecode[++index]));
					break; 	// iload index | › value loads an int value from a
							// variable #index

			case Const.ILOAD_0:	stack.write(localVars.GetLocalInt(0));
					break; 	// iload_0 | › value loads an int value from variable 0

			case Const.ILOAD_1:	stack.write(localVars.GetLocalInt(1));
					break; 	// iload_1 | › value loads an int value from variable 1

			case Const.ILOAD_2:	stack.write(localVars.GetLocalInt(2));
					break; 	// iload_2 | › value loads an int value from variable 2

			case Const.ILOAD_3:	stack.write(localVars.GetLocalInt(3));
					break; 	// iload_3 | › value loads an int value from variable 3

			case Const.IMUL:	stack.write(stack.readInt() * stack.readInt());
					break;	 // imul | value1, value2  result multiply two integers

			case Const.INEG:	stack.write((-1) * stack.readInt());
					break; 	// ineg | value › result negate int

			case Const.INSTANCEOF:
					int instanceofIndex = (((int)bytecode[++index] & 0xFF) << 8) | (bytecode[++index] & 0xFF);
					Reference instanceofRef = references.getReference(stack.readRef());
					if (instanceofRef.isInstanceOf(currentClass, instanceofIndex))
						stack.write((int)1);
					else stack.write((int)1);

				break; // instanceof indexbyte1, indexbyte2 | objectref ›
						// result determines if an object objectref is of a
						// given type, identified by class reference index in
						// constant pool (indexbyte1 << 8 + indexbyte2)

			case Const.INVOKEINTERFACE:	int interfaceMethodIndex = (((int)bytecode[++index] & 0xFF) << 8) | (bytecode[++index] & 0xFF);
						index += 2; //preskakuje count a zero byte
                        currentClass.runMethod(	interfaceMethodIndex,
												references,
												stack, localVars, false, true, process);

				break;
						// invokeinterface indexbyte1, indexbyte2, count, 0 |
						// objectref, [arg1, arg2, ...] › invokes an interface
						// method on object objectref, where the interface
						// method is identified by method reference index in
						// constant pool (indexbyte1 << 8 + indexbyte2)

			case Const.INVOKESPECIAL:
						int specialMethodIndex = ((bytecode[++index] & 0xFF) << 8) | (bytecode[++index] & 0xFF);

						currentClass.runMethod(specialMethodIndex,
												references,
												stack, localVars, false, false, process);
		/*
						MethodInfo specialMethod = currentClass.getMethod(specialMethodIndex);
						int specialArgsSize = 0;
						byte specialBytecode[] = null;
						int specialMaxLocals = 0;

						try
						{
							specialArgsSize 	= specialMethod.getArgumentsLength();
							specialBytecode 	= specialMethod.getByteCode();
							specialMaxLocals 	= specialMethod.getMaxLocals();
						}
						catch (Exception e)
						{
								e.printStackTrace();
						}


						int specialFrom = stack.stackPointer() - specialArgsSize - 4;

						ByteBuffer specialLocalVarsBuffer = ByteBuffer.wrap(Arrays.copyOfRange(stack.getData(), specialFrom, specialFrom + specialMaxLocals * 4));
						InterpreterLocalVars specialLocalVars = new InterpreterLocalVars(specialLocalVarsBuffer);

						this.Run(specialBytecode, specialMethod.getClassInfo(), references, stack, specialLocalVars, specialFrom);
*/
					break;

						// invokespecial indexbyte1, indexbyte2 | objectref,
						// [arg1, arg2, ...] › invoke instance method on object
						// objectref, where the method is identified by method
						// reference index in constant pool (indexbyte1 << 8 +
						// indexbyte2)

			case Const.INVOKESTATIC:
//						if (index == 24)
//							System.out.println();
						int staticMethodIndex = (((int)bytecode[++index] & 0xFF) << 8) | (bytecode[++index] & 0xFF);
						//Reference staticRef = references.getReference(0);
	/*					MethodInfo staticMethod = currentClass.getMethod(staticMethodIndex);
						int staticArgsSize = 0;
						byte staticBytecode[] = null;
						int staticMaxLocals = 0;

						try
						{
							staticArgsSize 	= staticMethod.getArgumentsLength();
							staticBytecode 	= staticMethod.getByteCode();
							staticMaxLocals = staticMethod.getMaxLocals();
						}
						catch (Exception e)
						{
								e.printStackTrace();
						}


						int staticFrom = stack.stackPointer() - staticArgsSize;
						ByteBuffer staticLocalVarsBuffer = ByteBuffer.wrap(Arrays.copyOfRange(stack.getData(), staticFrom, staticFrom + staticMaxLocals * 4));
						LocalVars staticLocalVars = new LocalVars(staticLocalVarsBuffer);

						this.Run(staticBytecode, staticMethod.getClassInfo(), references, stack, staticLocalVars, staticFrom);
						*/
						//if (staticMethodIndex == 785)
						//	System.out.println();
						currentClass.runMethod(staticMethodIndex,
								references,
								stack, localVars, true, false, process);
				break;
						// invokestatic indexbyte1, indexbyte2 | [arg1, arg2,
						// ...] › invoke a static method, where the method is
						// identified by method reference index in constant pool
						// (indexbyte1 << 8 + indexbyte2)
			case Const.INVOKEVIRTUAL:
						int virtualMethodIndex = (((int)bytecode[++index] & 0xFF) << 8) | (bytecode[++index] & 0xFF);
						/*
						MethodInfo virtualMethod = currentClass.getMethod(virtualMethodIndex);
						int virtualArgsSize = 0;
						byte virtualBytecode[] = null;
						int virtualMaxLocals = 0;

						try
						{
							virtualArgsSize 	= virtualMethod.getArgumentsLength();
							virtualBytecode 	= virtualMethod.getByteCode();
							virtualMaxLocals 	= virtualMethod.getMaxLocals();
						}
						catch (Exception e)
						{
								e.printStackTrace();
						}


						int virtualFrom = stack.stackPointer() - virtualArgsSize - 4;

						ByteBuffer virtualLocalVarsBuffer = ByteBuffer.wrap(Arrays.copyOfRange(stack.getData(), virtualFrom, virtualFrom + virtualMaxLocals * 4));
						LocalVars virtualLocalVars = new LocalVars(virtualLocalVarsBuffer);

						this.Run(virtualBytecode, virtualMethod.getClassInfo(), references, stack, virtualLocalVars, virtualFrom);
			*/
						currentClass.runMethod(virtualMethodIndex,
								references,
								stack, localVars, false, true, process);
					break;

						// invokevirtual indexbyte1, indexbyte2 | objectref,
						// [arg1, arg2, ...] › invoke virtual method on object
						// objectref, where the method is identified by method
						// reference index in constant pool (indexbyte1 << 8 +
						// indexbyte2)


			case Const.IOR:	int iorvalue2 = stack.readInt();
						int iorvalue1 = stack.readInt();

						stack.write(iorvalue1 | iorvalue2);
					break; 	// ior | value1, value2 › result logical int or

			case Const.IREM:	int iremvalue2 = stack.readInt();
						int iremvalue1 = stack.readInt();

						int irem = iremvalue1 % iremvalue2;
						stack.write(irem);
				break; // irem | value1, value2 › result logical int remainder

			case Const.IRETURN:	int iretvalue = stack.readInt();
//						if (log.isTraceEnabled()) {
//							//System.out.println("Metoda vraci int : "+ iretvalue);
//							log.trace("Metoda vraci int : "+ iretvalue);
//						}

						stack.setStackPointer(returnStackPointer);
						stack.write(iretvalue);
						execute = false;
				break; // ireturn | value › [empty] returns an integer from a
						// method

			case Const.ISHL:	int ishlvalue2 = stack.readInt();
						int ishlvalue1 = stack.readInt();

						stack.write(ishlvalue1 << ishlvalue2);
					break; 	// ishl | value1, value2 › result int shift left

			case Const.ISHR:	int ishrvalue2 = stack.readInt();
						int ishrvalue1 = stack.readInt();

					stack.write(ishrvalue1 >> ishrvalue2);
					break; 	// ishr | value1, value2 › result int shift right

			case Const.ISTORE:	localVars.AddLocalInt(bytecode[++index], stack.readInt());
					break; 	// istore index | value › store int value into variable
							// #index

			case Const.ISTORE_0:	localVars.AddLocalInt(0, stack.readInt());
					break; 	// istore_0 | value › store int value into variable 0

			case Const.ISTORE_1:	localVars.AddLocalInt(1, stack.readInt());
					break; 	// istore_1 | value › store int value into variable 1

			case Const.ISTORE_2:	localVars.AddLocalInt(2, stack.readInt());
					break; 	// istore_2 | value › store int value into variable 2

			case Const.ISTORE_3:	localVars.AddLocalInt(3, stack.readInt());
					break; 	// istore_3 | value › store int value into variable 3

			case Const.ISUB:
					int val2 = stack.readInt();
					int val1 = stack.readInt();
					stack.write(val1 - val2);
					break; 	// isub | value1, value2 › result int subtract

			case Const.IUSHR:
					int iushrvalue2 = stack.readInt();
					int iushrvalue1 = stack.readInt();
					stack.write(iushrvalue1 >> iushrvalue2);
					break; 	// iushr | value1, value2 › result int shift right

			case Const.IXOR:	stack.write(stack.readInt() ^ stack.readInt());
					break;	// ixor | value1, value2 › result int xor

			case Const.JSR:	int jsrindex = index;
						int jsraddr = (short)((bytecode[++index] << 8) | bytecode[++index]);
						stack.write(index);
						index = jsrindex + jsraddr - 1;
					break;	//	a8	jsr	branchbyte1, branchbyte2	→ address	jump to subroutine at branchoffset (signed short constructed from unsigned bytes branchbyte1 << 8 + branchbyte2) and place the return address on the stack

			case Const.JSR_W:	int jsrwindex = index;
						int jsrwaddr = (short)((bytecode[++index] << 24) | bytecode[++index] << 16 | bytecode[++index] << 8 | bytecode[++index]);
						stack.write(index);
						index = jsrwindex + jsrwaddr - 1;
					break;//	c9	jsr_w	branchbyte1, branchbyte2, branchbyte3, branchbyte4	→ address	jump to subroutine at branchoffset (signed int constructed from unsigned bytes branchbyte1 << 24 + branchbyte2 << 16 + branchbyte3 << 8 + branchbyte4) and place the return address on the stack

			case Const.L2D:	stack.write(new Long(stack.readLong()).doubleValue());
					break;//	8a	l2d		value → result	converts a long to a double

			case Const.L2F:	stack.write(new Long(stack.readLong()).floatValue());
					break;//	89	l2f		value → result	converts a long to a float

			case Const.L2I:	stack.write(new Long(stack.readLong()).intValue());
					break;//	88	l2i		value → result	converts a long to an int

			case Const.LADD:	stack.write(stack.readLong() + stack.readLong());
					break;//	61	ladd		value1, value2 → result	add two longs

			case Const.LALOAD:	int laloadIndex = stack.readInt();
						Reference laloadRef = references.getReference(stack.readInt());
						((Array)laloadRef).getVariable(currentClass, laloadIndex, stack, references);
					break;
							//	2f	laload		arrayref, index → value	load a long from an array

			case Const.LAND:	long landvalue2 = stack.readLong();
						long landvalue1 = stack.readLong();

						stack.write(landvalue1 & landvalue2);
					break;//	7f	land		value1, value2 → result	bitwise and of two longs

			case Const.LASTORE:	long lastoreValue = stack.readLong();
						int lastoreIndex = stack.readInt();
						Reference lastoreRef = references.getReference(stack.readInt());
						((Array)lastoreRef).setVariable(currentClass, lastoreIndex, stack, references, lastoreValue);
					break;//	50	lastore		arrayref, index, value →	store a long to an array

			case Const.LCMP:	long lcmpval2 = stack.readLong();
						long lcmpval1 = stack.readLong();

						if(lcmpval1 == lcmpval2)
							stack.write(0);
						else if(lcmpval1 > lcmpval2)
							stack.write(1);
						else
							stack.write(-1);

					break;//	94	lcmp		value1, value2 → result	compares two longs values

			case Const.LCONST_0:	stack.write(0L);
					break;//	9	lconst_0		→ 0L	pushes the long 0 onto the stack

			case Const.LCONST_1:	stack.write(1L);
					break;//	0a	lconst_1		→ 1L	pushes the long 1 onto the stack

			case Const.LDC: 	int ldc_index = bytecode[++index] & 0xFF;
						//int ldc_val = currentClass.getConstant4B(ldc_index);
						//stack.write(ldc_val);	//TODO select type to get
						currentClass.getConstant(ldc_index, stack, references);
					break;//	12	ldc	index	→ value	pushes a constant #index from a constant pool (String, int, float or class type) onto the stack

			case Const.LDC_W:	int ldcw_index = (short)(((bytecode[++index] & 0xFF) << 8) | (bytecode[++index] & 0xFF));
						//int ldcw_val = currentClass.getConstant4B(ldcw_index);
						//stack.write(ldcw_val); //TODO select type to get
						currentClass.getConstant(ldcw_index, stack, references);
					break;//	13	ldc_w	indexbyte1, indexbyte2	→ value	pushes a constant #index from a constant pool (String, int, float or class type) onto the stack (wide index is constructed as indexbyte1 << 8 + indexbyte2)

			case Const.LDC2_W: 	int ldcw2_index = (short)(((bytecode[++index] & 0xFF) << 8) | (bytecode[++index] & 0xFF));
						//long ldcw2_val = currentClass.getConstant8B(ldcw2_index);
						//stack.write(ldcw2_val);
						currentClass.getConstant(ldcw2_index, stack, references);
					break;//	14	ldc2_w	indexbyte1, indexbyte2	→ value	pushes a constant #index from a constant pool (double or long) onto the stack (wide index is constructed as indexbyte1 << 8 + indexbyte2)

			case Const.LDIV:	long ldivvalue2 = stack.readLong();
						long ldivvalue1 = stack.readLong();
						stack.write(ldivvalue1 / ldivvalue2);
					break;//	6d	ldiv		value1, value2 → result	divide two longs

			case Const.LLOAD:	stack.write(localVars.GetLocalLong(bytecode[++index]));
					break;//	16	lload	index	→ value	load a long value from a local variable #index

			case Const.LLOAD_0:	stack.write(localVars.GetLocalLong(0));
					break;//	1e	lload_0		→ value	load a long value from a local variable 0

			case Const.LLOAD_1:	stack.write(localVars.GetLocalLong(1));
					break;//	1f	lload_1		→ value	load a long value from a local variable 1

			case Const.LLOAD_2:	stack.write(localVars.GetLocalLong(2));
					break;//	20	lload_2		→ value	load a long value from a local variable 2

			case Const.LLOAD_3:	stack.write(localVars.GetLocalLong(3));
					break;//	21	lload_3		→ value	load a long value from a local variable 3

			case Const.LMUL:	stack.write(stack.readLong() * stack.readLong());
					break;//	69	lmul		value1, value2 → result	multiplies two longs

			case Const.LNEG:	stack.write((-1) * stack.readLong());
					break;//	75	lneg		value → result	negates a long

			case Const.LOOKUPSWITCH:
					int gotoIndex = index;
					for(int i = 0; i <= 3; i++)
						if((index + i + 1) % 4 == 0)
						{
							index += i;
							break;
						}

					int defOffset = (short)((bytecode[++index] << 24) | bytecode[++index] << 16 | bytecode[++index] << 8 | bytecode[++index]);
					int numPairs  = (short)((bytecode[++index] << 24) | bytecode[++index] << 16 | bytecode[++index] << 8 | bytecode[++index]);

					int key = stack.readInt();

					boolean matched = false;
					for(int i = 0; i < numPairs; i++)
					{
						int caseKey = (short)((bytecode[++index] << 24) | bytecode[++index] << 16 | bytecode[++index] << 8 | bytecode[++index]);
						if(key == caseKey)
						{
							int offset = (short)((bytecode[++index] << 24) | bytecode[++index] << 16 | bytecode[++index] << 8 | bytecode[++index]);
							index = gotoIndex + offset - 1;
							matched = true;
							break;
						}
						else
						{
							index += 4;
						}
					}

					if(!matched)
						index = gotoIndex + (defOffset & 0xFF) - 1;
					if (index < 0)
						System.out.println();
					break;


					//	ab	lookupswitch	<0-3 bytes padding>, defaultbyte1, defaultbyte2, defaultbyte3, defaultbyte4, npairs1, npairs2, npairs3, npairs4, match-offset pairs...	key →	a target address is looked up from a table using a key and execution continues from the instruction at that address

			case Const.LOR:	stack.write(stack.readLong() | stack.readLong());
					break;//	81	lor		value1, value2 → result	bitwise or of two longs

			case Const.LREM:	long lremvalue2 = stack.readLong();
						long lremvalue1 = stack.readLong();
						stack.write(lremvalue1 % lremvalue2);
					break;//	71	lrem		value1, value2 → result	remainder of division of two longs

			case Const.LRETURN:	long lretvalue = stack.readLong();
//						if (log.isTraceEnabled()) {
//							//System.out.println("Metoda vraci long: " + lretvalue);
//							log.trace("Metoda vraci long: " + lretvalue);
//						}

						stack.setStackPointer(returnStackPointer);
						stack.write(lretvalue);
						execute = false;
					break;//	ad	lreturn		value → [empty]	returns a long value

			case Const.LSHL:	long lshlvalue2 = stack.readInt();
						long lshlvalue1 = stack.readLong();

						stack.write(lshlvalue1 << lshlvalue2);
					break;//	79	lshl		value1, value2 → result	bitwise shift left of a long value1 by value2 positions

			case Const.LSHR:	long lshrvalue2 = stack.readInt();
						long lshrvalue1 = stack.readLong();

						stack.write(lshrvalue1 >> lshrvalue2);
					break;//	7b	lshr		value1, value2 → result	bitwise shift right of a long value1 by value2 positions

			case Const.LSTORE:	localVars.AddLocalLong(bytecode[++index], stack.readLong());
					break;//	37	lstore	index	value →	store a long value in a local variable #index

			case Const.LSTORE_0: localVars.AddLocalLong(0, stack.readLong());
					break;//	3f	lstore_0		value →	store a long value in a local variable 0

			case Const.LSTORE_1:	localVars.AddLocalLong(1, stack.readLong());
					break;//	40	lstore_1		value →	store a long value in a local variable 1

			case Const.LSTORE_2:	localVars.AddLocalLong(2, stack.readLong());
					break;//	41	lstore_2		value →	store a long value in a local variable 2

			case Const.LSTORE_3:	localVars.AddLocalLong(3, stack.readLong());
					break;//	42	lstore_3		value →	store a long value in a local variable 3

			case Const.LSUB:	long lsubvalue2 = stack.readLong();
						long lsubvalue1 = stack.readLong();

						stack.write(lsubvalue1 - lsubvalue2);
					break;//	65	lsub		value1, value2 → result	subtract two longs

			case Const.LUSHR:	long lushrvalue2 = stack.readInt();
						long lushrvalue1 = stack.readLong();

						stack.write(lushrvalue1 >>> lushrvalue2);
					break;//	7d	lushr		value1, value2 → result	bitwise shift right of a long value1 by value2 positions, unsigned

			case Const.LXOR:	long lxorvalue2 = stack.readLong();
						long lxorvalue1 = stack.readLong();

						stack.write(lxorvalue1 ^ lxorvalue2);
					break;//	83	lxor		value1, value2 → result	bitwise exclusive or of two longs

			case Const.MONITORENTER:
                    Reference enterMonitorReference = references.getReference(stack.readRef());
                    process.monitorEnter(MonitorValidator.validateReference(enterMonitorReference));
					break;//	c2	monitorenter		objectref →	enter monitor for object ("grab the lock" - start of synchronized() section)
			case Const.MONITOREXIT:
                    Reference exitMonitorReference = references.getReference(stack.readRef());
                    process.monitorExit(MonitorValidator.validateReference(exitMonitorReference));
					break;//	c3	monitorexit		objectref →	exit monitor for object ("release the lock" - end of synchronized() section)


			case Const.MULTIANEWARRAY:	int multiArrayIndex = (bytecode[++index] << 8) | bytecode[++index];
						int multiArrayDim = bytecode[++index];
						Reference multiArrayRef = classManager.createReferenceArray(currentClass, stack, multiArrayIndex, multiArrayDim);
						stack.write(references.putReference(multiArrayRef));
					break;//	c5	multianewarray	indexbyte1, indexbyte2, dimensions	count1, [count2,...] → arrayref	create a new array of dimensions dimensions with elements of type identified by class reference in constant pool index (indexbyte1 << 8 + indexbyte2); the sizes of each dimension is identified by count1, [count2, etc]

			case Const.NEW:
						int indexnew = (((int)bytecode[++index] & 0xFF) << 8) | (bytecode[++index] & 0xFF);
						Reference newRef = this.classManager.createReference(currentClass.getClassType(indexnew));
						int newRefIndex = references.putReference(newRef);
						stack.writeRef(newRefIndex);

					break;
							//	bb	new	indexbyte1, indexbyte2	→ objectref	creates new object of type identified by class reference in constant pool index (indexbyte1 << 8 + indexbyte2)

			case Const.NEWARRAY:	int newArrayAType = bytecode[++index];
						//int newarrayCount = stack.readInt();
						int newarrayRefToStack = references.putReference(this.classManager.createArray(currentClass, stack, newArrayAType, 1));
						stack.write(newarrayRefToStack);
					break;

						//	bc	newarray	atype	count → arrayref	creates new array with count elements of primitive type identified by atype
			case Const.NOP:
					break;//	0	nop		[No change]	performs no operation

			case Const.POP:	stack.readInt();
					break;//	57	pop		value →	discards the top value on the stack

			case Const.POP2:	stack.readInt();
						stack.readInt();
					break;//	58	pop2		{value2, value1} →	discards the top two values on the stack (or one value, if it is a double or long)

			case Const.PUTFIELD:

						int putFieldIndex = (((int)bytecode[++index] & 0xFF) << 8) | bytecode[++index] & 0xFF;
						int putfieldVarLen = currentClass.getVariableLength(putFieldIndex);
						long putfieldValue;

						if(putfieldVarLen > 4)
							putfieldValue = stack.readLong();
						else
							putfieldValue = stack.readInt();

						Reference putfieldRef = references.getReference(stack.readInt());
						putfieldRef.setVariable(currentClass, putFieldIndex, stack, references, putfieldValue);

					break;
						//	b5	putfield	indexbyte1, indexbyte2	objectref, value →	set field to value in an object objectref, where the field is identified by a field reference index in constant pool (indexbyte1 << 8 + indexbyte2)

			case Const.PUTSTATIC:	int putstaticIndex = (((int)bytecode[++index] & 0xFF) << 8) | bytecode[++index] & 0xFF;
						//int putstaticVarLen = currentClass.getVariableLength(putstaticIndex);
						//long putstaticValue;

						//if(putstaticVarLen > 4)
						//	putstaticValue = stack.readLong();
						//else
						//	putstaticValue = stack.readInt();
						classManager.setStaticVariable(currentClass, putstaticIndex, stack, references);
						//references.getReference(0).setVariable(currentClass, putstaticIndex, stack, references, putstaticValue);

					break;

					//	b3	putstatic	indexbyte1, indexbyte2	value →	set static field to value in a class, where the field is identified by a field reference index in constant pool (indexbyte1 << 8 + indexbyte2)
			case Const.RET:
					break;//	a9	ret	index	[No change]	continue execution from address taken from a local variable #index (the asymmetry with jsr is intentional)

			case Const.RETURN:	stack.setStackPointer(returnStackPointer);
//						if (log.isTraceEnabled()) {
//							//System.out.println("Return from void method.");
//							log.trace("Return from void method.");
//						}
						execute = false;;
					break;//	b1	return		→ [empty]	return void from method

			case Const.SALOAD:	int saloadIndex = stack.readInt();
						Reference saloadRef = references.getReference(stack.readInt());
						((Array)saloadRef).getVariable(currentClass, saloadIndex, stack, references);
					break;
						//	35	saload		arrayref, index → value	load short from array

			case Const.SASTORE:	short sastoreValue = stack.readShort();
						int sastoreIndex = stack.readInt();
						Reference sastoreRef = references.getReference(stack.readInt());
						((Array)sastoreRef).setVariable(currentClass, sastoreIndex, stack, references, (long)sastoreValue);
					break;
						//	56	sastore		arrayref, index, value →	store short to array

			case Const.SIPUSH:	int sipushval = ((bytecode[++index] << 8) | ((bytecode[++index]) & 0xFF));
						stack.write(sipushval);
					break;//	11	sipush	byte1, byte2	→ value	pushes a signed integer (byte1 << 8 + byte2) onto the stack

			case Const.SWAP:	int swapval2 = stack.readInt();
						int swapval1 = stack.readInt();
						stack.write(swapval2);
						stack.write(swapval1);
					break;//	5f	swap		value2, value1 → value1, value2	swaps two top words on the stack (note that value1 and value2 must not be double or long)

			case Const.TABLESWITCH:
					int tableSwitchGotoIndex = index;
					for(int i = 0; i <= 3; i++)
						if((index + i + 1) % 4 == 0)
						{
							index += i;
							break;
						}

					int tableSwitchdefOffset = (short)((bytecode[++index] << 24) | bytecode[++index] << 16 | bytecode[++index] << 8 | bytecode[++index]);
					int tableSwitchLow  = (short)((bytecode[++index] << 24) | bytecode[++index] << 16 | bytecode[++index] << 8 | bytecode[++index]);
					int tableSwitchHigh  = (short)((bytecode[++index] << 24) | bytecode[++index] << 16 | bytecode[++index] << 8 | bytecode[++index]);

					int tableSwitchVal = stack.readInt();

					if (tableSwitchVal < tableSwitchLow || tableSwitchVal > tableSwitchHigh)
					{
					    index = tableSwitchGotoIndex + tableSwitchdefOffset - 1;              // branch to default
					}
					else
					{
						index += 4 * (tableSwitchVal - tableSwitchLow);
						int tableSwitchOffset = (short)((bytecode[++index] << 24) | bytecode[++index] << 16 | bytecode[++index] << 8 | bytecode[++index]);
						index = tableSwitchGotoIndex + tableSwitchOffset - 1;     // branch to entry in table
					}

					break;
					//	aa	tableswitch	[0-3 bytes padding], defaultbyte1, defaultbyte2, defaultbyte3, defaultbyte4, lowbyte1, lowbyte2, lowbyte3, lowbyte4, highbyte1, highbyte2, highbyte3, highbyte4, jump offsets...	index →	continue execution from an address in the table at offset index
			case Const.WIDE:
					break;//	c4	wide	opcode, indexbyte1, indexbyte2

			}

            index++;
		} else {
            execute = false;
        }
        callStackLevel.setProgramCounter(index);
        if (!execute) {
            process.getInterpretingContext().popCall();
            if (process.getInterpretingContext().isCallStackEmpty()) {
                setReturnValue(process);
                return false;
            }
        }
        return true;
    }
}
