package san.interpreter.sanint.fiber.staticm;



import java.nio.ByteBuffer;
import java.util.HashMap;
import java.util.Map;
import san.interpreter.sanint.classLoading.Reference;
import san.interpreter.sanint.classLoading.ReferenceConverter;
import san.interpreter.sanint.classLoading.StaticVariableInfo;
import san.interpreter.sanint.fiber.InterpretingException;
import san.interpreter.sanint.interpreter.InterpreterStack;
import san.interpreter.sanint.interpreter.References;



/**
 *
 * @author Jiri Nohavec
 */
public class StaticManager {



    private static StaticManager staticManagerSingleton;



    public static StaticManager getInstance() {
        if (staticManagerSingleton == null) {
            staticManagerSingleton = new StaticManager();
        }
        return staticManagerSingleton;
    }



    private static final String KEY_FORMAT_STRING = "%s#%s#_:-)_#%c";
    private static final String[] EXCLUDE_PACKAGE_ARRAY = {
        "java",
        "javax",
        "org",
        "sun",
        "san"
    };

    private final Map<String, ByteBuffer> byteBufferMap;
    private final Map<String, Reference[]> referenceArrayMap;



    private StaticManager() {
        byteBufferMap = new HashMap<String, ByteBuffer>();
        referenceArrayMap = new HashMap<String, Reference[]>();
        for (int i = 0; i < EXCLUDE_PACKAGE_ARRAY.length; i++) {
            EXCLUDE_PACKAGE_ARRAY[i].replace(".", "/");
            if (!EXCLUDE_PACKAGE_ARRAY[i].endsWith("/")) {
                EXCLUDE_PACKAGE_ARRAY[i] += "/";
            }
        }
    }



    private boolean isExcludedClass(String className) {
        for (String jrePackageName: EXCLUDE_PACKAGE_ARRAY) {
            if (className.startsWith(jrePackageName)) {
                return true;
            }
        }
        return false;
    }



    private String getKey(String className, String variableName, StaticVariableInfo staticVariableInfo) {
        return String.format(KEY_FORMAT_STRING, className, variableName, staticVariableInfo.getType());
    }



    public synchronized boolean getStatic(String className,
                                          String variableName,
                                          StaticVariableInfo staticVariableInfo,
                                          InterpreterStack stack,
                                          References references,
                                          int dataLength,
                                          int referenceCount,
                                          ReferenceConverter referenceConverter) {
        if (isExcludedClass(className)) {
            return false;
        }
        String key = getKey(className, variableName, staticVariableInfo);
        ByteBuffer byteBuffer = null;
        Reference[] referenceArray = null;
        switch (staticVariableInfo.getType()) {
            case 'I':
            case 'F':
            case 'C':
            case 'S':
            case 'Z':
            case 'B':
            case 'J':
            case 'D':
                byteBuffer = byteBufferMap.get(key);
                if (byteBuffer == null) {
                    throw new InterpretingException("Accessing uninitialized static variable");
                }
                break;
            case '[':
            case 'L':
                referenceArray = referenceArrayMap.get(key);
                if (referenceArray == null) {
                    throw new InterpretingException("Accessing uninitialized static reference");
                }
                break;
        }
        switch (staticVariableInfo.getType()) {
            case 'I':
            case 'F':
                int intData = byteBuffer.getInt(staticVariableInfo.getOffset());
                stack.write(intData);
                break;
            case 'C':
            case 'S':
                short shortData = byteBuffer.getShort(staticVariableInfo.getOffset());
                stack.write((int) shortData);
                break;
            case 'Z':
            case 'B':
                byte byteData = byteBuffer.get(staticVariableInfo.getOffset());
                stack.write((int) byteData);
                break;
            case 'J':
            case 'D':
                long longData = byteBuffer.getLong(staticVariableInfo.getOffset());
                stack.write(longData);
                break;
            case '[':
            case 'L':
                Reference reference = referenceArray[staticVariableInfo.getOffset()];
                stack.write(references.putReference(reference));
                break;
        }
        return true;
    }



    public synchronized boolean setStatic(String className,
                                          String variableName,
                                          StaticVariableInfo staticVariableInfo,
                                          InterpreterStack stack,
                                          References references,
                                          int dataLength,
                                          int referenceCount,
                                          ReferenceConverter referenceConverter) {
        if (isExcludedClass(className)) {
            return false;
        }
        String key = getKey(className, variableName, staticVariableInfo);
        ByteBuffer byteBuffer = null;
        Reference[] referenceArray = null;
        switch (staticVariableInfo.getType()) {
            case 'I':
            case 'F':
            case 'C':
            case 'S':
            case 'Z':
            case 'B':
            case 'J':
            case 'D':
                byteBuffer = byteBufferMap.get(key);
                if (byteBuffer == null) {
                    byteBuffer = ByteBuffer.allocate(Math.max(1, dataLength));
                    byteBufferMap.put(key, byteBuffer);
                }
                break;
            case '[':
            case 'L':
                referenceArray = referenceArrayMap.get(key);
                if (referenceArray == null) {
                    referenceArray = new Reference[referenceCount];
                    referenceArrayMap.put(key, referenceArray);
                }
                break;
        }
        switch (staticVariableInfo.getType()) {
            case 'I':
            case 'F':
                int intData = stack.readInt();
                byteBuffer.putInt(staticVariableInfo.getOffset(), intData);
                break;
            case 'C':
            case 'S':
                short shortData = (short) stack.readInt();
                byteBuffer.putShort(staticVariableInfo.getOffset(), shortData);
                break;
            case 'Z':
            case 'B':
                byte byteData = (byte) stack.readInt();
                byteBuffer.put(staticVariableInfo.getOffset(), byteData);
                break;
            case 'J':
            case 'D':
                long longData = stack.readLong();
                byteBuffer.putLong(staticVariableInfo.getOffset(), longData);
                break;
            case '[':
            case 'L':
                int referenceIndex = stack.readRef();
                Reference oldReference = referenceArray[staticVariableInfo.getOffset()];
                Reference newReference = references.getReference(referenceIndex);
                try {
                    newReference = StaticReferenceConverter.convert(newReference, referenceConverter);
                } catch (Exception e) {
                }
                if (!StaticReferenceConverter.areBothSoftReferencesToObject(newReference, oldReference)) {
                    referenceArray[staticVariableInfo.getOffset()] = newReference;
                }
                break;
        }
        return true;
    }
}
