/*************************************************************************
 *                           Smart Active Node                           * 
 *                 copyright (c) since 2007 KIV-FAV-ZCU                  *
 *************************************************************************
 * The license for this code is included in LICENSE.txt.  If you did not *
 * get this file with this distribution, request a copy from:            *
 * txkoutny@kiv.zcu.cz                                                   *
 *************************************************************************/
package san.networkInterfaces;

import java.io.IOException;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.SocketException;

import san.ident.NetIdentifier;
import san.localization.Localization;
import san.logging.LoggerFactory;
import san.logging.ILogger.LogType;
import san.networkInterfaces.IncomingPacket.Head;

/**
 * <b>Title:</b> Smart Active Node<br>
 * <b>Description:</b> Describe remote network interface.<br>
 * @author Jan Boháč, Michal Rejda
 * @since 27.11.2006
 * @version 1.0
 */
public class RemoteInterface {

	private byte[] head;
	private InetAddress inetAddress;
	private int port; // port to sending
	private int communicationPort;
	/**
	 * @uml.property name="method"
	 * @uml.associationEnd multiplicity="(1 1)"
	 *                     inverse="remoteInterface:san.networkInterfaces.AccessibleMethod"
	 */
	private AccessibleMethod method = null;
	/**
	 * @uml.property name="metric"
	 */
	private int metric = 1;
	private NetIdentifier identifier;
	/**
	 * get the time of creating this network remote interface
	 */
	public long created;
	/**
	 * Associated connection with remote interface. If the connection is TCP is
	 * this object ServerSocket, if is connection UDP is this object
	 * DatagramSocket.
	 * @uml.property name="connection"
	 * @uml.associationEnd multiplicity="(1 1)"
	 *                     inverse="remoteInterface:san.networkInterfaces.Connection"
	 */
	private Connection connection = null;

	public RemoteInterface() {
		inetAddress = null;
		port = -1;
		communicationPort = -1;
		method = null;
		created = System.currentTimeMillis();
		head = null;
	}

	public RemoteInterface(InetAddress address, int port,
			int comunnicationPort, AccessibleMethod method, byte[] head) {
		this.inetAddress = address;
		this.port = port;
		this.communicationPort = comunnicationPort;
		this.method = method;
		this.head = head;
		this.created = System.currentTimeMillis();
	}

	/**
	 * Return a true if a head feature is set.
	 */
	public boolean isSet(Head headFeature) {
		return IncomingPacket.isSet(head, headFeature);
	}

	/**
	 * Create a object represent the connection to the remote interface.
	 */
	public Connection createConnection() {
		if (connection != null) {
			throw new RuntimeException(Localization.getLocalization()
					.getString("M000045"));
		}
		try {
			switch (method.number) {
				case AccessibleMethod.TCP :
					connection = new TcpConnection(0);
					break;
				case AccessibleMethod.UDP :
					connection = new UdpConnection();
					break;
			}
		}
		catch (SocketException e) {
			LoggerFactory.createLogger().Log(e, LogType.ERROR);
		}
		catch (IOException e) {
			LoggerFactory.createLogger().Log(e, LogType.ERROR);
		}
		return connection;
	}

	/**
	 * Cancel current connection associated to this interface.
	 */
	public void cancelConnection() {
		if (connection == null) {
			return;
		}
		try {
			if (connection instanceof ServerSocket)
				((ServerSocket) connection).close();
			else if (connection instanceof DatagramSocket)
				((DatagramSocket) connection).disconnect();
		}
		catch (IOException e) {
			LoggerFactory.createLogger().Log(e, LogType.ERROR);
		}
		finally {
			connection = null;
		}
	}
	
	void createInputOutput() {

		switch (getMethod().number) {
			case AccessibleMethod.TCP :
				netOut = new TcpOut(this);
				break;
			case AccessibleMethod.UDP :
				netOut = new UdpOut(this);
				break;
		}
		
		switch (getMethod().number) {
			case AccessibleMethod.TCP :
				netListener = new TcpListener(this);
//				LoggerFactory.createLogger().Log("TCP", LogType.DEBUG);
				break;
			case AccessibleMethod.UDP :
				netListener = new UdpListener(this);
//				LoggerFactory.createLogger().Log("UDP", LogType.DEBUG);
				break;
		}
		
		if(netListener != null)
			netListener.listen();
	}

	/**
	 * Getter of the property <tt>connection</tt>
	 * @return Returns the connection.
	 * @uml.property name="connection"
	 */
	public Connection getConnection() {
		return connection;
	}

	/**
	 * HashCalculation is from InetAddress.
	 */
	public int hashCode() {
		return inetAddress.hashCode();
	}

	/**
	 * Equals is from InetAddress.
	 */
	public boolean equals(Object arg0) {
		return inetAddress.equals(((RemoteInterface) arg0).inetAddress);
	}

	public NetIdentifier getIdentifier() {
		return identifier;
	}

	public void setIdentifier(NetIdentifier identifier) {
		this.identifier = identifier;
	}

	/**
	 * @return the inetAddress
	 */
	public InetAddress getInetAddress() {
		return inetAddress;
	}

	/**
	 * @param inetAddress the inetAddress to set
	 */
	public void setInetAddress(InetAddress address) {
		this.inetAddress = address;
	}

	/**
	 * Getter of the property <tt>method</tt>
	 * @return Returns the method.
	 * @uml.property name="method"
	 */
	public AccessibleMethod getMethod() {
		return method;
	}

	/**
	 * Setter of the property <tt>method</tt>
	 * @param method The method to set.
	 * @uml.property name="method"
	 */
	public void setMethod(AccessibleMethod method) {
		this.method = method;
	}

	/**
	 * @return the port
	 */
	public int getPort() {
		return port;
	}

	/**
	 * @param port the port to set
	 */
	public void setPort(int port) {
		this.port = port;
	}

	/**
	 * @return the head
	 */
	public byte[] getHead() {
		return head;
	}

	/**
	 * @param head the head to set
	 */
	public void setHead(byte[] head) {
		this.head = head;
	}

	/**
	 * Setter of the property <tt>connection</tt>
	 * @param connection The connection to set.
	 * @uml.property name="connection"
	 */
	public void setConnection(Connection connection) {
		this.connection = connection;
	}

	/**
	 * @return the communicationPort
	 */
	int getCommunicationPort() {
		return communicationPort;
	}

	/**
	 * @param communicationPort the communicationPort to set
	 */
	void setCommunicationPort(int communicationPort) {
		this.communicationPort = communicationPort;
	}

	/**
	 * @uml.property name="netOut"
	 * @uml.associationEnd readOnly="true" multiplicity="(1 1)"
	 *                     aggregation="composite"
	 *                     inverse="remoteInterface:san.networkInterfaces.NetOut"
	 */
	private NetOut netOut = null;

	/**
	 * Getter of the property <tt>netOut</tt>
	 * @return Returns the netOut.
	 * @uml.property name="netOut"
	 */
	public NetOut getNetOut() {
		return netOut;
	}

	/**
	 * @uml.property name="netListener"
	 * @uml.associationEnd readOnly="true" multiplicity="(1 1)"
	 *                     aggregation="composite"
	 *                     inverse="remoteInterface:san.networkInterfaces.NetListener"
	 */
	private NetListener netListener = null;

	/**
	 * Getter of the property <tt>netListener</tt>
	 * @return Returns the netListener.
	 * @uml.property name="netListener"
	 */
	public NetListener getNetListener() {
		return netListener;
	}

	/**
	 * Getter of the property <tt>metric</tt>
	 * @return Returns the metric.
	 * @uml.property name="metric"
	 */
	public int getMetric() {
		return metric;
	}

	/**
	 * Setter of the property <tt>metric</tt>
	 * @param metric The metric to set.
	 * @uml.property name="metric"
	 */
	public void setMetric(int metric) {
		this.metric = metric;
	}
}
